/*
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM.  Reproduction, modification and redistribution are prohibited.
 * TreeNode.js
 * (c) Copyright IBM Corporation 2012, 2016.  
 * U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
(function(){
dojo.provide("com.ibm.rdm.web.linksexplorer.TreeNode");

dojo.require("dijit.layout.ContentPane");
dojo.require("dijit._Widget");
dojo.require("dijit._Templated"); 
dojo.require("dojo.dnd.TimedMoveable");

dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.common.Menu");
dojo.require("com.ibm.rdm.web.common.MessageForm");
dojo.require("com.ibm.rdm.web.viewer.viewerutils");
dojo.require("com.ibm.rdm.web.project.EditManager");
dojo.require("com.ibm.rdm.web.links.LinkItemWidget");
dojo.require("com.ibm.rdm.web.links.LinkTypeUtils");
dojo.require("com.ibm.rdm.web.linksexplorer.NodeLinkItemWidget");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.server.client.utils.ResourceUtil");
dojo.require("com.ibm.rdm.web.linksexplorer.layout.TreeLayout");
dojo.require("com.ibm.rdm.web.server.cache.ProcessOperationsCache");
dojo.require("com.ibm.rdm.web.common.ResourceLinksHandler");
dojo.require("com.ibm.rdm.web.util.vvcutils");
dojo.require("com.ibm.rdm.web.server.Headers");

dojo.requireLocalization("com.ibm.rdm.web.linksexplorer", "LinksExplorer");

var RDMClient = com.ibm.rdm.web.server.client.RDMClient;
var RRMTypes = com.ibm.rdm.web.model.types.RRMTypes;
var EditManager = com.ibm.rdm.web.project.EditManager;
var ResourceUtil = com.ibm.rdm.web.server.client.utils.ResourceUtil;
var LinkTypeUtils = com.ibm.rdm.web.links.LinkTypeUtils;
var TreeLayout = com.ibm.rdm.web.linksexplorer.layout.TreeLayout;
var ProcessOperationsCache = com.ibm.rdm.web.server.cache.ProcessOperationsCache;
var ResourceLinksHandler = com.ibm.rdm.web.common.ResourceLinksHandler;
var VVCUtils = com.ibm.rdm.web.util.vvcutils;
var Headers = com.ibm.rdm.web.server.Headers;

dojo.declare("com.ibm.rdm.web.linksexplorer.TreeNode", [dijit._Widget, dijit._Templated], {

	// Constants
	NODE_LABEL_TRUNCATE_LENGTH: 150,
	COMPACT_NODE_LABEL_TRUNCATE_LENGTH: 20,
	
	// moveable: Boolean
	// if true, the node is movable by either the containerNode, or an optional node
	// found by the handle attribute
	moveable: true,
	
	// handle: String
	//leave blank to enable entire node as drag handle
	handle: "", 
	
	// checkSuspect: Boolean
	// if true, the node link will be checked for suspect status
	checkSuspect: false,
	
	// Tree Layout necessary properties:
	width: 175,		
	height: 75,	
	xPos: 0,
	yPos: 0,
	prelim: 0,
	modifier: 0,
	leftNeighbor: null,
	rightNeighbor: null,
	parent: null,
	children: null,
	
	link: null,
	modified: null,
	coreData: null,
	linkType: null,
	resourceInfo: null,
	mover: null,
	compactView: false,
	innerLink: null,
	resource: null,
	editLinkMode: false,
	locked: false,
	inputTag: null,
	oldTable: null,
	uncompactInnerLink: null,
	menu: null,
	linkTitleNode: null,
	originalLinkInfo: null,
	
	quickCreateTempNode: false,
	externalLink: false,
	refreshOnExpand: true,
	DMlink: false,	//true if this node represents a Design Manager link
	isCLMLink: false,
	
	_explored: true,
	_connectorRadius: 5,
	_strokeWidth: 2,
	_radius: 5,			// radius of the rounded rectangular corners
	_viewer: null,		
	_linkItemWidget: null,		//the interior link item widget
	_linkTypes: null,
	_connectHandles: null,
	
	// period of hover before we higlight link relationships. Highlighting
	// can be expensive so we don't want to trigger if the mouse is moving
	// across the node.
	HIGHLIGHT_DELAY: 350,
	
	templatePath: dojo.moduleUrl("com.ibm.rdm.web.linksexplorer", "templates/TreeNode.html"),
	
	constructor: function() {
		this.children = [];
		this._connectHandles = [];
	},
	
	startup: function(){
		var that = this;
			
		this.link = {
			line: null,
			label: null
		};
		
		var stopPropagation =  function (e) {
			dojo.stopEvent(e);
		};
		
		this.connect(this.innerContainer, "onmouseenter", function () {
			/*if (this.compactView) {
				dojo.removeClass(this.domNode, "compact");
				this._containerNode.getElementsByTagName('a')[0].innerHTML = this.innerLink.innerHTML;
				dojo.addClass(this.domNode.getElementsByTagName('div')[0],"innerStroke");
			}
			*/
			if (this._hoverTimer) {
				clearTimeout(this._hoverTimer);
				delete this._hoverTimer;
			}
			
			this._hoverTimer = setTimeout(dojo.hitch(this, function () {
				clearTimeout(this._hoverTimer);
				delete this._hoverTimer;
				dojo.addClass(this._expandCollapseBtn, "hover");
				dojo.publish("linksexplorer/highlight_links", [this]);
			}), this.HIGHLIGHT_DELAY); 
		});
		
		this.connect(this.innerContainer, "onmouseleave", function () {
			/*if (this.compactView) {
				dojo.addClass(this.domNode, "compact");
				this._containerNode.getElementsByTagName('a')[0].innerHTML = this.innerLink.innerHTML.substring(0,this.COMPACT_NODE_LABEL_TRUNCATE_LENGTH)+"...";
				dojo.removeClass(this.domNode.getElementsByTagName('div')[0],"innerStroke");
			}
			*/
			if (this._hoverTimer) {
				clearTimeout(this._hoverTimer);
				delete this._hoverTimer;
			}	else {
				dojo.removeClass(this._expandCollapseBtn, "hover");
				dojo.publish("linksexplorer/unhighlight_links", [this]);
			}
		});
		this.connect(this.innerContainer, "onclick", function () {
			com.ibm.rdm.web.linksexplorer.LinksDialogKeys.nodeClicked(this);
		});

		this.connect(this._expandCollapseBtn, "onmouseenter", function () {
			dojo.addClass(this._expandCollapseBtn, "hover");
		});
		
		this.connect(this._expandCollapseBtn, "onmouseleave", function () {
			dojo.removeClass(this._expandCollapseBtn, "hover");
		});

		
		this.connect(this._expandCollapseBtn, "onmouseup", stopPropagation);	
		this.connect(this._expandCollapseBtn, "onmousedown", stopPropagation);
		this.connect(this._expandCollapseBtn, "onclick", function () {
			that.setExplored(!that._explored);
			if (that._explored) {
				dojo.publish("linksexplorer/node_expansion", [this, this.refreshOnExpand]);
				this.refreshOnExpand = false;
			} else {
				dojo.publish("linksexplorer/node_collapse", [this]);
			}
		});
		
		dojo.style(this._treeNode, {
			top: (this.yPos) + 'px',
			left: (this.xPos) + 'px'
		});
		
		if (this.moveable) {
			this.mover = new dojo.dnd.Moveable(this.domNode,{skip: true});
		}
		
		//hide the toolbar icons for DM nodes
		if(this.DMlink) {			
			dojo.addClass(this._toolbar, "hidden");
			dojo.addClass(this._addLinkBtn, "hidden");
			dojo.addClass(this._deleteArtifactBtn, "hidden");
			dojo.addClass(this._editLinkBtn, "hidden");
		}
	},
	
	_disconnectHandles: function() {
		for (var i = 0 ; i < this._connectHandles.length ; i++) {
			dojo.disconnect(this._connectHandles[i]);
		}
		this._connectHandles = [];
		
	},
	
	//check artifact permissions and linking constraints and hide create/add/delete icons accordingly
	applyToolbarPermissions: function(quickCreateArtifactType, quickCreateLinkType) {
		// summary:
		//		Apply toolbar icons based on link constraints + permissions, etc.
		//
		// params:
		// 		quickCreateArtifactType - For Quick Edit, the default Artifact type 
		//		quickCreateLinkType - For Quick Edit, the default Link type
		
		var stopPropagation =  function (e) {
			dojo.stopEvent(e);
		};
		
		this._disconnectHandles();
		var linkTypesForProjectDeferred = LinkTypeUtils.getLinkTypesForProject(this.resourceInfo.componentURI, null);
		linkTypesForProjectDeferred.addCallback(dojo.hitch(this, function(linkTypesForProject){
			this._linkTypes = {};
			dojo.forEach(linkTypesForProject, dojo.hitch(this, function(linkType) {
				this._linkTypes[linkType.getResourceURI()] = linkType;
			}));
			var processArea = this.resourceInfo.resourceInfoObject && this.resourceInfo.resourceInfoObject.processArea;
			if (!processArea) {
				processArea = this.resourceInfo.componentURI;
			}

			var createLinkAllowed = LinkTypeUtils.getAvailableConstrainedCreationLinkTypesRoles(this.resourceInfo.componentURI, processArea, [this.resourceInfo.ofType]);
			createLinkAllowed.addCallback(this, function(permissibleRoles) {
				if((permissibleRoles.length === 0) || this.DMlink) {
					dojo.addClass(this._addLinkBtn, "hidden");
				} else {
					var addAllowed = true;
					if (quickCreateArtifactType && quickCreateLinkType) {
						addAllowed = false;
						// do roles include the quick create settings as a valid option?
						for (var i=0; i<permissibleRoles.length; i++) {
							var roleTuple = permissibleRoles[i];
							for (var j=0; j<roleTuple.length; j++) {
								var role = roleTuple[j];
								if (role.getLinkType().getResourceURI() === quickCreateLinkType.linkType.getResourceURI()) {
									if ((quickCreateLinkType.objOrSub === "subject" || quickCreateLinkType.objOrSub === "bidirectional") && role.isSubjectRole()) {
										if (role.isValidTargetArtifactTypeURI(quickCreateArtifactType.typeURI)) {
											addAllowed = true;
											break;
										}
									} else if (quickCreateLinkType.objOrSub === "object" && role.isObjectRole()) {
										if (role.isValidTargetArtifactTypeURI(quickCreateArtifactType.typeURI)) {
											addAllowed = true;
											break;
										}
									}
								}
							}
							if (addAllowed) {
								break;
							}
						}
					} 
					if (addAllowed) {
						dojo.removeClass(this._addLinkBtn, "hidden");
						this._connectHandles.push(this.connect(this._addLinkBtn, "onmouseup", stopPropagation));	
						this._connectHandles.push(this.connect(this._addLinkBtn, "onmousedown", stopPropagation));
						this._connectHandles.push(this.connect(this._addLinkBtn, "onclick", function (e) {
							dojo.publish("linksexplorer/add_link", [this]);
							dojo.stopEvent(e);
							com.ibm.rdm.web.linksexplorer.LinksDialogKeys.setUnderEdit(true);
						}));
					} else {
						dojo.addClass(this._addLinkBtn, "hidden");
					}
				}
				return permissibleRoles;
			});
			
			var deleteResourceAllowed = null;
			if (this.resourceInfo.resourceInfoObject && this.resourceInfo.resourceInfoObject.processArea) {
				deleteResourceAllowed = ProcessOperationsCache.isDeleteInstanceOfArtifactTypeAllowed(this.resourceInfo.resourceURI, this.resourceInfo.componentURI, this.resourceInfo.ofType, this.resourceInfo.resourceInfoObject.processArea, false);
				deleteResourceAllowed.addCallback(this, function(isAllowed){
					if (isAllowed && !this.DMlink) {
						this._connectHandles.push(this.connect(this._deleteArtifactBtn, "onmouseup", stopPropagation));
						this._connectHandles.push(this.connect(this._deleteArtifactBtn, "onmousedown", stopPropagation));
						this._connectHandles.push(this.connect(this._deleteArtifactBtn, "onclick", function(){
							var msgForm = new com.ibm.rdm.web.common.MessageForm({
								confirmMessage: dojo.string.substitute(this._messages.confirmDeleteDialogMessage, [this.resourceInfo.name]),
								dialogTitle: this._messages.confirmDeleteDialogTitle,
								dialogStyle: "yesNo",
								onOk: dojo.hitch(this, function(){
									dojo.publish("linksexplorer/set_node_refresh", [this.parent.resourceInfo.resourceURI, this.resourceInfo.resourceURI, true]);
									ResourceUtil.deleteResource({
										resourceURI: this.resourceInfo.resourceURI,
										componentURI: this.resourceInfo.componentURI,
										disableConfirmationDialog: true
									});
									dojo.publish("linksexplorer/link_delete", [this, false]);
								})
							});
							msgForm.show();
						}));
						
						
						this._connectHandles.push(this.connect(this._editLinkBtn, "onmouseup", stopPropagation));
						this._connectHandles.push(this.connect(this._editLinkBtn, "onmousedown", stopPropagation));
						this._connectHandles.push(this.connect(this._editLinkBtn, "onclick", function(){
							if (!dojo.hasClass(this._editLinkBtn, "saveEditMenu")) {
								this.enableEditArtifactMode();
								com.ibm.rdm.web.linksexplorer.LinksDialogKeys.setUnderEdit(true);
							}
						}));
					}
					else {
						dojo.addClass(this._deleteArtifactBtn, "hidden");
						dojo.addClass(this._editLinkBtn, "hidden");
					}
					return isAllowed;
				});
			}
			// Everything is hidden or hiddenLink... hide the toolbar
			var deferredList = new dojo.DeferredList([deleteResourceAllowed, createLinkAllowed]);
			deferredList.then(dojo.hitch(this, function() {
				if((dojo.hasClass(this._deleteArtifactBtn, "hidden") &&  dojo.hasClass(this._editLinkBtn, "hidden") && dojo.hasClass(this._addLinkBtn, "hidden")) ||
						this.externalLink) {
					dojo.addClass(this._toolbar, "hidden");
				}	
			}));
			
		}));
		if(this.linkType){
			//  Can not expand RM2RM links
			this.isRM2RM = LinkTypeUtils.isRMtoRMLinkType(this.linkType.getSameAs());
			if (this.isRM2RM) {
				dojo.addClass(this._expandCollapseBtn, "hidden");
				dojo.addClass(this._toolbar, "hidden");
			}
			//  Can not modify CLM nodes
			this.isCLMLink = LinkTypeUtils.isCalmLink(this.linkType.getSameAs());
			if (this.isCLMLink) {
				dojo.addClass(this._editLinkBtn, "hidden");
				dojo.addClass(this._deleteArtifactBtn, "hidden");
				dojo.addClass(this._toolbar, "hidden");	
			}
			
		}
		
		this.toolbarPermissionsApplied = true;
	},
	
	enableEditArtifactMode: function() {
		
		if(this.locked) {
			var msgForm = new com.ibm.rdm.web.common.MessageForm({
				confirmMessage : this._messages.lockedMessage,
				dialogTitle : this._messages.error,
				dialogStyle : com.ibm.rdm.web.common.MessageForm.DIALOG_TYPES.Ok
			});
			msgForm.show();
		}
		// the span that has the name
		this.linkTitleNode = dojo.query('.linked-resource-result', this.domNode)[0];
		
		if (!this.editLinkMode && !this.compactView && !this.locked && !this.DMlink) {
			this.editLinkMode = true;
			
			//lock all other instances of this node from being edited
			if(!this.quickCreateTempNode && !this.externalLink) {
				dojo.publish("linksexplorer/lock_nodes", [this.resource.identifier, true]);
			}
			
			var titleAttrDef = RRMTypes.getWellKnownAttributeURI(this.resourceInfo.componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Title);
			
			var widgetReady = dojo.hitch(this, function(editorWidgetDomNode){
				this.inputTag = editorWidgetDomNode;
				
				dojo.attr(this.inputTag, {value: this.resource.name});
				dojo.addClass(this.inputTag, "linkEditTextBox");
				// Save a copy of the original node incase the edit is canceled. 
				this.originalLinkInfo = dojo.clone(this.linkTitleNode.firstChild);
				this.linkTitleNode.replaceChild(this.inputTag, this.linkTitleNode.firstChild);
				this.inputTag.focus();
				this.inputTag.select();
			});
			
			var resourceRevision = {
				getRevision: function() {}
			};
			
			EditManager.getEditor({
				componentURI: this.resourceInfo.componentURI,
				artifactURI: this.resourceInfo.resourceURI,
				definitionURI: titleAttrDef,
				callback: widgetReady,
				resource: resourceRevision,
				disableOnBlur: true
			});
			
			EditManager.startEdit(this.resourceInfo.componentURI, this.resourceInfo.resourceURI, titleAttrDef);
			
			//change the 'edit name' icon to a save menu
			dojo.removeClass(this._editLinkBtn,'editLinkButton');
			dojo.addClass(this._editLinkBtn, 'saveEditMenu');
			dojo.attr(this._editLinkBtn, {title: this._messages.saveArtifact});
			
			this.menu = new com.ibm.rdm.web.common.Menu({
				leftClickToOpen: true
       		});
			
			this.menu.bindDomNode(this._editLinkBtn);
			var menuItem1 = new dijit.MenuItem({
           		label: this._messages.saveArtifact,
            	iconClass: "saveEditLinkButton",
            	onClick: dojo.hitch(this, function() {
					this.saveEdits();
					com.ibm.rdm.web.linksexplorer.LinksDialogKeys.setUnderEdit(false);
				})
        	});
			dojo.addClass(menuItem1, "saveEditLinkButton");
        	this.menu.addChild(menuItem1);
			
        	var menuItem2 = new dijit.MenuItem({
         		label: this._messages.cancelLink,
            	iconClass: "cancelEditLinkButton",
            	onClick: dojo.hitch(this, function() {
            		this.disableEditMode();
            		com.ibm.rdm.web.linksexplorer.LinksDialogKeys.setUnderEdit(false);
            		if(this.quickCreateTempNode) {
						//the parent node must now refresh its list of children before being expanded again
            			//this.parent.refreshOnExpand = true;
            			dojo.publish("linksexplorer/set_node_refresh", [this.parent.resourceInfo.resourceURI, this.resourceInfo.resourceURI, true]);
            			
            			//delete artifact
            			ResourceUtil.deleteResource({
							resourceURI: this.resourceInfo.resourceURI,
							componentURI: this.resourceInfo.componentURI,
							wrappedResourceURI: this.resourceInfo.resourceInfoObject.wrappedResourceURI,
							disableConfirmationDialog: true
						});
						
						//delete the link
						dojo.publish("linksexplorer/link_delete", [this, false]);
					}
					else {
						dojo.publish("linksexplorer/lock_nodes", [this.resource.identifier, false]);
					}
            	})
        	});
        	this.menu.addChild(menuItem2);
			
			if (!this.quickCreateTempNode) {
				var menuItem3 = new dijit.MenuItem({
					label: "Save all current changes",
					iconClass: "saveEditLinkButton",
					onClick: dojo.hitch(this, function(){
						com.ibm.rdm.web.linksexplorer.LinksDialogKeys.setUnderEdit(false);
						EditManager.saveAll(this.resourceInfo.componentURI);
						dojo.publish("linksexplorer/save_all_edited_links", []);
						dojo.publish("linksexplorer/lock_nodes", [this.resource.identifier, false]);
					})
				});
				this.menu.addChild(menuItem3);
			}
		}
	},
	
	saveEdits: function() {
		var deferred = new dojo.Deferred();
		deferred = EditManager.saveArtifacts(this.resourceInfo.componentURI, [this.resourceInfo.resourceURI]);
		
		if (this.inputTag.value.length > this.NODE_LABEL_TRUNCATE_LENGTH) {
			this.inputTag.value = this.inputTag.value.substring(0, this.NODE_LABEL_TRUNCATE_LENGTH - 3) + "...";
		}
		
		dojo.publish("linksexplorer/update_edited_links", [this.resource.identifier, this.inputTag.value, this.resourceInfo.resourceURI]);
		if(!this.quickCreateTempNode) {
			dojo.publish("linksexplorer/lock_nodes", [this.resource.identifier, false]);
		}
		else {
			dojo.publish("linksexplorer/set_node_refresh", [this.parent.resourceInfo.resourceURI, this.resourceInfo.resourceURI, true]);
		}
		this.quickCreateTempNode = false;
		return deferred;
	},
	
	_positionExpander: function () {
		/*var myHeight = this.domNode.clientHeight,
			btnHeight = this._expandCollapseBtn.clientHeight;
		
		//myHeight = Math.ceil(myHeight);
		//btnHeight = Math.ceil(btnHeight);
		//console.debug("Middle is "+(myHeight / 2)+" - "+(btnHeight / 2)+" to give "+((myHeight / 2) - (btnHeight / 2)));
		dojo.style(this._expandCollapseBtn, "top", ((Math.floor(myHeight / 2) - Math.floor(btnHeight / 2))) +"px");
		*/
	},
	
	postMixInProperties: function () {
		this._messages = dojo.i18n.getLocalization("com.ibm.rdm.web.linksexplorer", "LinksExplorer");
	},
	
	uninitialize: function(){		
	
		
		if (this.link) {
			if (dojo.isArray(this.link.line)) {
				dojo.forEach(this.link.line, function(line){
					line.removeShape();
				});
			}
			else 
				if (this.link.line) {
					this.link.line.removeShape();
					this.link.line = null;
				}
			
			
			if (this.link.label) {
				this.link.label.destroy();
				this.link.label = null;
			}
		}
		
		// There is a problem with link item widgets destroy method; it destroys
		// all instances of the widget when we want to destroy just one of them.
		if (this._linkItemWidget) {
			this._linkItemWidget.destroyRecursive();
		}
		
	},
	
	setExplored: function(explored) {
		/* summary: 
		 * 		Set the visuals on this node to explored. Does NOT change the 
		 * 		core data to explored, only changes visuals
		 */
		dojo.attr(this._expandCollapseBtn, "title", explored ? this._messages.collapseNode : this._messages.expandNode);
		//if there is a change in explored state
		if (this._explored !== explored) {
			if (explored) {
				this._explored = true;
				if (dojo.hasClass(this.domNode, 'unexplored')) {
					dojo.removeClass(this.domNode, 'unexplored');
				}
				
			}
			else {
				this._explored = false;
				dojo.addClass(this.domNode, 'unexplored');
			}
		}
	},
	
	isSuspect: function() {
		/*
		 * summary:
		 * 		Determine whether this link is suspect or not. A suspect
		 * 		link is one in which either end of the link has been
		 * 		modified since this link was last modified.
		 */		
		if (this.checkSuspect) {
			var linkDate = new Date(this.modified);
			var resourceDate = new Date(this.resourceInfo.modified);
			
			if (resourceDate > linkDate) {
				return true;
			}
			
			if (this.parent) {
				var parentDate = new Date(this.parent.resourceInfo.modified);
				
				if (parentDate > linkDate) {
					return true;
				}
			}
		}
		
		return false;
	},
	
	clearSuspect: function() {
		/*
		 * summary:
		 *		update the link so that it will no longer be marked suspect
		 */
		
		var deferred = new dojo.Deferred();
		
		var getDeferred = RDMClient.doGet({
			uri: this.coreData.linkURI[0],
			componentURI: this.resourceInfo.componentURI,
			skipCache: true // Need to be sure we're working with the latest data
		});
		
		getDeferred.addCallback(this, function(response) {
			var putDeferred = RDMClient.doPut({
				uri: this.coreData.linkURI[0],
				componentURI: this.resourceInfo.componentURI,
				postData: response.doc
			});
			
			putDeferred.addCallback(this, function(response) {
				this.modified = response.ioArgs.xhr.getResponseHeader("Last-Modified");
				deferred.callback();
			});
		});

		return deferred;
	},
	
	setLayout: function (layout) {
		this.layout = layout;
		
		if(layout.name === TreeLayout.Type['RADIAL']) { //disable edit mode for radial layout
			dojo.addClass(this._editLinkBtn,'hidden');
		}
		else { //enable edit mode for all other layouts
			dojo.removeClass(this._editLinkBtn,'hidden');
		}
	},
	
	reinitTreeNode: function() {
		if(!this.externalLink) {
			this.locked = false;
			this._linkItemWidget = //new com.ibm.rdm.web.links.LinkItemWidget({
				new com.ibm.rdm.web.linksexplorer.NodeLinkItemWidget({
					"resource": this.resource,
					"linkType": this.linkType,
					"label": this.resourceInfo.name,
					"resourceViewer": this._viewer,
					"elementDescription": "",
					"displayResourceIdOnly":this.isRadialLayout(),
					"invokedByLinksExplorer": true,
					_handleLinkClicked: this._handleLinkClicked
				});			
		}else { //external link
			this.locked = false;
			this._linkItemWidget = new com.ibm.rdm.web.links.LinkItemWidget({
				"uri": this.resourceInfo.resourceURI,
				"linkType": this.linkType,
				"label": this.resourceInfo.name,
				"invokedByLinksExplorer": true
			});
			this.externalLink = true;
			dojo.addClass(this._addLinkBtn,'hidden');
		}

		
		var table = dojo.create("table");
		var tbody = dojo.create("tbody");
		table.appendChild(tbody);
		this._linkItemWidget.startup();
		tbody.appendChild(this._linkItemWidget.domNode);
		this._containerNode.replaceChild(table, this.oldTable);
		this.innerLink = this._containerNode.getElementsByTagName('a')[0];
		this.oldTable = table;
		
		//change the toolbar icon back to the edit icon
		dojo.removeClass(this._editLinkBtn,'saveEditMenu');
		dojo.addClass(this._editLinkBtn, 'editLinkButton');
		dojo.attr(this._editLinkBtn, {title: this._messages.editArtifact});
		this.editLinkMode = false;
		if (this.menu) {
			this.menu.uninitialize();
		}
		
		
	},
	
	disableEditMode: function() {
		dojo.removeClass(this._editLinkBtn,'saveEditMenu');
		dojo.addClass(this._editLinkBtn, 'editLinkButton');
		dojo.attr(this._editLinkBtn, {title: this._messages.editArtifact});
		this.menu.uninitialize();
		
		this.editLinkMode = false;
		// Cancel the edit, put back original resource row
		this.linkTitleNode.replaceChild(this.originalLinkInfo, this.inputTag);
		setTimeout(dojo.hitch(this, function () {
			EditManager.cancelArtifacts(this.resourceInfo.componentURI, [this.resourceInfo.resourceURI], true);
		}), 0);
	},
	
	switchCompactMode: function() {
		//adjust the length of the display text in the node, as well as the icon size.
		this.compactView = !this.compactView;
		
		//this._linkItemWidget.setCompactMode(this.compactView);
		
		if(this.compactView) {
			//disable editing mode from within compact mode
			dojo.addClass(this._editLinkBtn, 'hidden');
			if(this.editLinkMode) {
				this.disableEditMode();
			}
			this.locked = false;
			this.uncompactInnerLink = this.innerLink.innerHTML;
			dojo.addClass(this.domNode, "compact");
			if(this.innerLink.innerHTML.length > this.COMPACT_NODE_LABEL_TRUNCATE_LENGTH) {
				this._containerNode.getElementsByTagName('a')[0].innerHTML = this.innerLink.innerHTML.substring(0,this.COMPACT_NODE_LABEL_TRUNCATE_LENGTH-3)+"...";
			}
			else {
				this._containerNode.getElementsByTagName('a')[0].innerHTML = this.innerLink.innerHTML;
			}
			dojo.removeClass(this.domNode.getElementsByTagName('div')[0],"innerStroke");
		}
		else {
			dojo.removeClass(this.domNode, "compact");
			this._containerNode.getElementsByTagName('a')[0].innerHTML = this.uncompactInnerLink;
			dojo.addClass(this.domNode.getElementsByTagName('div')[0],"innerStroke");
			if(!this.externalLink) {
				dojo.removeClass(this._editLinkBtn, "hidden");
			}
		}
	},
	
	createInteriorWidget: function(viewer, count) {
		
		var that = this;
		this._viewer = viewer;
		if (!this._viewer.resourceURI) {
			this._viewer.resourceURI = this.resourceInfo.resourceURI;
		}
		
		//create link item widget and add it to the container's _containerNode	

		if (this.resourceInfo) { //if this is a resource link and not an external link
			
			if (count === 0) {
				dojo.publish("linksexplorer/update_title", [this.resourceInfo.identifier]);
			}
			
			if (this.resourceInfo.identifier) {
				
				//truncate artifact label if it is longer than threshold char length
				var truncatedName = this.resourceInfo.name;
				
				if (truncatedName.length > this.NODE_LABEL_TRUNCATE_LENGTH) {
					truncatedName = truncatedName.substring(0, this.NODE_LABEL_TRUNCATE_LENGTH - 3);
					truncatedName += "...";
				}
				
				//(truncatedName);
				this.resource = {
					name: truncatedName,
					identifier: this.resourceInfo.identifier,
					contentType: this.resourceInfo.contentType,
					resourceURI: this.resourceInfo.resourceURI,
					componentURI: this.resourceInfo.componentURI,
					usedInModuleArgs: this.resourceInfo.usedInModuleArgs,
					baselineArgs: this.resourceInfo.baselineArgs,
					ofType: this.resourceInfo.ofType
				};
				
				
				var currentGlobalConfig = VVCUtils.getGlobalConfigurationUri();
				var currentConfiguration = VVCUtils.getCurrentConfigurationUri();

				if (currentGlobalConfig) {
					this.resource.resourceURI += "?" + Headers.VVC_OSLC_CONFIGURATION+ "=" + encodeURIComponent(currentGlobalConfig);
				} else {
					this.resource.resourceURI += "?" + Headers.VVC_CONFIGURATION + "=" + encodeURIComponent(currentConfiguration);
				}
				
				this._linkItemWidget = //new com.ibm.rdm.web.links.LinkItemWidget({
					new com.ibm.rdm.web.linksexplorer.NodeLinkItemWidget({
					"resource": this.resource,
					"linkType": this.linkType,
					"label": this.resourceInfo.name,
					"resourceViewer": viewer,
					"elementDescription": "",
					"displayResourceIdOnly":this.isRadialLayout(),
					"invokedByLinksExplorer": true,
					_handleLinkClicked: this._handleLinkClicked
				});
				
			} else { //external link
				var internalURI = ResourceLinksHandler.isInternalResourceLink( this.resourceInfo.resourceURI );
				var notFoundInCC = null;
				if ( internalURI == true ) {
					notFoundInCC = true;
				}
				this._linkItemWidget = new com.ibm.rdm.web.links.LinkItemWidget({
					"uri": this.resourceInfo.resourceURI,
					"linkType": this.linkType,
					"label": this.resourceInfo.name,
					"invokedByLinksExplorer": true,
					"isNotFoundCC": notFoundInCC
				});
				this.externalLink = true;
				dojo.addClass(this._addLinkBtn,'hidden');
			}
			
			this.oldTable = dojo.create("table");
			var tbody = dojo.create("tbody");
			this.oldTable.appendChild(tbody);
			this._linkItemWidget.startup();
			tbody.appendChild(this._linkItemWidget.domNode);
			this._containerNode.appendChild(this.oldTable);
			this.innerLink = this._containerNode.getElementsByTagName('a')[0];
			
		}
		else if(this.DMlink) {
			//this.resourceInfo will be null for DM links
			//create a very basic interior
		}
		else {
			this._containerNode.innerHTML = this._messages.errorLoadingNode;
		}
		
		this._positionExpander();
		
	},
			
	createLinkLabel: function () {
		var linkType = this.linkType,
			processArea;
		
		if (linkType) {
			
			// the direction label is either the subject or object label of the link
			// depending on which the link points to
			var dirLabel = this.direction === "subject" ? linkType.subjectToObjectLabel : linkType.objectToSubjectLabel;
			
			// if labels have not been created for the graphic node yet
			if (!this.link.label) {
				this.linkObject = linkType.linkObject;
				if (!this.DMLink) {
					processArea = this.resourceInfo.resourceInfoObject && this.resourceInfo.resourceInfoObject.processArea;
				}

				this.link.label = new com.ibm.rdm.web.linksexplorer.LinkLabels({
					container: this.domNode.parentNode,
					primaryLabel: linkType.name,
					dirLabel: dirLabel,
					object: this,
					viewer: this._viewer,
					processArea: processArea,
					componentURI: this.resourceInfo.componentURI,
					DMlink: this.DMlink,
					RM2RM: this.isRM2RM,
					isCLMLink: this.isCLMLink
				});
			}
			else {
				this.link.label.updateLinkLabel(dirLabel);
			}
			
		}
		
	},
	
	cachePosition: function () {
		// Summary: 
		//  Remember the position of this node
		//  so that we can check it actually was moved
		//  during updatePosition.
		this._tmpY = this.yPos;
		this._tmpX = this.xPos;
	},
	
	updatePosition: function() {
		// Summary:
		//  Position the node in the DOM using it's
		//  xPos and yPos.
		if (this._tmpX !== this.xPos) {
			dojo.style(this.domNode, "left", this.xPos + "px");
		}
		if (this._tmpY !== this.yPos) {
			dojo.style(this.domNode, "top", this.yPos + "px");
		}
	},
	
	positionChanged: function () {
		// Summary:
		//  Did the position actually change during layout ?
		return ((this._tmpX !== this.xPos) || (this._tmpY !== this.yPos));
	},
	
	highlight: function (on) {
		if (on) {
			dojo.addClass(this.domNode, "highlight");
			dojo.addClass(this._expandCollapseBtn, "hover");
		} else {
			dojo.removeClass(this.domNode, "highlight");
			dojo.removeClass(this._expandCollapseBtn, "hover");
		}
		
		
	},
	
	removeChild: function(resourceURI) {
		for(var i=0; i<this.children.length; i++) {
			if(this.children[i].resourceInfo.resourceURI === resourceURI) {
				this.children.splice(i, 1);
			}
		}
	},
	
	isRadialLayout: function(){		
		//Display only the artifact ID when using RadialTreeLayout.
		return (this.layout && (this.layout.name === TreeLayout.Type.RADIAL))?true:false;	

	},
	
	_handleLinkClicked: function(event) {
		// 95619: Previously, we needed some special code to handle opening these links in snapshots.
		//	That should no longer be needed, becuase the link should be opened appropriately given its
		//	configuration URI. Simply check to see if we are trying to open the link in the current window,
		//	and then close the dialog if we are. Let LinkItemWidget handle opening the link.
		if (!event.ctrlKey && !event.shiftKey && !event.altKey) {
			dojo.publish('linksexplorer/dialog_close');
		}
	}
	
});

})(); 