/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2007, 2016.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
(function() {
dojo.provide("com.ibm.rdm.web.commenting.ReviewCommentForm");

dojo.require("com.ibm.rdm.web.review.reviewTypes");
//CIRC: dojo.require("com.ibm.rdm.web.editor.CKEditorFactory");
dojo.require("jazz.ui.MessageBox");
dojo.require("dojo.i18n");
dojo.require("dijit.form.Form");
dojo.require("dijit.form.Button");
dojo.require("dijit.form.SimpleTextarea");
dojo.require("com.ibm.rdm.web.editor.editorutils");
dojo.require("com.ibm.rdm.web.editor.linkutils");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.Constants");

dojo.requireLocalization("com.ibm.rdm.web.commenting", "ReviewCommentForm");
dojo.requireLocalization("com.ibm.rdm.web.common", "DialogForm");
dojo.requireLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");

(function () {

var MessageHandler = com.ibm.rdm.web.common.MessageHandler;

dojo.declare("com.ibm.rdm.web.commenting.ReviewCommentForm", com.ibm.rdm.web.common.DialogForm, {
	formData: null,
	templatePath: dojo.moduleUrl("com.ibm.rdm.web", "commenting/templates/ReviewCommentForm.html"),
	widgetsInTemplate: true,
	commentText: null,
	artifactStatus: null,
	isMultipleArtifacts: null,
	componentURI: null,
	ckEditor: null,

	// performCastArtifactVotesAndAddComments: Function
	//		The code that makes use of this form must provide this function.  It should
	//		provide the following signature:
	//			performCastArtifactVotesAndAddComments(onSaveComplete, onSaveError)
	performCastArtifactVotesAndAddComments: null,

	_reviewCommentForm: null,
	_commentTextArea: null,
	_commentLabelNode: null,
	_okButtonNode: null,
	_cancelButtonNode: null,
	_messages: null,
	_commonMessages: null,
	_reviewMessages: null,
	
	
	postCreate: function() {
		this.inherited("postCreate", arguments);
		this._messageBox = new jazz.ui.MessageBox(null,this._messageBoxContainer);
	},
	
	postMixInProperties: function() {
			this._messages = dojo.i18n.getLocalization("com.ibm.rdm.web.commenting", "ReviewCommentForm");
			this._commonMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.common", "DialogForm");
			this._reviewMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");
	},
	
	_getDialogTitle: function() {
		var dialogTitle = null;
		switch (this.artifactStatus) {
		case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Approved:
			if (this.isMultipleArtifacts) {
				dialogTitle = this._messages.approveArtifacts;
			} else {
				dialogTitle = this._messages.approveArtifact;
			}
			break;
		case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Disapproved:
			if (this.isMultipleArtifacts) {
				dialogTitle = this._messages.disapproveArtifacts;
			} else {
				dialogTitle = this._messages.disapproveArtifact;
			}
			break;
		case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Abstained:
			if (this.isMultipleArtifacts) {
				dialogTitle = this._messages.abstainArtifacts;
			} else {
				dialogTitle = this._messages.abstainArtifact;
			}
			break;
		case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Reviewed:
			if (this.isMultipleArtifacts) {
				dialogTitle = this._messages.reviewedArtifacts;
			} else {
				dialogTitle = this._messages.reviewedArtifact;
			}
			break;
		}
		return dialogTitle;
	},
	
	_switchButtonBarToShowingOnlyClose: function() {
		dojo.style(this._okButtonNode, "display", "none");
		dojo.style(this._cancelButtonNode, "display", "none");
		dojo.style(this._closeButtonNode, "display", "");
		this._closeButtonNode.focus();
	},

	_switchButtonBarToShowingOnlyOkAndCancel: function() {
		dojo.style(this._okButtonNode, "display", "");
		dojo.style(this._cancelButtonNode, "display", "");
		dojo.style(this._closeButtonNode, "display", "none");
	},

	_onClose: function(/*Event?*/event) {
		if (event) {
			dojo.stopEvent(event);
		}
		this.hide();
	},

	_ok: function() {
		var setEnablementForOkAndCancelButton = dojo.hitch(this, function(doSetAsEnabled){
			dojo.attr(this._okButtonNode, "disabled", !doSetAsEnabled);
			dojo.attr(this._cancelButtonNode, "disabled", !doSetAsEnabled);
		});

		this.commentText = this.ckEditor.getData();

		// Disable OK/Cancel while we're waiting for a response from the server.
		setEnablementForOkAndCancelButton(false);

		var onSaveComplete = dojo.hitch(this, function(){
			setEnablementForOkAndCancelButton(true);
			this._switchButtonBarToShowingOnlyClose();
			this._messageBox.showMessage({
				type: jazz.ui.MessageBox.STATUS_TYPES.OK,
				summary: this._reviewMessages.done
			});
			// Automatically remove the dialog after showing the completion message 
			// for a second or so.
			this.defer(
				this.hide,
				com.ibm.rdm.web.common.Constants.MILLISECONDS_TO_WAIT_BEFORE_AUTO_CLOSING
			);
		});
		var onSaveError = dojo.hitch(this, function(){
			setEnablementForOkAndCancelButton(true);
			this._switchButtonBarToShowingOnlyClose();
			// Note: we've taken pains to make sure the error is displayed in this dialog.
			// We do this so that the user has the opportunity to see the error AND THEN
			// copy the comment content to the clipboard so that he/she can "try again"
			// after clicking on the app refresh button (after waiting long enough for
			// the review to move from the Paused state back to the Started state).
			this._messageBox.showMessage(MessageHandler.getMessageBoxInfo(7302));
		});
		this._messageBox.showMessage({
			type: jazz.ui.MessageBox.STATUS_TYPES.INFO,
			summary: this._reviewMessages.savingChangesToReview
		});
		this.performCastArtifactVotesAndAddComments(
			onSaveComplete,
			onSaveError
		);
	},
	
	insertLink: function (editor) {
		com.ibm.rdm.web.editor.linkutils.createCommentLink(this, editor);
	},
	
	_onDialogOpen: function() {
		// Note: ReviewCommentForm.getInstance() reuses the same form content instance
		// whenever the containing dialog is asked to show itself and it _may_ be the
		// case that the last time it was used the OK/Cancel buttons were removed and
		// replaced with a Close button ... we account for that here (by putting these
		// buttons back).
		this._switchButtonBarToShowingOnlyOkAndCancel();

		this._messageBox.hideMessage();
		if (!this.ckEditor) {
			var types = com.ibm.rdm.web.editor.CKEditorFactory.EditorTypes;
			var onCreate = dojo.hitch(this, function(ckeditorInstance){
				this.ckEditor = ckeditorInstance;
				ckeditorInstance.richTextWidget = this;
				ckeditorInstance.on("instanceReady", function (){
					// console.log('ckeditor create? ', ckeditorInstance.window.$.document.body);
					dojo.addClass(ckeditorInstance.window.$.document.body, 'comment-editor');
					window.setTimeout(dojo.hitch(this, function(){
						this.ckEditor.focus();
					}), 160);
				}, this);
				ckeditorInstance.on("paste", function (evt) {
					var pasteHtml = com.ibm.rdm.web.editor.editorutils.handleCommonOnPaste(evt);	
					if (!pasteHtml) {
						return;
					}
					pasteHtml = com.ibm.rdm.web.editor.editorutils.handleCommentOnPaste(pasteHtml, MessageHandler, this._messageBox);
					evt.data.dataValue = pasteHtml;
					var validationError = com.ibm.rdm.web.editor.editorutils.validatePastedXHTMLFragmentReturnFixed(evt, null, false);
					if (validationError) {
						// if the paste does not pass server validation...				
						com.ibm.rdm.web.editor.editorutils.showInvalidPasteDialogWithFixed(evt, validationError.error, validationError.fixed);
					}   
				}, this);
			});
			var createArgs = {
				"node": this._commentTextArea,
				"config": types.REVIEW_COMMENT_EDITOR,
				"callback": onCreate,
				"inline": false
			};
			com.ibm.rdm.web.editor.CKEditorFactory.create(createArgs);
		}
		switch (this.artifactStatus) {
			case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Approved:
				this._okButtonNode.innerHTML = this._messages.approve;
				break;
			case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Disapproved:
				this._okButtonNode.innerHTML = this._messages.disapprove;
				break;
			case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Abstained:
				this._okButtonNode.innerHTML = this._messages.abstain;
				break;
			case com.ibm.rdm.web.review.reviewTypes.ARTIFACT_STATUS_TYPES.Reviewed:
				this._okButtonNode.innerHTML = this._messages.reviewed;
				break;
		}
		var commentLabel = null;
		if (this.isMultipleArtifacts) {
			commentLabel = this._messages.commentLabelMultiple;
		} else {
			commentLabel = this._messages.commentLabel;
		}
		if (dojo.isIE) {
			this._commentLabelNode.innerText = commentLabel;
		} else {
			this._commentLabelNode.textContent = commentLabel;
		}
	},
	
	show: function() {
		dojo.style(this.domNode, "display","block");
		this._dialog = new jazz.ui.Dialog({
			contentNode: this._reviewCommentForm,
			primaryTitle: this._getDialogTitle(),
			destroyContents: false,
			width: "auto",   
			onClose: dojo.hitch(this, this._onDialogClose),
			onOpen: dojo.hitch(this, this._onDialogOpen)
		});
	},
	
	_onDialogClose: function() {
		dojo.style(this._dialog.domNode, "display","none");
		this.unregisterNavigateAwayHandler();
		if (this.onDialogClose) {
			this.onDialogClose();
		}
		if (this.ckEditor)  {
			// CKE4 Adoption
			// CKE4 Seems to lose necessary data once it is lost from the dom, because of this we can not
			// reuse the instance and we have to destroy it.
			this.ckEditor.destroy(true);
			this.ckEditor = null;
		}
	}
});
	
com.ibm.rdm.web.commenting.ReviewCommentForm.getInstance = function (args) {
	var ReviewCommentForm = com.ibm.rdm.web.commenting.ReviewCommentForm;
	if (!ReviewCommentForm.instance) {
		ReviewCommentForm.instance = new ReviewCommentForm(args);
	} else {
		ReviewCommentForm.instance.registerNavigateAwayHandler();
		ReviewCommentForm.instance.artifactStatus = args.artifactStatus;
		ReviewCommentForm.instance.componentURI = args.componentURI;
		ReviewCommentForm.instance.isMultipleArtifacts = args.isMultipleArtifacts;
		ReviewCommentForm.instance.formData = args.formData;
		ReviewCommentForm.instance.onDialogClose = args.onDialogClose;
		ReviewCommentForm.instance.onOk = args.onOk;
		ReviewCommentForm.instance.onCancel = args.onCancel;
		ReviewCommentForm.instance.performCastArtifactVotesAndAddComments =
			args.performCastArtifactVotesAndAddComments;
		
		// make sure the previous state of the form is forgotten
		ReviewCommentForm.instance._okHandled = false;
		ReviewCommentForm.instance._cancelHandled = false;
	}
	
	return ReviewCommentForm.instance;
};
})();
})();