/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2007, 2016.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
(function() {
dojo.provide("com.ibm.rdm.web.commenting.utils");

dojo.require("dojo.string");
dojo.require("com.ibm.rdm.web.commenting.EditCommentForm");
dojo.require("jazz.util.date");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.UriHelper");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.common.HoverPopup");
dojo.require("com.ibm.rdm.web.common.ResourceLinksHandler");
dojo.require("com.ibm.rdm.web.mail.SendEmailForm");
dojo.require("com.ibm.rdm.web.mail.mailutils");
dojo.require("com.ibm.rdm.web.viewer.viewerutils");
dojo.require("com.ibm.rdm.web.util.queryutils");
dojo.require("com.ibm.rdm.web.common.UserHover");
dojo.require("com.ibm.rdm.web.server.client.utils.UserCache");
dojo.require("com.ibm.rdm.web.review.ReviewHoverPopup");
dojo.require("com.ibm.rdm.web.common.ProjectCache");
dojo.require("com.ibm.rdm.web.util.bidiutils"); //bidi
dojo.require("com.ibm.rdm.web.common.ArtifactIcons");
dojo.require("com.ibm.rdm.web.util.vvcutils");
dojo.require("com.ibm.rdm.web.commenting.CommentingClient");
dojo.require("com.ibm.rdm.web.notifications.EventTopics");

dojo.requireLocalization("com.ibm.rdm.web.commenting", "commenting");
dojo.requireLocalization("com.ibm.rdm.web.common", "common");

// ---------- COMMENTING CONSTANTS --------
com.ibm.rdm.web.commenting.LOCATION_PANE_ID_PREFIX = "rdm-ui-annotation-annotationpane-location-";
com.ibm.rdm.web.commenting.XMLNS_COMMENTING = "http://com.ibm.rdm.comments";
com.ibm.rdm.web.commenting.XMLNS_RECENT_COMMENTING = "http://com.ibm.rdm/recentevents/comments";
com.ibm.rdm.web.commenting.COMMENT_SLUG = "comment";
com.ibm.rdm.web.commenting.QUERY_URI = "projects/rdm/query";
com.ibm.rdm.web.commenting.MODIFIED_PROPERTY = "Last-Modified";
com.ibm.rdm.web.commenting.MODIFIED_BY_PROPERTY = "X-Last-Modified-User";
com.ibm.rdm.web.commenting.LOCATION_PROPERTY = "Location";
com.ibm.rdm.web.commenting.CONTENT_LOCATION_PROPERTY = "Content-Location";
com.ibm.rdm.web.commenting.DIRECTED_TO_TAG = "directedTo";
com.ibm.rdm.web.commenting.TEXT_TAG = "text";

// Group key for non-review comments
com.ibm.rdm.web.commenting.NON_REVIEW_COMMENTS = -1;

// Values of priority fields
// NOTE - these values match the stored values for priority
com.ibm.rdm.web.commenting.PRIORITY_LOW = "Low";
com.ibm.rdm.web.commenting.PRIORITY_MEDIUM = "Medium";
com.ibm.rdm.web.commenting.PRIORITY_HIGH = "High";
com.ibm.rdm.web.commenting.PRIORITY_DEFAULT = com.ibm.rdm.web.commenting.PRIORITY_MEDIUM;
com.ibm.rdm.web.commenting.PRIORITY_UNSPECIFIED = "Unspecified";

// Date ranges
com.ibm.rdm.web.commenting.UPDATED_UNSPECIFIED = 0;
com.ibm.rdm.web.commenting.UPDATED_TODAY = 1;
com.ibm.rdm.web.commenting.UPDATED_YESTERDAY = 2;
com.ibm.rdm.web.commenting.UPDATED_PAST_WEEK = 3;
com.ibm.rdm.web.commenting.UPDATED_PAST_MONTH = 4;
com.ibm.rdm.web.commenting.UPDATED_EARLIER = 5;

// if directedTo users for a comment exceeds this value, abbreviate the directedTo names
com.ibm.rdm.web.commenting.ABBREVIATE_DIRECTEDTO_SIZE = 3;
com.ibm.rdm.web.commenting.MAX_DIRECTEDTO_SIZE = 7;

var messages = dojo.i18n.getLocalization("com.ibm.rdm.web.commenting", "commenting");
var commonMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.common", "common");
var UserCache = com.ibm.rdm.web.server.client.utils.UserCache;
var ResourceLinksHandler = com.ibm.rdm.web.common.ResourceLinksHandler;
var MailUtils = com.ibm.rdm.web.mail.mailutils;
var DomUtils = com.ibm.rdm.web.util.domutils;
var ReviewHoverPopup = com.ibm.rdm.web.review.ReviewHoverPopup;
var COMMENT_TRUNCATION_LENGTH = 250;
var AUTHOR_IMAGE_SIZE = 25;
var BidiUtils = com.ibm.rdm.web.util.bidiutils;  //bidi
var ArtifactIcons = com.ibm.rdm.web.common.ArtifactIcons;
var VvcUtils = com.ibm.rdm.web.util.vvcutils;

// ---------- COMMENTING UTILS ----------
com.ibm.rdm.web.commenting.utils = {
	_UPDATED_OPTIONS: null,
	_PRIORITY_OPTIONS_FORM: null,
	_REQ_UPDATED_OPTIONS: null,
	_PRIORITY_OPTIONS_FILTER: null,
	
	_blankGif: dojo.moduleUrl("dojo", "resources/blank.gif").toString(),
	
	getFilterPriorityOptions: function() {
		if (this._PRIORITY_OPTIONS_FILTER == null) {
			this._PRIORITY_OPTIONS_FILTER = {};
			this._PRIORITY_OPTIONS_FILTER[com.ibm.rdm.web.commenting.PRIORITY_UNSPECIFIED]= {text: messages.all, iconUri: null, optionClass: "filter-form-defaults"};
			this._PRIORITY_OPTIONS_FILTER[com.ibm.rdm.web.commenting.PRIORITY_LOW] = {text: messages.low, iconUri: null};
			this._PRIORITY_OPTIONS_FILTER[com.ibm.rdm.web.commenting.PRIORITY_MEDIUM] = {text: messages.medium, iconUri: null};
			this._PRIORITY_OPTIONS_FILTER[com.ibm.rdm.web.commenting.PRIORITY_HIGH] =  {text: messages.high, iconUri: null};
		}
		return this._PRIORITY_OPTIONS_FILTER;
	},
	
	getFormPriorityOptions: function() {
		if (this._PRIORITY_OPTIONS_FORM == null) {
			this._PRIORITY_OPTIONS_FORM = {};
			this._PRIORITY_OPTIONS_FORM[com.ibm.rdm.web.commenting.PRIORITY_LOW] = {text: messages.low, iconUri: null};
			this._PRIORITY_OPTIONS_FORM[com.ibm.rdm.web.commenting.PRIORITY_MEDIUM] = {text: messages.medium, iconUri: null};
			this._PRIORITY_OPTIONS_FORM[com.ibm.rdm.web.commenting.PRIORITY_HIGH] =  {text: messages.high, iconUri: null};
		}
		return this._PRIORITY_OPTIONS_FORM;
	},
	
	getUpdatedOptions: function() {
		if (this._UPDATED_OPTIONS == null) {
			this._UPDATED_OPTIONS = {};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_UNSPECIFIED]= {text: messages.all, iconUri: null};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_TODAY]= {text: messages.today, iconUri: null};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_YESTERDAY]= {text: messages.yesterday, iconUri: null};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_PAST_WEEK]= {text: messages.pastWeek, iconUri: null};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_PAST_MONTH]= {text: messages.pastMonth, iconUri: null};
			this._UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_EARLIER]= {text: messages.earlier, iconUri: null};
		}
		return this._UPDATED_OPTIONS;
	},
	
	getReqUpdatedOptions: function() {
		if (this._REQ_UPDATED_OPTIONS == null) {
			this._REQ_UPDATED_OPTIONS = {};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_UNSPECIFIED]= {text: messages.all, iconUri: null};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_TODAY]= {text: messages.today, iconUri: null};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_YESTERDAY]= {text: messages.yesterday, iconUri: null};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_PAST_WEEK]= {text: messages.pastWeek, iconUri: null};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_PAST_MONTH]= {text: messages.pastMonth, iconUri: null};
			this._REQ_UPDATED_OPTIONS[com.ibm.rdm.web.commenting.UPDATED_EARLIER]= {text: messages.earlier, iconUri: null};
		}
		return this._REQ_UPDATED_OPTIONS;
	},
	
	getUpdatedRangeTitle: function(type) {
		if (type == com.ibm.rdm.web.commenting.UPDATED_UNSPECIFIED) {
			return messages.all;
		}
		if (type == com.ibm.rdm.web.commenting.UPDATED_TODAY) {
			return messages.today;
		}
		if (type == com.ibm.rdm.web.commenting.UPDATED_YESTERDAY) {
			return messages.yesterday;
		}
		if (type == com.ibm.rdm.web.commenting.UPDATED_PAST_WEEK) {
			return messages.pastWeek;
		}
		if (type == com.ibm.rdm.web.commenting.UPDATED_PAST_MONTH) {
			return messages.pastMonth;
		}
		if (type == com.ibm.rdm.web.commenting.UPDATED_EARLIER) {
			return messages.earlier;
		}
		
		return messages.unknownRange;
	},
	
	getUnknownUserName: function() {
		// summary:
		//		Returns a name to use for unknown user.
		// description:
		//		Return a string to display for user references that are not found.
		return messages.unknownUser;
	},
	
	getWindowHeight: function() {
		var r;
		try {
			r = window.innerHeight;
		}
		catch (e) {
			// attempt to handle the error by returning a legitimate value
			return 100;
		}
		r = document.body.clientHeight;
		return r;
	},
	spanHtml: function(text, className) {
		if (!text) {
			return "";
		}
		else {
			return "<span class=\"" + className + "\"" + ">" + text + "</span>";
		}
	},
	getPriorityImgPathOfComment: function(comment) {
		return this.getPriorityImgPath(comment.priority, comment.isResolved);
	},
	getPriorityImgPath: function(priority, isResolved) {
		var fname = "priority-low.gif";
		if (com.ibm.rdm.web.commenting.PRIORITY_MEDIUM == priority) {
			fname = "priority-medium.gif";
		}
		else if (com.ibm.rdm.web.commenting.PRIORITY_HIGH == priority) {
			fname = "priority-high.gif";
		}
		
		if (isResolved) {
			fname = "res-" + fname;
		}
		return dojo.moduleUrl("com.ibm.rdm.web.commenting", "images/icons/obj16/" + fname);
		
	},
	fixContentPane: function(contentPane) {
		contentPane._createSubWidgets = dojo.hitch(contentPane, function(){
			return;
		});
	},
	
	createCommentText: function(widget, popupDiv, comment, popupWidget, w, h) {
		var divText = document.createElement("div");
		var wPx = w + "px;";
		var hPx = h + "px;";
		var contentPaneText = new dijit.layout.ContentPane({style: "height: " + hPx + " width: " + wPx}, divText);
		if (comment.isResolved) {
			dojo.addClass(divText, "com-ibm-rdm-resolved-comment-text");
		}
		var text = comment.text;
		divText.innerHTML = "<span>" + text + "</span>";
		this._connectCommentLinksHandler(widget, divText, popupWidget);
		popupDiv.appendChild(contentPaneText.domNode);
	},
	
	createCommentByline: function(widget, appendTo, comment, showUserHovers, /*optional*/ resourceTitle) {
		// summary:
		//		Creates the "byline" for a comment, showing details of the author, directed-to users,
		//		creation date etc.
		// widget:
		//		The widget that will contain the comment byline.
		// appendTo:
		//		A DOM node to which the byline content is appended.
		// comment:
		//		The comment to be displayed.
		// showUserHovers:
		//		Indicates whether hovers should be shown on the author and directed-to user names.
		// resourceTitle:
		//		The name of the resource relating to the comment.  Included in any email message
		//		created by clicking on a user name.  If not supplied the URI of the resource is
		//		included instead.
		var userHovers = [];
		var userToStyle = "com-ibm-rdm-comment-entry-username";
		var userFromStyle = "com-ibm-rdm-comment-entry-username";
		var reviewIconStyle = "com-ibm-rdm-comment-entry-review-icon";
		var baseIconStyle = "com-ibm-rdm-comment-entry-base-icon";
		var timeStyle = "j-date";
		var toTextStyle = "j-hint";
		
		var fromNode = document.createElement("span");
		var self = this;
		
		var handleUserInfo = function(mapUserUriToUserInfo) {
			var userInfo = mapUserUriToUserInfo[UserCache.userIdToURI(comment.author)];
			
			// The author's name with a hover.
			var mailparams = {
				subject: comment.title,
				bodyHtml: self._makeCommentEmailBody(userInfo.name, comment, resourceTitle),
				cssClassName: userFromStyle
			};
			var authorNode = MailUtils.createMailTo(userInfo, comment.componentURI, mailparams);
			fromNode.appendChild(authorNode);
			
			if (showUserHovers) {
				var authorHover = new com.ibm.rdm.web.common.UserHover({componentURI : comment.componentURI}, null);
				authorHover.createHoverPopup(authorNode, userInfo);
				userHovers.push(authorHover);
			}
			
			// This function is used by recent comments and comments, so the test for review comments is
			// different.  Do recent comments test first.
			if (comment.reviewId || (comment.isReviewComment && comment.isReviewComment())) {
				var commentReviewIcon = document.createElement("img");
				commentReviewIcon.src = this._blankGif;
				if (comment.isBaseComment) {
					commentReviewIcon.className = "base-review-comment-icon";
				} else {
					commentReviewIcon.className = "review-comment-icon";
				}
				var reviewURI = com.ibm.rdm.web.FRONTING_SERVICE_REVIEW_AND_APPROVAL + "/" + com.ibm.rdm.web.util.queryutils.getLastPathSegment(comment.reviewId);
				widget.connect(commentReviewIcon, "onclick", function(event) {
					dojo.stopEvent(event);
					ResourceLinksHandler.showInternalResource(reviewURI, {"isRemoveFromCache":false, "useWindowOpen":true});
				});
				// TODO (LPK) - tooltip like rich client shows
				//commentReviewIcon.setAttribute("title", "Review comment.  Use click to open the review.");
				// TODO (LPK) - retrieve review via deferred, put review name in alt tag
				//var review = this.commentingWidget.getReview(this.annotation.reviewId);
				//var msg = null;
				//if (review == null) {
				//	msg = dojo.string.substitute(this._messages["reviewComment"], [this._messages["reviewUnknown"]]);
				//} else {
				//	msg = dojo.string.substitute(this._messages["reviewComment"], [review.getReviewName()]);
				//}
				//commentReviewIcon.setAttribute("alt", msg);
				dojo.addClass(commentReviewIcon, reviewIconStyle);
				dojo.place(commentReviewIcon, authorNode, "before");
			} else if (comment.isBaseComment) {
				var baseCommentIcon = document.createElement("img");
				baseCommentIcon.src = dojo.moduleUrl("com.ibm.rdm.web", "project/images/shared_ovr.gif");
				dojo.addClass(baseCommentIcon, baseIconStyle);
				dojo.place(baseCommentIcon, authorNode, "before");
			}
			
			var abbreviateUserNames = comment.directedToUsers.length > com.ibm.rdm.web.commenting.ABBREVIATE_DIRECTEDTO_SIZE;
			var length = Math.min(comment.directedToUsers.length, com.ibm.rdm.web.commenting.MAX_DIRECTEDTO_SIZE);
			for (var i = 0; i<length; i++) {
				var directedTo = comment.directedToUsers[i];
				var directedToUserInfo = mapUserUriToUserInfo[UserCache.userIdToURI(directedTo)];
				// skip users who are unidentified
				var directedToUser = (directedToUserInfo ? directedToUserInfo.name : "");
				if (directedToUser != "") {
					mailparams = {
						subject: comment.title,
						bodyHtml: self._makeCommentEmailBody(userInfo.name, comment, resourceTitle),
						linkText: abbreviateUserNames ? self._abbreviatedUserName(directedToUser) : directedToUser,
						cssClassName: userToStyle
					};
					var directedToNode = MailUtils.createMailTo(directedToUserInfo, comment.componentURI, mailparams);
					
					if (showUserHovers) {
						var directedToHover = new com.ibm.rdm.web.common.UserHover({componentURI : comment.componentURI}, null);
						directedToHover.createHoverPopup(directedToNode, directedToUserInfo);
						userHovers.push(directedToHover);
					}
						
					if (i==0) {
						var toTextElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode(toTextStyle, " " + messages["to"] + " ");
						fromNode.appendChild(toTextElement);
					} else {
						var commaElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode(toTextStyle, ", ");
						fromNode.appendChild(commaElement);
					}
					fromNode.appendChild(directedToNode);
				}
			}
			
			if (comment.directedToUsers.length > com.ibm.rdm.web.commenting.MAX_DIRECTEDTO_SIZE) {
				var moreElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode(toTextStyle, "...");
				fromNode.appendChild(moreElement);
			}
			
			var timeStr =  jazz.util.date.formatDif(new Date(comment.updated), false, true);
			var timeElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode(timeStyle, " " + timeStr);
			fromNode.appendChild(timeElement);
			return mapUserUriToUserInfo;
		};
		
		var requestedUsers = [];
		
		requestedUsers.push(UserCache.userIdToURI(comment.author));
		
		for (var i = 0 ; i < comment.directedToUsers.length; i++) {
			var userId = comment.directedToUsers[i];
			requestedUsers.push(UserCache.userIdToURI(userId));
		}
		
		UserCache.getUsers(requestedUsers).addCallback(this, handleUserInfo);
			
		appendTo.appendChild(fromNode);
		return userHovers;
		
	},
	
	createAuthorImage: function(appendTo, comment) {
		
		var authorURI = UserCache.userIdToURI(comment.author);
		var authorDeferred = UserCache.getUser(authorURI, true);
		
		authorDeferred.then(dojo.hitch(this, function(userInfo){
			var photoUri = com.ibm.rdm.web.common.UriHelper.makeUserPhotoUrl(userInfo);
			var img = DomUtils.createImageFixAspect(photoUri, AUTHOR_IMAGE_SIZE, "com-ibm-rdm-comment-author-image");
			img.setAttribute("alt", commonMessages["userImage"]);
			appendTo.appendChild(img);
		}));
	},
	
	_abbreviatedUserName: function(userName) {
		if (userName.length == 0) {
			return userName;
		}
		var substrings = [];
		var pos = userName.indexOf(" ");
		var name = userName;
		if (pos != -1) {
			while (pos != -1) {
				var subName = name.substring(0, pos);
				if (subName != "") {
					substrings.push(subName);
				}
				name = name.slice(pos + 1);
				pos = name.indexOf(" ");
			}
		}
		substrings.push(name);
		var abbreviatedName = null;
		if (substrings.length == 1) {
			// take first two characters of the string
			abbreviatedName = substrings[0].substring(0,2);
		} else {
			// take first character of each substring
			abbreviatedName = "";
			for (var i=0; i<substrings.length; i++) {
				abbreviatedName = abbreviatedName + substrings[i].charAt(0);
			}
		}
		return abbreviatedName;
	},
	
	getPriorityImageSrc: function(priority, isResolved) {
		var imagePath = null;
		if (priority) {
			switch (priority) {
				case "High" :
					imagePath = (isResolved ?
						"commenting/images/icons/obj16/res-priority-high.gif" :
						"commenting/images/icons/obj16/priority-high-pd.gif"
					);
					break;
				case "Medium" :
					imagePath = (isResolved ?
						"commenting/images/icons/obj16/res-priority-medium.gif" :
						"commenting/images/icons/obj16/priority-medium.gif"
					);
					break;
				case "Low" :
					imagePath = (isResolved ?
						"commenting/images/icons/obj16/res-priority-low.gif" :
						"commenting/images/icons/obj16/priority-low.gif"
					);
					break;
			}
		}
		return dojo.moduleUrl("com.ibm.rdm.web", imagePath);
	},
	
	_connectCommentLinksHandler: function(widget, commentTextNode, popupWidget) {
		var self = this;
		var links = dojo.query("a", commentTextNode);
		for (var i = 0; i < links.length; i++) {
			var link = links[i];
			var linkHref = link.attributes.getNamedItem("href").value;
			if (ResourceLinksHandler.isLinkInternal(linkHref)) {
				var href = com.ibm.rdm.web.FRONTING_SERVICE_RESOURCES + '/' + com.ibm.rdm.web.util.queryutils.getLastPathSegment(linkHref);
				link.setAttribute("href", href);
			}
		}
	},
	
	setCommentText: function(commentTextNode, comment) {
		commentTextNode.innerHTML = comment.text;
		if (comment.isResolved) {
			dojo.addClass(commentTextNode, "com-ibm-rdm-resolved-comment-text");
		}
	},
	
	connectCommentLinksAndRichHover: function(widget, commentTextNode) {
		// summary:
		//		Hookup navigation and rich hover for links in the comment text.
		// widget:
		//		The widget that contains the commentTextNode.
		// commentTextNode:
		//		DOM element that contains the text with the links.
		// returns:
		//		An array of ResourceLinks, these links should be destroyed when the widget
		//		is destroyed.

		var self = this;
		var resourceLinks = [];
		var connection = null;
		var links = dojo.query("a", commentTextNode);
		for (var i = 0; i < links.length; i++) {

			var link = links[i];
			// Fix for 83952: Error opening business process map: "The comments for an artifact could not be loaded ID CRRRW7209E..."
			// Remove id attributes that are unnecessarily being added by server.
			var idAttr = link.getAttribute("id");
			if (idAttr) {
				dojo.removeAttr(link, "id");
			}
			var uri = dojo.attr(link, "href");
			if (!uri) {
				continue;
			}
			if(dojo.isIE < 8) {
				// In IE7, getAttribute (and hence dojo.attr) always gives an absolute value for href, even if
				// the actual value is relative.  So we have to work around this.
				// At this point the URI may be either of these forms:
				//
				//		https://server:port/rm/resource/ABCDEFGH		(when the value has been created on the client)
				//		https://server:port/resource/ABCDEFGH			(when the value has been read from the server)
				//
				// In the first case, we can leave it as it is.
				//
				// In the second case, we convert it to a link of the form ../resource/ABCDEFGH, similar to that
				// for other browsers.

				var baseUri = com.ibm.rdm.web.FRONTING_SERVICE_CONTEXT_ROOT;
				if (uri.indexOf(baseUri) == -1) {
					// does not start with "https://server:port/rm/"
					baseUri = baseUri.replace(net.jazz.ajax._contextRoot+'/', '/');
					if (uri.indexOf(baseUri) == 0) {
						// starts with "https://server:port/"
						uri = uri.replace(baseUri, "../");
					}
				}
			}
			if (uri.toString().indexOf("http") == -1) {
				
				var href = com.ibm.rdm.web.util.queryutils
					.getFullyQualifiedUriFromRelativeLink(
						com.ibm.rdm.web.FRONTING_SERVICE_COMMENTS, uri);
				dojo.attr(link, "href", href);
				// Refresh the working URI now that we've updated the anchor
				uri = href;
			}
			
			var resourceLink;
			if(uri.toString().indexOf("/reviews/") != -1) {
				// Create a review hover
				resourceLink = new ReviewHoverPopup({
					viewType: ReviewHoverPopup
						.VIEW_TYPE.OVERALL_REVIEW_SUMMARY_STATUS,
					orientation: "L" // Popup to the left of the target node.
				});
				connection = dojo.connect(link, "onclick", this, function(event){
					dojo.stopEvent(event);
					ResourceLinksHandler.showInternalResource(uri.toString());
				});
				resourceLink.createHoverPopup(link,
					com.ibm.rdm.web.util.queryutils.getLastPathSegment(
						uri.toString()));
			} else {
				// Create a standard Resource Link
				// Point to the resource in the currently loaded configuration.
				var config = VvcUtils.getCurrentConfigurationUri();
				if (config) {
					uri = VvcUtils.replaceConfigurationInUri(uri, config);
				}
				resourceLink = new jazz.ui.ResourceLink({
					uri: uri,
					retainLinkText: true,
					lazyFetch: true
				}, link);
			}
			resourceLinks.push(resourceLink);
		}
		return {resourceLinks: resourceLinks, connection: connection};
		
	},
	
	hookupReplyToLink: function(widget, replyNode, comment, orientation, commentingWidget) {
		// summary:
		//		Connects a hover popup to a node, such that hovering shows detail of
		//		a comment.
		// description:
		//		Note that the resultant hover does not include the link to goto the
		//		comment - it just displays it.
		// widget:
		//		The widget that the hover node is on.
		// replyNode:
		//		A dom node to which the hover action is to be applied.
		// comment:
		//		The comment to be displayed in the hover.
		// orientation:
		//		A string to force the hover to a particular orientation.  Valid values are
		//		"R", "r", "L", "l".  Any other value, or null lets the hover open whichever
		//		side has more space.
		var self = this;
		var createFunction = function createPopupContents(popupWidget) {
		return self._getReplyTooltipContent(widget, comment, commentingWidget, popupWidget, false);
		};
		
		return new com.ibm.rdm.web.common.HoverPopup({
			around: [replyNode],
			createContents: createFunction,
			orientation: orientation,
			destroyContents: true
		}, null);
	},
	
	_hookupReplyToLink: function(widget, replyNode, comment, commentingWidget) {
		var self = this;
		var createFunction = function createPopupContents(popupWidget) {
		return self._getReplyTooltipContent(widget, comment, commentingWidget, popupWidget, true);
		};
		
		return new com.ibm.rdm.web.common.HoverPopup({
			around: [replyNode],
			createContents: createFunction,
			destroyContents: true
		}, null);
	},
	
	_getReplyTooltipContent: function(widget, comment, commentingWidget, popupWidget, includeGoToLink){
		var self = this;
		var popup = document.createElement("span");
		var popupDiv = document.createElement("div");
		dojo.addClass(popupDiv, "com-ibm-rdm-comment-flyout");
		popup.appendChild(popupDiv);
			
		if (includeGoToLink) {
			// links section at top of tooltip
			var linksDiv = this._createPopupLinksDiv(popupDiv);
		
			// go to comment link
			var goToLink = document.createElement("A");
			goToLink.appendChild(document.createTextNode(messages.goToComment));
			widget.connect(goToLink, "onclick", function(event) {
				popupWidget.close();
				self._goToCommentClicked(event, comment, commentingWidget);});
			linksDiv.appendChild(goToLink);
		}
			
		// WI 70442 - Attempting to pass a resourceTitle to the edit comment form.
		// This is an optional param, and if it is undefined it will fall back to the resourceURI
		var resourceDisplayName;
		if (comment.displayName) {
			resourceDisplayName = comment.displayName;
		}
		else if (commentingWidget && commentingWidget.getResourceInfo()) {
			resourceDisplayName = commentingWidget.getResourceInfo().displayName;
		}
		// comment byline
		com.ibm.rdm.web.commenting.utils.createCommentByline(widget, popupDiv, comment, false, resourceDisplayName);
		dojo.addClass(popupDiv, "info");
		
		// comment title
		this._createPopupTitleDiv(comment, popupDiv);
		
		// comment text
		this._createPopupTextDiv(widget, comment, popupDiv, popupWidget);

		return popup;
	},
	
	_createPopupLinksDiv: function(popupDiv) {
		// links section at top of tooltip
		var linksDiv = document.createElement("div");
		dojo.addClass(linksDiv, "linksDiv");
		popupDiv.appendChild(linksDiv);
		return linksDiv;
	},
	
	_createPopupTextDiv: function(widget, comment, popupDiv, popupWidget) {
		var divText = document.createElement("div");
		var contentPaneText = new dijit.layout.ContentPane({style: "height: 200px; width: 350px;"}, divText);
		dojo.addClass(contentPaneText, "content");
		dojo.addClass(divText, "text");
		if (comment.isResolved) {
			dojo.addClass(divText, "resolvedText");
		}
		var text = comment.text;
		divText.innerHTML = "<span>" + text + "</span>";
		this._connectCommentLinksHandler(widget, divText, popupWidget);
		popupDiv.appendChild(contentPaneText.domNode);
	},
	
	_createPopupTitleDiv: function(comment, popupDiv) {
		var divTitle = document.createElement("div");
		if (comment.isResolved) {
			dojo.addClass(divTitle, "resolvedTitle");
		} else if (comment.priority == com.ibm.rdm.web.commenting.PRIORITY_HIGH) {
			dojo.addClass(divTitle, "title-high-priority");
		} else {
			dojo.addClass(divTitle, "title");
		}
		divTitle.appendChild(document.createTextNode(BidiUtils.setTextDirUCC(comment.title)));
		popupDiv.appendChild(divTitle);
	},
		
	_replyToCommentClicked: function(event, comment, commentingWidget) {
		if (commentingWidget) {
			commentingWidget.replyToButtonClick(comment);
		} else {
			this.replyToComment(comment);
		}
	},
	
	_goToCommentClicked: function(event, comment, commentingWidget) {
		commentingWidget._selectCommentPane(comment.linkSelfUri, true);
	},
	
	replyToComment: function(comment) {
		var title = this.makeReplyTitle(comment.title);
		var formData = {
			formTitle: messages.replyToComment,
			resourceLocationId: comment.locationId,
			commentTitle: title,
			directedToUsers: this.makeCombinedDirectedToList(comment),
			comment: comment
		};
		this.openEditCommentForm(formData, comment.componentURI);
	},
	
	createArtifactInReviewNode: function(widget, appendTo, artifactName, artifactType, artifactRevision, comment, createRichText, popupWidget) {
		// artifact image
		var artifactImg = this._createArtifactImage(artifactType, comment.componentURI);
		appendTo.appendChild(artifactImg);
			
		var resourceHyperlink = null;
		// artifact name
		if (artifactType == null) {
			var node = com.ibm.rdm.web.util.domutils.createSpannedTextNode("com-ibm-rdm-comment-artifact-not-exist", messages.artifactNotExist);
			appendTo.appendChild(node);
		} else {
			var href = com.ibm.rdm.web.common.UriHelper.getResourceInReviewContextJafLink(
				comment.resourceUri,
				artifactRevision,
				comment.reviewId
			);
			var linkProperties = ResourceLinksHandler.createCommentResourceLink(href, artifactName, true, true, comment.resourceUri);
			resourceHyperlink = linkProperties.resourceLinkNode;
			widget.connect(resourceHyperlink, "onclick", function(event) {
				if (popupWidget) {
					popupWidget.close();
				}
				dojo.stopEvent(event);
				ResourceLinksHandler.showArtifactInReview(href, comment.linkSelfUri);
			});
			appendTo.appendChild(resourceHyperlink);
			dojo.addClass(resourceHyperlink, "com-ibm-rdm-comment-artifact-hyperlink");
		}
		return resourceHyperlink;
	},
	
	createArtifactNode: function(
			widget, appendTo, artifactName, artifactType, comment, popupWidget, viewlet, configurationUri) {
		var self = this;
		// artifact image
		var artifactImg = this._createArtifactImage(artifactType, comment.componentURI, comment.resourceUri);
		appendTo.appendChild(artifactImg);
		
		var artifactFullLabel = BidiUtils.setTextDirUCC(artifactName); //includes module related information

		var resourceHyperlink = null;
		// artifact name
		if (artifactType == null) {
			var node = com.ibm.rdm.web.util.domutils.createSpannedTextNode("com-ibm-rdm-comment-artifact-not-exist", messages.artifactNotExist);
			appendTo.appendChild(node);
		} else {
			var lc = null;
			var gc = null;
			if (VvcUtils.isGlobalConfigurationUri(configurationUri)) {
				gc = configurationUri;
			} else {
				lc = configurationUri;
			}
			resourceHyperlink = ResourceLinksHandler.createLink({
				uri: comment.resourceUri,
				label: artifactFullLabel,
				commentURI: comment.linkSelfUri,
				viewlet: viewlet,
				usedInModuleArgs:{uri:comment.module, title:comment.moduleTitle},
				configurationUsed: lc,
				globalConfigurationURI: gc
			});
			// Using own will ensure that the resource link is destroyed when widget is destroyed.
			 widget.own(dojo.connect(resourceHyperlink, "onclick", function(event) {
				if (popupWidget) {
					popupWidget.close();
				}
			}));
			
			var wrapper = dojo.create('div', {"class": "com-ibm-rdm-comment-artifact-link-wrapper"}, appendTo);
			wrapper.appendChild(resourceHyperlink);
			dojo.addClass(resourceHyperlink, "com-ibm-rdm-comment-artifact-hyperlink");
		}
		return resourceHyperlink;
	},
	
	_createArtifactImage: function(artifactFormat, componentURI, artifactURI) {
		//Creates an Artifact Image for an artifact type only. Used in reviews, using the artifact format only. DO NOT USE for general icon as
		//it does not reflect the chosen image based on artifact type
		var artifactImg = document.createElement("img");
		var imgURI;
		if (!artifactFormat) {
			// Reviews do not have an artifact format.  determine whether the artifact is a review and give
			// it an appropriate icon.  Would be better if we could get a some indication of the type from the,
			// server as part of the comment, but that is not available, so we need to look at the artifat URI
			// to deduce that we are dealing with a review.  Work item 51490.
			if (artifactURI && artifactURI.indexOf("/reviews/")) {
				imgURI = dojo.moduleUrl("com.ibm.rdm.web", "review/images/icons/obj16/review.gif");
			}
		} else {
			imgURI =
				ArtifactIcons.getSmallImageURI(componentURI, artifactFormat, artifactFormat);
		}
		if (!imgURI) {
			imgURI = dojo.moduleUrl("com.ibm.rdm.web", "common/images/resources/unknown_obj.gif");
		}
		
		artifactImg.src = imgURI;
		dojo.addClass(artifactImg, "com-ibm-rdm-comment-artifact-image");
		return artifactImg;
	},
	
	
	createTitleNode: function(widget, appendTo, comment, replyComment, commentingWidget, isOrphaned) {
		// priority image
		if (comment.priority=="High"){
			var priorityImg = document.createElement("img");
			var icon = com.ibm.rdm.web.commenting.utils.getPriorityImageSrc(comment.priority, comment.isResolved);
			priorityImg.src = icon;
			dojo.addClass(priorityImg, "com-ibm-rdm-comment-priority-imag");
			appendTo.appendChild(priorityImg);
		}
			
		// title
		var title = comment.title;
		if (!comment.title || comment.title.length == 0) {
			title = messages.noTitle;
		}
		var commentTitleNode = com.ibm.rdm.web.search.domutils.createSpannedTextNode("com-ibm-rdm-comment-title", title);
		if (comment.priority=="High") {
			dojo.addClass(commentTitleNode, "com-ibm-rdm-comment-title-high-priority");
		}
		BidiUtils.applyTextDirSetting(commentTitleNode, title);
		
		appendTo.appendChild(commentTitleNode);
		if (comment.isResolved) {
			dojo.addClass(commentTitleNode, "com-ibm-rdm-resolved-comment-text");
		}
		
		if (replyComment) {
			var message = dojo.string.substitute(messages.replyCommentLink, [replyComment.commentNumber]);
			var reElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode("rdm-annotation-commentReplyFor j-hint", " " + message);
			appendTo.appendChild(reElement);
			var replyHover = this._hookupReplyToLink(widget, reElement, replyComment, commentingWidget);
			return replyHover;
		} else if (isOrphaned) {
			var reElement = com.ibm.rdm.web.search.domutils.createSpannedTextNode("rdm-annotation-commentReplyFor j-hint", " " + messages.orphanedCommnentLinkLabel);
			appendTo.appendChild(reElement);
		}
		
		return null;
	},
		
	_handleCommentArtifactLinkClicked: function(event, commentURI, viewletContext) {
		ResourceLinksHandler.handleCommentArtifactAnchorClicked(event, commentURI, null, viewletContext);
	},
	
	openEditCommentForm: function(formData, componentURI, commentingWidget, resourceURI, closeCallback, openCallback) {
		var form =  com.ibm.rdm.web.commenting.EditCommentForm.getInstance({
				commentingWidget: commentingWidget,
				componentURI: componentURI,
				resourceURI: resourceURI,
				formData: formData,
				onDialogClose: closeCallback,
				onDialogOpen: openCallback
			});
		form.show();
	},
	
	createArtifactComment: function(resourceURI, componentURI, displayName) {
		// summary:
		//		Display a form to create a new comment and, when OK'd, create
		//		the comment.
		// description:
		//		This method should only be called when it's true that the current
		//		page DOES NOT contain a "Comments" sidebar section (e.g. from the
		//		Artifacts page).  This method enables the user to create a comment
		//		and, after successfully saving a new comment, it doesn't attempt
		//		to update a "Comments" sidebar section.

		var formData = {
			formTitle: dojo.string.substitute(
				messages.artifactFormTitle, [displayName]
			),
			commentTitle: "",
			resourceLocationId: null,
			locationPane: null,
			directedToUsers: []
		};
		var skeletalCommentingWidget = {
			getCommentReviewId: function() {
				return "";
			},
			saveNewComment: dojo.hitch(this, function(newComment) {
				var deferred = com.ibm.rdm.web.commenting.CommentingClient.saveNewComment(
					newComment, componentURI
				);
				deferred.then(
					dojo.hitch(this, function(comment){
						com.ibm.rdm.web.notifications.ChangeNotifier.notifyListeners({
							eventTopic: com.ibm.rdm.web.notifications.EventTopics.COMMENTS_CHANGED,
							resourceUriScope: comment.resourceUri
						});
					})
				);
				return deferred;
			})
		};
		this.openEditCommentForm(
			formData, componentURI, skeletalCommentingWidget, resourceURI
		);
	},

	_makeCommentEmailBody: function(authorName, comment, /*optional*/ resourceTitle) {
		// summary:
		//		Creates an HTML fragment suitable for inclusion in an email message.
		//		The message consists of an optional link, followed by some optional text content.
		// authroName:
		//		The full name of the comment author.
		// comment:
		//		The comment that the email concerns.
		// resourceTitle:
		//		The title of the resource relating to the comment.
	
		// Put in an empty paragraph at the top to give them a place to start typing.
		var body = "<p></p>";
		
		var resourceLink = DomUtils.makeAnchorString(comment.resourceUri, resourceTitle);
		var message = dojo.string.substitute(messages["commentedOn"], [authorName, resourceLink]);
		body += '<p style="color:#777777">' + message + "</p>";
		
		if (comment.title || comment.text) {
			body += '<blockquote style="margin:0.5em 1em;padding:0.5em 1em;background-color:#eeeeee">';
			if (comment.title) {
				body += '<p style="color:#3087B3;font-weight:bold;">' + comment.title + '</p>';
			}
			if (comment.text) {
				var text = DomUtils.convertLinksToAbsolute(comment.text, com.ibm.rdm.web.FRONTING_SERVICE_COMMENTS);
				body += text;
			}
			body += '</blockquote>';
		}
		return body;
	},
	
	truncateText: function(text, maxLength) {
		// summary:
		//		Truncates a text string to a suitable length for display.
		// text: String
		//		A string to be truncated if necessary
		// maxLength: int
		//		The maximum length of the string before it is truncated.
		// returns:
		//		The supplied string if less than maxLength, or the string truncated
		//		to maxLength with an elipsis appended.

		if (text == null || text == "") {
			return "";
		}
		if (text.length > maxLength) {
			text = text.substring(0,maxLength);
			return dojo.string.substitute(messages.valueMore, [text]);
		} else {
			return text;
		}
	},
	

	makeCombinedDirectedToList: function(comment) {
		// summary:
		//		Creates a combined list of users to direct a comment to, consisting of the author and
		//		the directed-to list from a previous comment.
		// comment:
		//		An existing comment
		// returns:
		//		An array populated with names of the comment author, and the directed to users, and
		//		excluding the current user.  There may be duplicates in this list.
		
		var thisUser = com.ibm.rdm.web.AUTHENTICATED_USER.userId;
		var author = comment.author;
		var combinedList = dojo.filter(comment.directedToUsers, function(item){
			return item != thisUser;
		});
		if (thisUser != author) {
			combinedList.push(author);
		}
		return combinedList;
	},
	
	makeReplyTitle: function(origCommentTitle) {
		// summary:
		//		Generates a title for a comment which is a rely to an existing comment.
		// origCommentTitle: string
		//		The title of the comment being replied to.
		// returns:
		//		The original title with a RE: prefix added, unless it already has one.
		var prefix = messages.reBarePrefix;
		// There is an assumption here that the prefix really is a prefix. If it is not
		// at the start of the title it may be added multiple times on successive replies,
		// which is not so neat, but still safe.
		if (origCommentTitle.indexOf(prefix) == 0) {
			return origCommentTitle;
		}
		return dojo.string.substitute(messages.re, [origCommentTitle]);
	}
	
};

})();
