/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2013, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function(){
 
dojo.provide("com.ibm.rdm.web.compare.ModuleStructureCompareWidget");
 
dojo.require("dijit._Templated");
dojo.require("dijit._Widget");
dojo.require("dojox.gfx");

dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.common.OverlayBorder");
dojo.require("com.ibm.rdm.web.compare.CompareConstants");
dojo.require("com.ibm.rdm.web.compare.compareutils");
dojo.require("com.ibm.rdm.web.compare.moduleCompareUtils");
dojo.require("com.ibm.rdm.web.compare.compareTagUtils");
dojo.require("com.ibm.rdm.web.compare.compareLinksUtils");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.server.client.utils.WrapperResourceUtil");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.util.keyutils");
dojo.require("com.ibm.rdm.web.comparevvc.vvccompareutils");
dojo.require("com.ibm.rdm.web.compare._ModuleCompareWidgetBase");

dojo.requireLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");

var CompareConstants = com.ibm.rdm.web.compare.CompareConstants,
	CompareUtils = com.ibm.rdm.web.compare.compareutils,
	ModuleCompareUtils = com.ibm.rdm.web.compare.moduleCompareUtils,
	DomUtils = com.ibm.rdm.web.util.domutils,
	VVCCompareUtils = com.ibm.rdm.web.comparevvc.vvccompareutils,
	_ModuleCompareWidgetBase = com.ibm.rdm.web.compare._ModuleCompareWidgetBase,
	KeyUtils = com.ibm.rdm.web.util.keyutils,
	messages;  // initialised on creation of first instance of the widget.


var ROW_URI_ATTR = "data-rdm-uri",
	ROW_CONTENT_WIDTH_ADJUST_ATTR = "data-rdm-width-adjust",
	CONTENT_COL_HEADING_TAG_ATTR = "data-rdm-heading-tag",
	MOVED = "moved",
	MOVED_FROM = "movedFrom",
	ADDED = "added",
	REMOVED = "removed",
	CONNECTION_COLOR_MOVED = "#999999",
	CONNECTION_COLOR_MOVED_FADED = "#cccccc",
	CONNECTION_COLOR_REMOVED = "#999999",
	CONNECTION_COLOR_REMOVED_FADED = "#cccccc",
	CONNECTION_COLOR_ADDED = "#999999",
	CONNECTION_COLOR_ADDED_FADED = "#cccccc",
	CONNECTION_SURFACE_WIDTH = 100,
	CONNECTION_SURFACE_DEFAULT_HEIGHT = 600,
	CONNECTION_CURVE_CONTROL_LEFT = 40,
	CONNECTION_CURVE_CONTROL_RIGHT = CONNECTION_SURFACE_WIDTH - CONNECTION_CURVE_CONTROL_LEFT;
	
// A dummy partial resource that is used to represent the root node of the tree represented on
// one of the grids.  The root node is not provided as part of the view results we get from the
// server, but we do need to display this notional node to give a place for insertion of
// placeholder nodes above the first item on the grid.
// Contains just the fields needed to keep the row generation code happy.
var ROOT_ROW_DATA = {
	getResourceURI: function() {
		return CompareConstants.STRUCTURE_DIFF_ROOT_URI;
	},
	depth: 0,
	// Additional field not in partial resource for easy check for root node.
	isRoot: true
};

// "Private" helpers
function decrementDepthCounters(depthCounterArray) {
	// Decrements each entry in an array.  If any entry becomes zero it is removed.
	// If all works well in the data we are given each value should be equal to or higher than
	// all values that follow it in the array.
	//
	// returns true if at least array entry was removed after reaching zero.
	
	var i = depthCounterArray.length,
		removedEntry = false;
	while (i--) {
		if (--depthCounterArray[i] <= 0) {
			// this counter is zero, so remove it.
			depthCounterArray.splice(i, 1);
			removedEntry = true;
		}
	}
	return removedEntry;
}

function addMovedRowDecorations(cellChangeData, depthCounterArray, classArray, contentParent) {
	// Calculates the classes to add to a cell to indicate that it
	// is part of one or more moves.  Also, adds wrapping nodes to suit the current move
	// depth.  The new nodes are added inside the contentParent, and the new contentParent
	// (the deepest wrapper) is returned.
	// Only for use on nodes that may be moves.  Adds and removes are known not to be moves
	// so should use addSimpleMoveBlockNesting() instead.
	var isBlockStart;
	if (cellChangeData && cellChangeData.movedBlock) {
		// starting a new block.  Add its count to the stack, and display as a
		// blockStart
		depthCounterArray.push(cellChangeData.movedBlock.size);
		classArray.push("blockStart");
		isBlockStart = true;
	}
	
	// Make up an array corresponding to the depth counter array.  This will hold the classes to be
	// applied to each nesting level
	var nestClasses = [];
	var moveDepth = depthCounterArray.length;
	while (moveDepth--) {
		nestClasses.push("nest");
	}
	nestClasses[nestClasses.length-1] += " deepestNest";
	
	
	if (isBlockStart) {
		// deepest nesting is a block start.
		nestClasses[nestClasses.length-1] += " nestStart";
	}
	
	// The number of moved blocks this node is in is indicated by the stack size.
	// But exclude added and removed blocks from any move marking - they are not
	// included in the count.
	var initialMoveDepth = depthCounterArray.length;
	if (initialMoveDepth) {
		classArray.push("moved");
		classArray.push("moveDepth" + ((initialMoveDepth >= 4) ? "4plus" : initialMoveDepth));
		if (decrementDepthCounters(depthCounterArray)) {
			classArray.push("blockEnd");
		}
		
		// compare the initialMoveDepth with the new value, and set the
		// corresponding values in the nestClasses array, depending on how many blocks have closed.
		var finalMoveDepth = depthCounterArray.length;
		if (finalMoveDepth < initialMoveDepth) {
			var closedCount = initialMoveDepth - finalMoveDepth;
			moveDepth = initialMoveDepth;
			while (closedCount--) {
				moveDepth--;
				nestClasses[moveDepth] += " nestEnd";
			}
		}
		
		// Create the nesting wrappers with the appropriate classes
		for (moveDepth = 0 ; moveDepth < initialMoveDepth ; moveDepth++) {
			contentParent = dojo.create('div', {"class": nestClasses[moveDepth]}, contentParent);
		}
	}
	return contentParent;
}

function addSimpleMoveBlockNesting(depthCounterArray, contentParent) {
	// summary:
	//		Adds nesting wrappers corresponding to the move depth array.  This does not
	//		change the move nesting level, so is not for use on nodes that may be a move
	//		themselves (so they may start or end blocks.
	// depthCounterArray:
	//		Array of move block length counters.
	// contentParent:
	//		The parent node to contain the nesting nodes.
	// returns:
	//		The deepest nesting node added, or the supplied contentParent
	//		if no nesting is added.
	var nestDepth = depthCounterArray.length;
	while(nestDepth--) {
		contentParent = dojo.create('div', {"class": "nest"}, contentParent);
	}
	return contentParent;
}

function buildChangeMaps(diffRecords) {
	// summary:
	//		Builds maps containing the changes to be displayed on the left and right sides
	// returns:
	//		An object contining two hashes (one left and one right).  Each hash is keyed by
	//		URI, and the value is like:
	//			{
	//				addedBlock: {size: 1},
	//				removedBlock: {size: 1},
	//				movedBlock: {size: 1},
	//				addMarkers: array of URIs.  Each URI represents an added block.
	//				removeMarkers: array of URIs.  Each URI represents an added block.
	//			}
	//		The position markers indicate that there is an insert/remove point (i.e. line
	//		between the rows) after the row of the entry.  There can be many position markers.
	//		Only one of addBlock/removeBlock/moveBlock is set at once.  The size of a move block
	//		excludes any contained add and remove blocks.
	
	var totalRecords = diffRecords.length,
		left = {},		// Maps of URI to change for nodes on each side of the display.
		right = {},
		record, leftChange, rightChange, positionMarkers;
		
	for (var i = 0 ; i < totalRecords ; i++) {
		record = diffRecords[i];
		switch(record.type) {
			case CompareConstants.STRUCTURE_DIFF_TYPE_ADD:
				// start of an add block on the right side
				rightChange = right[record.binding] || (right[record.binding] = {});
				rightChange.addedBlock = {
					size: record.afterBlockSize
				};
				// add a postion marker on the left to indicate the location of the addition		
				leftChange = left[record.positionBinding] || (left[record.positionBinding] = {});
				positionMarkers = leftChange.addMarkers || (leftChange.addMarkers = []);
				positionMarkers.push(record.binding);
				break;
			case CompareConstants.STRUCTURE_DIFF_TYPE_REMOVE:
				// start of a remove block on the left side
				leftChange = left[record.binding] || (left[record.binding] = {});
				leftChange.removedBlock = {
					size: record.beforeBlockSize
				};
				// add a position marker on the right to indicate the location of the removal
				rightChange = right[record.positionBinding] || (right[record.positionBinding] = {});
				positionMarkers = rightChange.removeMarkers || (rightChange.removeMarkers = []);
				positionMarkers.push(record.binding);
				break;
			case CompareConstants.STRUCTURE_DIFF_TYPE_MOVE:
				// start of a move block on the left side
				leftChange = left[record.binding] || (left[record.binding] = {});
				leftChange.movedBlock = {
					size: record.beforeBlockSize
				};
				// start of a move block on the right side
				rightChange = right[record.binding] || (right[record.binding] = {});
				rightChange.movedBlock = {
					size: record.afterBlockSize
				};
				break;
		}
	}
	
	return {
		left: left,
		right: right
	};
}

dojo.declare("com.ibm.rdm.web.compare.ModuleStructureCompareWidget", [dijit._Widget, dijit._Templated, _ModuleCompareWidgetBase], {
		
	templatePath: dojo.moduleUrl("com.ibm.rdm.web", "compare/templates/ModuleStructureCompareWidget.html"),
	widgetsInTemplate: true,
	
	// set in markup on construction
	width: 400,
	height: 400,
	
	_gridArea: null,
	_leftGrid: null,
	_rightGrid: null,
	_leftSizeWrapper: null,
	_rightSizeWrapper: null,
	_leftSelectionOverlay: null,
	_rightSelectionOverlay: null,
	_leftHoverOverlay: null,
	_rightHoverOverlay: null,
	_leftGridScroller: null,
	_rightGridScroller: null,
	_connectionSurfaceNode: null,
	_connectionSurface: null,
	_displayedConnections: null,
	_scrollAnimation: null,
	
	// Flags indicating what changes are being displayed.
	_showingAdds: true,
	_showingRemoves: true,
	_showingMoves: true,
	_showingMovesFrom: false,
	
	// Description of the selection for each grid.  Null or undefined if there is no selection.
	// Each entry is of the form:
	// 		{rowNode: node, hightlightNode: node}
	_leftSelection: null,
	_rightSelection: null,
	
	// Indicates whether to show a single grid (add or remove whole module) or dual grid (compare module)
	_isSingleGrid: false,
	
	// The current height of the connection surface.
	_connectionSurfaceHeight: CONNECTION_SURFACE_DEFAULT_HEIGHT,
	
	constructor: function() {
		if (!messages) {
			messages = dojo.i18n.getLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");
		}
	},
	
	postCreate: function(){
		this._connectionSurface = dojox.gfx.createSurface(this._connectionSurfaceNode, CONNECTION_SURFACE_WIDTH, this._connectionSurfaceHeight);
		this._displayedConnections = [];
				
		this.connect(this._leftGridScroller, "onscroll", this._handleLeftPaneScroll);
		this.connect(this._rightGridScroller, "onscroll", this._handleRightPaneScroll);
		this.connect(this._leftGrid, "onclick", dojo.hitch(this, this._onGridClick, true));
		this.connect(this._rightGrid, "onclick", dojo.hitch(this, this._onGridClick, false));
		this.connect(this._leftGrid, "onmouseleave", dojo.hitch(this, this._onGridMouseLeave, true));
		this.connect(this._rightGrid, "onmouseleave", dojo.hitch(this, this._onGridMouseLeave, false));
		this.connect(this._leftGrid, "onmousemove", dojo.hitch(this, this._onGridMouseMove, true));
		this.connect(this._rightGrid, "onmousemove", dojo.hitch(this, this._onGridMouseMove, false));
		
		// Define key handlers for intercepting keypresses.  In addition to the values specified in
		// KeyUtils, we also add an isLeftGrid entry to each.  This allows us to use common handler
		// functions that behave slightly differently for left and right grids.
		var leftKeyHandlers = [
			{
				keyCode: dojo.keys.UP_ARROW,
				onKeyPress: dojo.hitch(this, this._handleKeyUpDown, true),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.DOWN_ARROW,
				onKeyPress: dojo.hitch(this, this._handleKeyUpDown, false),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.UP_ARROW,
				ctrl: true,
				onKeyPress: dojo.hitch(this, this._handleKeyNextPrevChange, true),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.DOWN_ARROW,
				ctrl: true,
				onKeyPress: dojo.hitch(this, this._handleKeyNextPrevChange, false),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.LEFT_ARROW,
				onKeyPress: dojo.hitch(this, this._handleKeyLeftRight, true),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.RIGHT_ARROW,
				onKeyPress: dojo.hitch(this, this._handleKeyLeftRight, false),
				isLeftGrid: true
			},
			{
				keyCode: dojo.keys.ENTER,
				onKeyPress: dojo.hitch(this, this._handleKeyAction),
				isLeftGrid: true
			},
			{
				keyChar: ' ',	// SPACEBAR - do the same as ENTER
				onKeyPress: dojo.hitch(this, this._handleKeyAction),
				isLeftGrid: true
			}
		];
		// Make an identical handlers array for right grid, but with isLeftGrid false in each entry.  Make sure
		// we don't change the leftKeyHandlers array or entries within it.
		var rightKeyHandlers = dojo.map(leftKeyHandlers, function(keyDef){
			keyDef = dojo.mixin({}, keyDef);	// shallow copy of entry.
			keyDef.isLeftGrid = false;
			return keyDef;
		});
		this.connect(this._leftGrid, "onkeypress", KeyUtils.getKeyHandler(leftKeyHandlers));
		this.connect(this._rightGrid, "onkeypress", KeyUtils.getKeyHandler(rightKeyHandlers));
		
		this._resizeImpl();
	},
	
	onColumnResize: function (leftWidth, rightWidth, rightPosition) {
		// summary:
		//		Overrideable method that will be called when the columns have been resized.
		//		This is useful for when we want to resize elements not contained within this
		//		widget based upon the sizes of the columns.
		// params:
		//		leftWidth: the width of the left column
		//		rightWidth: the width of the right column
		//		rightPosition: the x-position within the parent at which the right column is placed
	},
	
	resize: function(newWidth, newHeight) {
		this.width = newWidth < 300 ? 300 : newWidth;
		this.height = newHeight < 10 ? 10 : newHeight;
		this._resizeImpl();
	},
	
	_resizeImpl: function() {
		var totalWidthAvailable = this.width,
			leftWidth = Math.floor((totalWidthAvailable - CONNECTION_SURFACE_WIDTH) / 2),
			totalHeightAvailable = this.height,
			gridHeight = totalHeightAvailable - 1, // allow for 1px top border
			gridHeightStr = gridHeight + "px",
			rightWidth, rightLeftPosition;
			
		dojo.style(this.domNode, {
			width: totalWidthAvailable+"px",
			height: totalHeightAvailable+"px"
		});
		
		dojo.style(this._gridArea, {
			width: totalWidthAvailable+"px",
			height: gridHeightStr
		});

		dojo.style(this._leftSizeWrapper, {
			width: leftWidth+"px",
			height: gridHeightStr
		});
		
		dojo.style(this._leftGrid, {
			height: gridHeightStr
		});

		
		if (this._isSingleGrid) {
			// right grid is full width
			rightWidth = totalWidthAvailable;
			rightLeftPosition = 0;
		} else {
			// right grid is on the right
			rightWidth = totalWidthAvailable - CONNECTION_SURFACE_WIDTH - leftWidth;
			rightLeftPosition = leftWidth + CONNECTION_SURFACE_WIDTH;
		}

		dojo.style(this._rightSizeWrapper, {
			width: rightWidth+"px",
			height: gridHeightStr,
			left: rightLeftPosition + "px"
		});
		dojo.style(this._rightGrid, {
			height: gridHeightStr
		});
		
		this._resizeWideContentNodes();
		
		dojo.style(this._connectionSurfaceNode, {
			width: CONNECTION_SURFACE_WIDTH,
			left: leftWidth+"px"
		});
		
		this._connectionSurface.setDimensions(CONNECTION_SURFACE_WIDTH, gridHeight);
		this._connectionSurfaceHeight = gridHeight;
		this._showConnections();
		this._redrawSelectionOverlays();
		
		this.onColumnResize(leftWidth, rightWidth, rightLeftPosition);
	},
	
	_selectSingleGridDisplay: function() {
		this._isSingleGrid = true;
		dojo.addClass(this.domNode, "singleGrid");
	},
	
	_selectDoubleGridDisplay: function() {
		this._isSingleGrid = false;
		dojo.removeClass(this.domNode, "singleGrid");
	},
	
	onSelect: function(params) {
		// summary:
		//		Called when an artifact is selected.
		// params:
		//		object of the form:
		//		{
		//			fromArtifactURI: artifactURI1,
		//			toArtifactURI: artifactURI2,
		//			displayType: one of the DISPLAY_AS_XXXX constants
		//			resourceDeltas: hash of delta types keyed by concept URI
		//		}
	},

	
	show: function(params){
		// summary:
		//		Display a structure difference, using supplied view and difference information.
		//		It is possible for there to be only a single view, and no structure diff information, in
		//		the case where the module has been created or deleted.  In such a case we just display the
		//		single view.
		this._sourceOnLeft = params.sourceOnLeft;
		this._compareContext = params.compareContext;
		this._resourceDeltas = params.resourceDeltas;
		this._displayType = params.displayType;
		var left = params.sourceOnLeft ? params.sourceModuleView : params.targetModuleView,		// ordered array of partial resources for left column
			right = params.sourceOnLeft ? params.targetModuleView : params.sourceModuleView,	// ordered array of partial resources for right column
			changes, valueChangeMap;
		// If we are showing an addition or deletion, we only have left or right.  Make sure that we either have both or
		// only right.
		if (!right) {
			right = left;
			left = null;
		}
		if (params.diff) {
			// calculate hashes for changes to display on left and right, and value changes
			changes = buildChangeMaps(params.diff);
			valueChangeMap = ModuleCompareUtils.buildValueChangeMap(left, right, params.resourceDeltas, params.moduleURI);	// Hash with key for rows that are different on left and right
			this._selectDoubleGridDisplay();
		} else {
			// There is no diff to display.  To get here we should only be displaying the right-hand grid.
			valueChangeMap = {};		// indicates no value changes.
										// TODO: Maybe we should be highlighting artifacts that
										// have changed between the configurations.
			this._selectSingleGridDisplay();
		}
		
		this.onColumnResize = params.onColumnResize || this.onColumnResize;
		
		this.clear();
		
		var rowClasses, row, contentWrapper, change, movedBlockCountStack, contentParent, previousRow,
			rowURI, removedBlockSize, addedBlockSize, depth, positionMarkers, rowData, i, valueChanged;
		
		if (left) {
			// display left grid
			// start our loop counter at -1 so that we go round the loop once for the
			// root node, which is not in the array.
			movedBlockCountStack = [];
			for (i = -1 ; i < left.length ; i++) {
				rowData = i < 0 ? ROOT_ROW_DATA : left[i];
				if (rowData) {
					rowURI = rowData.getResourceURI();
					// Create the outline of the row
					rowClasses = ["row"];
					row = dojo.create("div", {
						id: this._makeLeftNodeId(rowURI)
					}, this._leftGrid);
					dojo.attr(row, ROW_URI_ATTR, rowURI);
					rowClasses.push((rowData.depth >= 10) ? "depth10plus" : "depth"+rowData.depth);
					
					change = changes.left[rowURI];
					contentParent = row;
					
					
					// Add classes for removed rows
					if (change && change.removedBlock) {
						removedBlockSize = change.removedBlock.size;
						rowClasses.push("blockStart");
					}
					if (removedBlockSize) {
						removedBlockSize--;
						rowClasses.push("removed");
						if (removedBlockSize === 0) {
							rowClasses.push("blockEnd");
						}
						// add nesting to allow for this operation being contained in a move block.
						contentParent = addSimpleMoveBlockNesting(movedBlockCountStack, contentParent);
					} else {
						// Add classes and nesting wrappers for moved rows and blocks contained in move blocks.
						contentParent = addMovedRowDecorations(change, movedBlockCountStack, rowClasses, contentParent);
					}
					
					// Add the row content
					if (!rowData.isRoot) {
						contentParent.appendChild(this._buildRowContent(true/*isLeftGrid*/, rowData));
					}
					
					// Add placeholders for the positions of added blocks		
					if (change && change.addMarkers) {
						var addedBlockPlaceholder = dojo.create("div", {"class": "addedPlaceholder"}, contentParent);
						positionMarkers = change.addMarkers;
						dojo.forEach(positionMarkers, function(markerURI){
							dojo.create("div", {
								id: this._makeLeftNodeId(markerURI)
							}, addedBlockPlaceholder);
						}, this);
					}
					
					// Add indication of changed value
					if (valueChangeMap[rowURI]) {
						rowClasses.push("valueChanged");
					}
	
					dojo.addClass(row, rowClasses);
	
					// suppress the dividing line immediately before the start of a block.
					if (previousRow && dojo.indexOf(rowClasses, "blockStart") >= 0) {
						dojo.addClass(previousRow, "noDivider");
					}
					previousRow = row;
				}
			}
		}
		
		// right grid
		// start our loop couter at -1 so that we go round the loop once for the
		// root node, which is not in the array.
		movedBlockCountStack = [];
		for (i = -1 ; i < right.length ; i++) {
			rowData = i < 0 ? ROOT_ROW_DATA : right[i];
			if (rowData) {
				rowURI = rowData.getResourceURI();
				// create the outline of the row
				rowClasses = ["row"];
				row = dojo.create("div", {
					id: this._makeRightNodeId(rowURI)
				}, this._rightGrid);
				dojo.attr(row, ROW_URI_ATTR, rowURI);
				rowClasses.push((rowData.depth >= 10) ? "depth10plus" : "depth"+rowData.depth);
				
				change = changes && changes.right[rowURI]; // tolerate there being no changes - for added or removed module.
				contentParent = row;
								
				// Add classes for added rows
				if (change && change.addedBlock) {
					addedBlockSize = change.addedBlock.size;
					rowClasses.push("blockStart");
				}
				if (addedBlockSize) {
					addedBlockSize--;
					rowClasses.push("added");
					if (addedBlockSize === 0) {
						rowClasses.push("blockEnd");
					}
					// add nesting to allow for this operation being contained in a move block.
					contentParent = addSimpleMoveBlockNesting(movedBlockCountStack, contentParent);
				} else {
					// Add classes and nesting wrappers for moved rows and blocks contained in move blocks.
					contentParent = addMovedRowDecorations(change, movedBlockCountStack, rowClasses, contentParent);
				}
				
				// Add the row content
				if (!rowData.isRoot) {
					valueChanged = valueChangeMap[rowURI];

					contentParent.appendChild(this._buildRowContent(false/*isLeftGrid*/, rowData, valueChanged));
					// Allow styling to indicate changed value
					if (valueChanged) {
						var thisDelta = this._resourceDeltas[rowURI];
			
						if (VVCCompareUtils.hasOnlyDeltaType(thisDelta, CompareConstants.DELTA_TYPE_LINKS_CHANGE)) {
							rowClasses.push("linkEndpointOnly");
						}
							
						rowClasses.push("valueChanged");
					}
				}
				
				// Add placeholders for the positions of removed blocks
				if (change && change.removeMarkers) {
					var removedBlockPlaceholder = dojo.create("div", {"class": "removedPlaceholder"}, contentParent);
					positionMarkers = change.removeMarkers;
					dojo.forEach(positionMarkers, function(markerURI){
						dojo.create("div", {
							id: this._makeRightNodeId(markerURI)
						}, removedBlockPlaceholder);
					}, this);
				}
				
				// Add placeholders for the moved from positions of moved blocks
//				if (change && change.movedBlocks) {
//					var movedBlockPlaceholder = dojo.create("div", {"class": "movedPlaceholder"}, contentParent);
//					var movedBlocks = change.movedBlocks;
//					dojo.forEach(movedBlocks, function(movedBlock){
//						dojo.create("div", {
//							id: this._makeRightMovedFromNodeId(movedBlock.uri)
//						}, movedBlockPlaceholder);
//					}, this);
//				}

				dojo.addClass(row, rowClasses);
				
				// suppress the dividing line immediately before the start of a block.
				if (previousRow && dojo.indexOf(rowClasses, "blockStart") >= 0) {
					dojo.addClass(previousRow, "noDivider");
				}
				previousRow = row;
			}
		}
		
		this._showConnections();
		this._selectInitialRow();
		this._connectImgLoadToAnnotationRedraw();
		this._showAnnotations();
	},
	
	clear: function() {
		// summary:
		//		Clears the widget's display
		dojo.empty(this._leftGrid); 
		dojo.empty(this._rightGrid);
		this._clearConnections();
		this._leftSelectionOverlay.hide();
		this._rightSelectionOverlay.hide();
		this._leftHoverOverlay.hide();
		this._rightHoverOverlay.hide();
		this._clearAnnotations();
	},
	
	_buildRowContent: function(isLeftGrid, rowData, valueChanged) {
		// summary:
		//		Builds a DOM node to represent the content of a row.  This does not include
		//		the nesting wrapppers for change outline boxes.
		// isLeftGrid:
		//		Boolean indicating which grid the row will be on.
		// rowData:
		//		Partial Requirement to be displayed.
		// valueChanged:
		//		true-ish to indicate that the content has changed.
		// returns:
		//		A DOM node containing a display representation of the row.
		
		var contentWrapper = dojo.create("div", {"class": "contentWrapper"}),
			innerBox = dojo.create("div", {"class": "innerBox"}, contentWrapper),
			table = dojo.create('table', {"class": "rowCells"}, innerBox),
			tbody = dojo.create('tbody', null, table),
			row = dojo.create('tr', null, tbody),
			idCell = dojo.create('td', {"class": "idColumn"}, row),
			ptCell = dojo.create('td', {"class": "contentColumn"}, row),
			displayContent = dojo.create('div', {"class": "displayContent initialContent show"}, ptCell),
			wrappedResourceMimeType, ptWrapper;
		
		if (!rowData.getID()) {
			// The target configuration does not contain the artifact for this row.  Display a simple
			// row containing just a message to indicate a binding without a matching artifact.
			ptWrapper = dojo.create('div', {"class":"textContent"}, displayContent);
			ptWrapper.appendChild(document.createTextNode(messages.unboundArtifact));
			dojo.addClass(contentWrapper, "unboundArtifactNode");
			return contentWrapper;
		}
		
		idCell.appendChild(document.createTextNode(rowData.getID()));
		
		// Check to see if the heading info has changed. If it has, decorate
		//	it accordingly.
		var showHeadingChange = false,
			isSourceGrid = (!this._sourceOnLeft && !isLeftGrid) || (this._sourceOnLeft && isLeftGrid);
		if (isSourceGrid && this._resourceDeltas[rowData.resourceURI]) {
			var deltaEntry = this._resourceDeltas[rowData.resourceURI];
				
			// The row has changed, and we have a delta entry for it. Check to see if it exists in both configs and if the heading info potentially changed.
			if(VVCCompareUtils.hasDeltaType(deltaEntry,
					[
					 	CompareConstants.DELTA_TYPE_SOURCE_CHANGE,
					 	CompareConstants.DELTA_TYPE_TARGET_CHANGE,
					 	CompareConstants.MOVED_IN_SOURCE,
					 	CompareConstants.MOVED_IN_TARGET
					 ])) {
				showHeadingChange = true;
			}
		}
		
		
		if (!isLeftGrid && this._displayType !== CompareConstants.DISPLAY_AS_DELETED) {
			// Add the Show Changes link
			var selector = dojo.create('div', {"class":"contentSelector"});
			var link = dojo.create('a', {"class":"showChangesLink","tabIndex":"-1"}, selector);
			link.setAttribute('href', '#');
			if (valueChanged) {
				link.appendChild(document.createTextNode(messages.showChanges));
			} else {
				link.appendChild(document.createTextNode(messages.showContent));
			}
			displayContent.appendChild(selector);
		}
		
		
		
		if ((wrappedResourceMimeType = rowData.getWrappedResourceMimeType())) {
			// It's a wrapped resource
			var wrappedResourceConfiguration;
			if (isLeftGrid) {
				wrappedResourceConfiguration = this._sourceOnLeft ? this._compareContext.sourceConfiguration : this._compareContext.targetConfiguration;
			} else {
				wrappedResourceConfiguration = this._sourceOnLeft ? this._compareContext.targetConfiguration : this._compareContext.sourceConfiguration;
			}
			
			this._displayWrappedResource(displayContent, rowData, wrappedResourceConfiguration);
		} else if (rowData.isHeading() || showHeadingChange) {
			// 95455: We still display any heading numbers if showHeadingChange is true, but only
			//	decorate the row as a heading if rowData.isHeading() is true.
			var headingTag = rowData.isHeading() ? 'h' + (rowData.depth > 6 ? 6 : rowData.depth) : "div"; 
			
			var headingNode = dojo.create(headingTag, {"class":"textContent"});
			var sectionNumberElement = dojo.create('div', {"class": "sectionNumber"}, headingNode);
			
			if (showHeadingChange) {
				// Find the opposite corresponding row, and use that to compare when displaying header information
				var oppositeId = isLeftGrid ? this._getRightIdFromLeft(rowData.resourceURI) : this._getLeftIdFromRight(rowData.resourceURI);
				var oppositeRowNode = dojo.query("#"+oppositeId)[0],
					oppositeHeadingText = "";
				if (oppositeRowNode) {
					var oppositeHeadingNode = dojo.query(".sectionNumber", oppositeRowNode)[0];
					if (oppositeHeadingNode) {
						oppositeHeadingText = oppositeHeadingNode.textContent;
					}
				}
				sectionNumberElement.appendChild(CompareUtils.getValueDiff(rowData.isHeading() ? rowData.getModuleSection() : "", oppositeHeadingText).node);
			} else {
				sectionNumberElement.appendChild(document.createTextNode(rowData.getModuleSection()));
			}
			var headingTextElement = dojo.create('div', {"class": "headingText"}, headingNode);
			headingTextElement.innerHTML = rowData.getPrimaryText();
			displayContent.appendChild(headingNode);
			
			// Add the heading tag as an attribite to the contentColumn node so that
			// it can be found for re-display of the column content when we don't have
			// the full rowData available.
			dojo.attr(ptCell, CONTENT_COL_HEADING_TAG_ATTR, headingTag);
		} else {
			ptWrapper = dojo.create('div', {"class":"textContent"}, displayContent);
			ptWrapper.innerHTML = rowData.getPrimaryText();
		}
		return contentWrapper;
	},
		
	_clearConnections: function() {
		// summary:
		//		Clear all connections from display
		dojo.forEach(this._displayedConnections, function(connection){
			connection.removeShape();
			//connection.destroy();
		});
		this._displayedConnections = [];
	},
	
	_showConnections: function() {
		// summary:
		//		Display the connection lines between corresponding blocks
		//		on the left and right grids.
		// returns:
		//		An object indicating the topmost visible objects on left and right sides.
		
		// start with an empty canvas
		this._clearConnections();
		
		// for each block start on the left, find the corresponding one on the right.
		dojo.query(".blockStart, .addedPlaceholder>div", this._leftGrid).forEach(function(leftBlockStartNode){
			var connectionType, connectionVisibility,
				rightNodeId = this._getRightIdFromLeft(leftBlockStartNode.id),
				rightBlockStartNode = dojo.query("#"+rightNodeId, this._rightGrid)[0];
			if (!rightBlockStartNode) {
				// no match - should never get here!
				return;
			}
			if (this._showingMoves && dojo.hasClass(leftBlockStartNode, "moved")) {
				connectionType = MOVED;
			} else if (this._showingRemoves && dojo.hasClass(leftBlockStartNode, "removed")) {
				connectionType = REMOVED;
			} else if (this._showingAdds && dojo.hasClass(rightBlockStartNode, "added")){
				connectionType = ADDED;
			}
			if (connectionType) {
				// the connection is visible
				connectionVisibility = this._drawConnection(leftBlockStartNode, rightBlockStartNode, connectionType);
			}
			
//			// For a MOVED block on the left, connect to the corresponding movedFrom position on the right.
//			if (this._showingMovesFrom && connectionType === MOVED) {
//				rightNodeId = this._getRightMovedFromIdFromLeft(leftBlockStartNode.id);
//				var rightMovedFromPositionNode = dojo.query("#"+rightNodeId, this._rightGrid)[0];
//				if (!rightBlockStartNode) {
//					// no match - should never get here!
//					return;
//				}
//				this._drawConnection(leftBlockStartNode, rightMovedFromPositionNode, MOVED_FROM);
//			}
			
		}, this);
	},
	
	_drawConnection: function(leftIdNode, rightIdNode, connectionType) {
		// summary:
		// 		Draw a connecting line between two points
		// leftIdNode:
		//		Node in the left tree that has an id
		// rightIdNode:
		//		Node in the right tree that has an id.
		// returns:
		//		An object indicating the visibility of both ends of the connection:
		//			{
		//				leftVisible: 	true/false,
		//				rightVisible:	true/false 
		//			}
		var leftConnectionNodeInfo = this.getConnectionNode(leftIdNode);
		var rightConnectionNodeInfo = this.getConnectionNode(rightIdNode);
		var leftEndpointY = leftConnectionNodeInfo.node.offsetTop - leftConnectionNodeInfo.node.offsetParent.scrollTop + leftConnectionNodeInfo.connectionOffset;
		var rightEndpointY = rightConnectionNodeInfo.node.offsetTop - rightConnectionNodeInfo.node.offsetParent.scrollTop + rightConnectionNodeInfo.connectionOffset;
		var lineColor, lineStyle = "Solid";
		
		var endpointInfo = {
			leftVisible: (leftEndpointY >= 0 && leftEndpointY < this._connectionSurfaceHeight),
			leftPosition: leftEndpointY,
			leftOffset: leftConnectionNodeInfo.connectionOffset,
			rightVisible: (rightEndpointY >= 0 && rightEndpointY < this._connectionSurfaceHeight),
			rightPosition: rightEndpointY,
			rightOffset: rightConnectionNodeInfo.connectionOffset
		};
		

		var boldConnector = endpointInfo.leftVisible && endpointInfo.rightVisible;
		switch (connectionType) {
			case MOVED:
				lineColor = boldConnector ? CONNECTION_COLOR_MOVED : CONNECTION_COLOR_MOVED_FADED;
				break;
			case MOVED_FROM:
				lineColor = boldConnector ? CONNECTION_COLOR_MOVED : CONNECTION_COLOR_MOVED_FADED;
				lineStyle = "Dash";
				break;
			case REMOVED:
				lineColor = boldConnector ? CONNECTION_COLOR_REMOVED : CONNECTION_COLOR_REMOVED_FADED;
				break;
			case ADDED:
				lineColor = boldConnector ? CONNECTION_COLOR_ADDED : CONNECTION_COLOR_ADDED_FADED;
				break;
		}
		
		// draw a line connecting these points		
		var curve = this._connectionSurface.createPath();
		curve = curve.moveTo(0, leftEndpointY)
				.curveTo(CONNECTION_CURVE_CONTROL_LEFT, leftEndpointY, CONNECTION_CURVE_CONTROL_RIGHT, rightEndpointY, CONNECTION_SURFACE_WIDTH, rightEndpointY)
				.setStroke({
					color: lineColor,
					width: 1,
					style: lineStyle
				});
		if (!boldConnector) {
			curve.moveToBack();
		}
		this._displayedConnections.push(curve);
		return endpointInfo;
		
	},
		
	_handleRightPaneScroll: function(event) {
		dojo.stopEvent(event);
		this._showConnections();
	},
	
	
	_handleLeftPaneScroll: function(event) {
		dojo.stopEvent(event);
		this._showConnections();
	},
	
	_onGridClick: function(isLeftGrid, event) {
		// summary:
		//		Handles a click within either the left or the right grid which
		//		selects an artifact for which to display a value.
		// isLeftGrid:
		//		true if the click is in the left grid.  false if click is in the right grid.
		// event: standard click event.
		var target = event.target,
			gridNode = isLeftGrid ? this._leftGrid : this._rightGrid,
			contentColumnNode, rowNode;
		
		dojo.stopEvent(event);
		
		// Check for clicks on show/hide links in the row
		if (dojo.hasClass(target, "showChangesLink")){
			// Show the changes inline
			contentColumnNode = DomUtils.findAncestorByClassInclusive(target, "contentColumn", gridNode);
			rowNode = DomUtils.findAncestorByAttr(target, ROW_URI_ATTR, gridNode);
			this._showChanges(isLeftGrid, contentColumnNode, rowNode, target);
			return;
		}
		if (dojo.hasClass(target, "hideChangesLink")){
			contentColumnNode = DomUtils.findAncestorByClassInclusive(target, "contentColumn", gridNode);
			this._hideChanges(contentColumnNode);
			return;
		}
		
		// It's a click somewhere else in the grid.
		rowNode = DomUtils.findAncestorByAttr(target, ROW_URI_ATTR, gridNode);
		if (!rowNode) {
			return;
		}
		var placeholderNode = DomUtils.findAncestorByClassInclusive(target, "addedPlaceholder", rowNode)
				|| DomUtils.findAncestorByClassInclusive(target, "removedPlaceholder", rowNode);
		if (placeholderNode) {
			// click was on placeholder line between rows.  Use the placeholder that is within a row.
			rowNode = placeholderNode;
		}
		this._selectRow(isLeftGrid, rowNode);
		
		// Scroll the corresponding row on the opposite grid into view if it is not already visible.
		// _selectRow has just set this._rightSelection and this._leftSelection, so use them to identify the
		// row to be scrolled into view.
		var oppositeSelection = isLeftGrid ? this._rightSelection : this._leftSelection;
		if (oppositeSelection) {
			var selectedConnectionNodeInfo = this.getConnectionNode(rowNode);
			var oppositeSelectedConnectionNodeInfo = this.getConnectionNode(oppositeSelection.rowNode);
			// Find out if the opposite row is visible  Only force a scroll of the opposite pane if the row being highlighted there is not visible. 
			var scrollNode = isLeftGrid ? this._rightGridScroller : this._leftGridScroller,
				connectionOffset = oppositeSelectedConnectionNodeInfo.connectionOffset - selectedConnectionNodeInfo.connectionOffset;
				
			var isVisible = this.isRowFullyVisible(oppositeSelectedConnectionNodeInfo.node, scrollNode);
			
			if (!isVisible) {
				// Scroll the opposite pane to bring the corresponding row into view.
				this._scrollPane(selectedConnectionNodeInfo.node, oppositeSelectedConnectionNodeInfo.node, scrollNode, connectionOffset);
			}
		}
	},

	_selectInitialRow: function() {
		// summary:
		//		Make an initial row selections so that something is
		//		selected when the structure is first displayed.
		//		Selects top node on the right grid if there is one, otherwise
		//		top node on the left grid.
		
		// If a grid contains any rows, we can guarantee that there is an initial child  which
		// is the invisible root node, so skip that and use the first proper node.
		var rowNode = this._rightGrid.firstChild.nextSibling,
			isLeftGrid = false;
		if (!rowNode && !this._isSingleGrid) {
			rowNode = this._leftGrid.firstChild.nextSibling;
			isLeftGrid = true;
		}
		// rowNode could be undefined, but _selectRow copes with that.
		this._selectRow(isLeftGrid, rowNode, true/*dontTakeFocus*/);
	},
	
	_onGridMouseMove: function(isLeftGrid, event) {
		// summary:
		//		Handle mouse moves over the grid, which cause a highlight to appear on
		//		hovered rows and placeholders.
		var target = event.target,
			gridNode, hoverOverlay, selectionNode;
		if (isLeftGrid) {
			gridNode = this._leftGrid;
			hoverOverlay = this._leftHoverOverlay;
		} else {
			gridNode = this._rightGrid;
			hoverOverlay = this._rightHoverOverlay;
			
		}
		// Search for a selection node, stopping when the first possible one is found.
		selectionNode =    DomUtils.findAncestorByClassInclusive(target, "deepestNest", gridNode)
						|| DomUtils.findAncestorByClassInclusive(target, "innerBox", gridNode)
						|| DomUtils.findAncestorByClassInclusive(target, "addedPlaceholder", gridNode)
						|| DomUtils.findAncestorByClassInclusive(target, "removedPlaceholder", gridNode);
		if (!selectionNode) {
			hoverOverlay.hide();
			return;
		}
		hoverOverlay.move(dojo.marginBox(selectionNode));
		hoverOverlay.show();
	},
	
	_onGridMouseLeave: function(isLeftGrid, event) {
		// summary:
		//		Handler for when the mouse leaves the grid area.  Hides the hover overlay.
		var hoverOverlay = isLeftGrid ? this._leftHoverOverlay : this._rightHoverOverlay;
		hoverOverlay.hide();
	},
	
	_makeLeftNodeId: function(uri) {
		return this.id + "_L_" + uri.substr(uri.lastIndexOf('/')+1);
	},
	
	_makeRightNodeId: function(uri) {
		return this.id + "_R_" + uri.substr(uri.lastIndexOf('/')+1);
	},
	
	_makeRightMovedFromNodeId : function(uri) {
		return this.id + "_RMF_" + uri.substr(uri.lastIndexOf('/')+1);
	},
	
	_getRightIdFromLeft: function(leftNodeId) {
		var leftPrefix = this.id + "_L_";
		if (leftNodeId.indexOf(leftPrefix) === 0) {
			leftNodeId = leftNodeId.substr(leftPrefix.length);
		}
		return this._makeRightNodeId(leftNodeId);
	},
	
	_getLeftIdFromRight: function(rightNodeId) {
		var rightPrefix = this.id + "_R_";
		if (rightNodeId.indexOf(rightPrefix) === 0) {
			rightNodeId = rightNodeId.substr(rightPrefix.length);
		}
		return this._makeLeftNodeId(rightNodeId);
	},
	
	_getRightMovedFromIdFromLeft: function(leftNodeId) {
		var leftPrefix = this.id + "_L_";
		if (leftNodeId.indexOf(leftPrefix) === 0) {
			leftNodeId = leftNodeId.substr(leftPrefix.length);
		}
		return this._makeRightMovedFromNodeId(leftNodeId);
	},
	
	_handleKeyUpDown: function(isUp, event, handlerEntry) {
		// summary:
		//		Handle a keypress that moves focus to the next/previous row.
		// isUp:
		//		true for UP key, false for DOWN key.
		// event:
		//		The key event.
		// handlerEntry:
		//		The keyboard handler entry from which this handler was invoked.
		var isLeftGrid = handlerEntry.isLeftGrid,
			selection = isLeftGrid ? this._leftSelection : this._rightSelection,
			oldFocusRow = selection.rowNode,
			newFocusRow = isUp ? oldFocusRow.previousSibling : oldFocusRow.nextSibling;
		if (!newFocusRow || dojo.attr(newFocusRow, ROW_URI_ATTR) === CompareConstants.STRUCTURE_DIFF_ROOT_URI) {
			// at the bottom or at the fake root node that is at the start of the list on display - nowhere to go
			return;
		}
		this._selectRow(isLeftGrid, newFocusRow);
	},
	
	_handleKeyNextPrevChange: function(isUp, event, handlerEntry) {
		// summary:
		//		Handle a keypress that moves focus to the next/previous "interesting" row.  An interesting row is
		//		one that is at the start of a block or has a change (i.e. "Show changes" button is displayed.
		// isUp:
		//		true for UP key, false for DOWN key.
		// event:
		//		The key event.
		// handlerEntry:
		//		The keyboard handler entry from which this handler was invoked.
		var isLeftGrid = handlerEntry.isLeftGrid,
			grid, currentSelection, newFocusRow;
		if (isLeftGrid) {
			grid = this._leftGrid;
			currentSelection = this._leftSelection;
		} else {
			grid = this._rightGrid;
			currentSelection = this._rightSelection;
		}
		// traverse through the rows, looking for the next "interesting" row
		if (!currentSelection) {
			return;
		}
		newFocusRow = isUp ? currentSelection.rowNode.previousSibling : currentSelection.rowNode.nextSibling;
		while ( newFocusRow && (dojo.attr(newFocusRow, ROW_URI_ATTR) !== CompareConstants.STRUCTURE_DIFF_ROOT_URI) ) {
			if (dojo.hasClass(newFocusRow, "blockStart") || dojo.hasClass(newFocusRow, "valueChanged")) {
				// found an interesting row
				this._selectRow(isLeftGrid, newFocusRow);
				return;
			}
			// try the next row
			newFocusRow = isUp ? newFocusRow.previousSibling : newFocusRow.nextSibling;
		}
		// Reached the top or bottom without finding an interesting row
	},
	
	_handleKeyAction: function(event, handlerEntry) {
		// summary:
		//		Handle a keypress that invokes the default action on a row (show or hide content) 
		// event:
		//		The key event.
		// handlerEntry:
		//		The keyboard handler entry from which this handler was invoked.
		if (handlerEntry.isLeftGrid) {
			// we never have show/hide changes links on the left grid, so nothing to do.
			return;
		}
		var selection = this._rightSelection;
		if (!selection) {
			// no row selected
			return;
		}
		var isLeftGrid = false,
			rowNode = selection.rowNode,
			contentColumnNode = dojo.query(".contentColumn", rowNode)[0],
			showHideChangesLinkNode = dojo.query(".displayContent.show .showChangesLink", contentColumnNode)[0];
		if (showHideChangesLinkNode) {
			// There is a show changes link on display
			this._showChanges(isLeftGrid, contentColumnNode, rowNode, showHideChangesLinkNode);
			return;
		}
		showHideChangesLinkNode = dojo.query(".displayContent.show .hideChangesLink", contentColumnNode)[0];
		if (showHideChangesLinkNode) {
			// There is a hide changes link on display
			this._hideChanges(contentColumnNode);
			return;
		}
	},
	
	_handleKeyLeftRight: function(isLeftKey, event, handlerEntry) {
		// summary:
		//		Handle a keypress that moves focus to the left or right between the grids.  It is also
		//		possible to move between grids using the browser's standard TAB behaviour.
		// isUp:
		//		true for LEFT key, false for RIGHT key.
		// event:
		//		The key event.
		// handlerEntry:
		//		The keyboard handler entry from which this handler was invoked.
		var isLeftGrid = handlerEntry.isLeftGrid;
		if (isLeftGrid === isLeftKey) {    // same as ((isLeftGrid && isLeftKey) || (!isLeftGrid && !isLeftKey))
			// cannot move left from left grid or right from right grid.
			return;
		}
		
		// move to selection on the other grid
		var selection = isLeftGrid ? this._rightSelection : this._leftSelection;
		if (selection) {
			this._selectRow(!isLeftGrid, selection.rowNode);
		}
	}
});
})();
