/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2013, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
 
(function() {
dojo.provide("com.ibm.rdm.web.compare.compareLinksUtils");

dojo.require("com.ibm.rdm.web.compare.CompareConstants");
//CIRC: dojo.require("com.ibm.rdm.web.compare.compareutils");
dojo.require("com.ibm.rdm.web.comparevvc.vvccompareutils");
dojo.require("com.ibm.rdm.web.server.client.utils.LinkUtils");
dojo.require("com.ibm.rdm.web.links.linkutils");
dojo.require("com.ibm.rdm.web.links.LinkTypeUtils");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.server.Headers");
dojo.require("com.ibm.rdm.web.util.deferredutils");
dojo.require("com.ibm.rdm.web.util.vvcutils");

dojo.requireLocalization("com.ibm.rdm.web.compare", "compareutils");

var CompareConstants = com.ibm.rdm.web.compare.CompareConstants,
    ResourceType     = CompareConstants.ResourceType,
	DeferredUtils    = com.ibm.rdm.web.util.deferredutils,
	LinkUtils        = com.ibm.rdm.web.links.linkutils,
	LinkTypeUtils    = com.ibm.rdm.web.links.LinkTypeUtils,
	ResourceManager  = com.ibm.rdm.web.model.resources.ResourceManager,
	RDMClient = com.ibm.rdm.web.server.client.RDMClient,
	Headers = com.ibm.rdm.web.server.Headers,
	ServerLinkUtils  = com.ibm.rdm.web.server.client.utils.LinkUtils,
	VVCCompareUtils  = com.ibm.rdm.web.comparevvc.vvccompareutils,
	VVCUtils         = com.ibm.rdm.web.util.vvcutils,
	messages		 = dojo.i18n.getLocalization("com.ibm.rdm.web.compare", "compareutils"),
	ID_SEPARATOR     = ': ',
	LABEL_SEPARATOR  = ': ',
	LIST_SEPARATOR   = ', ';

// ========= 'PRIVATE' functions ================

function addModuleDeltaEntries(resourceDeltas, linkDeltaEntries, map) {
    
    // Summary:
    //  Given a map of bindingURI => module URI
    //  a list of all resource deltas and a list of all the delta entries
    //  for links, this function adds in an entry for the module into the
    //  resourceDeltas if required. In addition it updates the link delta
    //  entries "deliveryControlledBy" property so that the delivery of 
    //  the module controls delivery of the link. 
    //
    //  See RelatedArtifactsDeliveryManager._manageRelatedLinkMergeStates
    
    var moduleURI, // the containing module for each entry in the map
        idx, 
        binding, // key in the map 
        related, // related artifacts object for the module entry we'll
                 // add to resourceDeltas
        moduleDeltaEntry,
        linkDeltaEntry; // link delta entry that refers to the modulefact
    // find the link delta entry that has as an endpoint, the module in the 
    // artifact.
    for (binding in map) { 
       moduleURI = map[binding];
       linkDeltaEntry = (linkDeltaEntries.filter(function(e) {
               return (e.deliveryControlledBy[0] === binding || e.deliveryControlledBy[1] === binding);
           })[0]);
       
       
       idx = linkDeltaEntry.deliveryControlledBy.indexOf(binding);
       // replace the URI that "controls" delivery of this link, i.e. 
       // we are saying that delivery of the link is controlled by the
       // module.  That is the user must deliver the module and the other
       // end of the link.
       linkDeltaEntry.deliveryControlledBy[idx] = map[binding];
       // Also keep the binding URI that was in there before - this is not used to determine
       // delivery, but is used in module compare to determine that there has been a
       // change of link to a row in the module.  Keep it in an array, in the position that
       // corresponds to the deliverControlledBy entry that we have just replaced.
       linkDeltaEntry.linkedBindings = linkDeltaEntry.linkedBindings || [];
       linkDeltaEntry.linkedBindings[idx] = binding;
       // and put the module into the resource deltas if it is 
       // not already there.
       moduleDeltaEntry = resourceDeltas[moduleURI];
       if (!moduleDeltaEntry) {
           moduleDeltaEntry = resourceDeltas[moduleURI] = {
               concept: map[binding],
               type: ResourceType.ARTIFACT
           };
       }
       
       related = moduleDeltaEntry.related = moduleDeltaEntry.related || {};
       related.links = related.links || [];
       related.links.push(linkDeltaEntry);
       
       VVCCompareUtils.addDeltaType(moduleDeltaEntry,  CompareConstants.DELTA_TYPE_LINKS_CHANGE);
   }

   return resourceDeltas;
}


function addLinkEndpointDeltas(resourceDeltas, parsedLinks) {
	// summary:
	//		Creates or updates entries in resourceDeltas to indicate artifacts that have
	//		had links added or deleted.  Following this call, resourceDeltas will have
	//		an entry with a LINKS_CHANGED deltaType flag for each resource that
	//		is an endpoint of the links in parsedLinks. 
	// resourceDeltas:
	//		A hash of resource deltas.
	// parsedLinks:
	//		An array of parsed parsed links.
	
	var linkDeltaEntries = [],
	    // requiring this gave a circular dependancy
	    CompareUtils = com.ibm.rdm.web.compare.compareutils,
	    urisToCheckIfInModule; // a list of link endpoints that we need to check
	                           // if a binding.
	parsedLinks.forEach(function(link){
		var linkDOM = link.xmlDOM,
		    endpoints = [linkDOM.RDF.Link.object.xmlAttributes.resource, 
                         linkDOM.RDF.Link.subject.xmlAttributes.resource];

		// Link endpoints...
		// Only include link endpoints that are at same location as the link so that
		// we are not gathering external resources such as http://www.ibm.com.
		endpoints.forEach(function(endpointURI) {
		    var deltaEntry,
		        linkDeltaEntry,
                related,
                subjectURI,
                objectURI;
                		
    		if (LinkUtils.getURLInfo(endpointURI).sameLocation) {
    			// resource at one end of the link
    			deltaEntry = resourceDeltas[endpointURI];
    			if (!deltaEntry) {
    				// no existing entry - create a new one
    				deltaEntry = resourceDeltas[endpointURI] = {
    					concept: endpointURI,
    					type: ResourceType.ARTIFACT
    				};
    			} 
    			
    			related = deltaEntry.related = deltaEntry.related || {};
                related.links = related.links || [];
                // find the link delta entry.
                linkDeltaEntry = resourceDeltas[link.linkURI];
                if (linkDeltaEntry) {
                    related.links.push(linkDeltaEntry);
                    // maybe this is not a great name ? 
                    // record the uris of the two artifacts that must be delivered
                    // in order for the link to be delivered.  
                    // see RelatedArtifactsDeliveryManager
					subjectURI = linkDOM.RDF.Link.subject.xmlAttributes.resource;
					objectURI = linkDOM.RDF.Link.object.xmlAttributes.resource;
					// Entries in this array may be modified later if it is found that the link is to
					// a binding in a module, so the link is delivered with the structure.
					linkDeltaEntry.deliveryControlledBy = linkDeltaEntry.deliveryControlledBy || [];
					linkDeltaEntry.deliveryControlledBy.push(endpointURI);
					// Entries in this array are kept as a record of the true endpoints.
					linkDeltaEntry.endpoints = [subjectURI, objectURI];
                }
                if (linkDeltaEntries.indexOf(linkDeltaEntry) === -1) {
                    linkDeltaEntries.push(linkDeltaEntry);
                }
    			VVCCompareUtils.addDeltaType(deltaEntry,  CompareConstants.DELTA_TYPE_LINKS_CHANGE);
    		}
        });
	});
	
	// for each link delta, we need to check whether any "deliveryControlledBy" things are modulefacts
	// if they are, replace them with the module URI.
	
	// this might look scary - but it just transforms the array :
	// [[a, b], [c, d], [e, f]] => [a, b, c, d, e, f]
	
	urisToCheckIfInModule = [].concat.apply([], linkDeltaEntries.map(function(e) { 
	    return e.deliveryControlledBy; 
	}));
	
	// so we end up with an array of uris of link end points, we need to check if 
	// any of those are modules.
	// requiring compareutils causes a circular dependancy error.
	return CompareUtils.getContainingModules(urisToCheckIfInModule)
	   .then(dojo.partial(addModuleDeltaEntries, resourceDeltas, linkDeltaEntries));

}


function parseLinkEndpoints(resourceDeltas, linksFromServer) {
	// summary:
	//		Update the resourceDeltas to add/update the endpoints of some links.
	// resourceDeltas:
	//		A hash of resource deltaEntry objects.
	// linksFromServer:
	//		Array of return objects given to us by the GET call for the link resources.
	// returns:
	//		The supplied resourceDeltas hash with some existing entries updated to include a links change
	//		deltaType flag. New entries may also be added to indicate newly found artifacts that have
	//		links added/removed.
	var i;
	var parsedResults = [];
	var fromParsedObjects,
		toParsedObjects;
	
	// Parse our results
	for (i = 0; i < linksFromServer.length; i+=1) {
		// Expand our parsedResults array, even if we aren't going to put anything in it
		parsedResults.push();
		// Make sure we've got an object to parse
		if (linksFromServer[i].xml.doc) {
			parsedResults[i] = new com.ibm.rdm.web.parsers.xml2json(linksFromServer[i].xml.doc);
		
			// Replace our parser object with the parsed JSON object
			parsedResults[i] = {
			    xmlDOM: parsedResults[i].parse(false),
			    linkURI: linksFromServer[i].linkURI
			};
		}
	}
	
	// There may be duplicates amongst the linked artifacts, but that is OK - addLinkDeltas
	// is tolerant of duplicates.

	// Make sure endpoints of all links appear in resourceDeltas, and have a
	// DELTA_TYPE_LINKS_CHANGE deltaTyps on them.
	return addLinkEndpointDeltas(resourceDeltas, parsedResults);
}	

com.ibm.rdm.web.compare.compareLinksUtils = {

	getLinksDiff: function(values, compact, showAll) {
		// summary:
		//		This function takes a structured object, processes it, and then returns DOM nodes
		//		with markup according to whether a link was added, deleted, or unchanged.
		// values:
		//		An object with links combined from both to and from links, and build
		//		such as this:
		//
		//		{
		//			linkType1URI : 
		//			{
		//				"a2bLinks" : 
		//				{
		//					linkAURI : {"id" : id,  "title" : title,  "type" : type},
		//					linkBURI : {"id" : id2, "title" : title2, "type" : type}
		//				},
		//				"a2bName"  : "Link To",
		//				"b2aLinks" : 
		//				{
		//					linkCURI: {"id" : id3, "title" : title2, "type" : type}
		//				},
		//				"b2aName" : "Link From"
		//			},
		//			linkType2URI : 
		//			{
		//				"a2bLinks" : 
		//				{
		//					linkDURI : {"id" : id,  "title" : title,  "type" : type}
		//				},
		//				"a2bName"  : "Embeds",
		//				"b2aLinks" : 
		//				{
		//					linkCURI: {"id" : id3, "title" : title2, "type" : type}
		//				},
		//				"b2aName" : "Embedded From"
		//			}
		//		}
		//
		// compact:
		//		A boolean whether to show the compact links, or their full names
		// showAll:
		//		A boolean that, if true, shows all links regardless of whether they've changed.
		//		If false, only changed links will show.
		// returns:
		//		A DOM node which contains the links in rows, interleaved in markup, in the this format:
		//		{
		//			node:       DOM node representing the value, with diff markup.
		//			hasChanged: True if the value has changed.
		//			isEmpty:    True if not set in either the fromValue and toValue.  Should
		//					    Only ever be true if hasChanged is not true.
		//			count:      The number of nodes (links) we are returning.
		//		}		
		
		var hasChanged = false;
		var isFirst    = true;
		var totalNodes = 0;
		var nodeCount  = 0;
		var isEmpty;

		// Containers for various DOM nodes which will be used to display the links
		// See createLinkNode() below for usage.
		var rowValues;
		var rowWrapper;
		var cellWrapper;
		var linkNodes;
		var overallWrapper = dojo.create("table");

		// Helper function to determine which markup (strikethrough, highlight, none) 
		// will be added to the particular links in singleTypeLinks, placed in innerWrapper
		var addMarkup = function(singleTypeLinks, innerWrapper) {
			var valueNode,
				elementName,
				hoverText,
				first = true,
				linkInfo;

			var i = 0;

			for (var resource in singleTypeLinks) {
				linkInfo = singleTypeLinks[resource];
				
				// We can still show a both, but only do it if
				// we're showing all, or if we have an ADD or DELETE
				if (showAll || linkInfo.type !== "BOTH") {
					if (!first) {
						innerWrapper.appendChild(document.createTextNode(LIST_SEPARATOR));
					}
					first = false;
					switch(linkInfo.type) {
						case "FROM":
							if (linkInfo.notInDeltas) {
								// do not show elements that are only in the target if
								// we are doing a delivery.
								// (they will only have this notInDeltas flag if doing a deliver)
								elementName = null;
								break;
							} else {
								elementName = 'del';
								hasChanged = true;
							}
							break;
						case "TO":
							if (linkInfo.notInDeltas) {
								// do not show elements that are only in the source and are not
								// being delivered.
								// (they will only have this notInDeltas flag if doing a deliver)
								elementName = null;
								break;
							}
							elementName = 'ins';
							hasChanged = true;
							break;
						default:		// this is really case: "BOTH"
							elementName = 'span';
							break;
					}
					
					if (elementName) {
						valueNode = dojo.create(elementName, null, innerWrapper);
						if (!linkInfo.id) {
							hoverText = messages.linkNotInConfig;
						} else {
							hoverText = linkInfo.id + ID_SEPARATOR + linkInfo.title;
						}
						dojo.setAttr(valueNode, "title", hoverText);
						valueNode.appendChild(document.createTextNode(compact && linkInfo.id ? linkInfo.id : hoverText));
						++i;
					}
				}
			}
			
			return i;
		};
		
		var createLinkNode = function(rowValue, rowName) {
			// Create a new row if the last one created was in the second
			// column, or we're using the compact (module) view which
			// does not have columns
			if (compact || isFirst || !rowWrapper) {
				rowWrapper = dojo.create("tr", null, overallWrapper);
			}
			
			linkNodes = dojo.create("td", {'class':'valueCol'});

			// Don't place our markup in the row yet, it could be empty
			nodeCount = addMarkup(rowValue, linkNodes);
			
			if (nodeCount > 0) {
				dojo.create("td", {'class':'j-label nameCol'}, rowWrapper)
					.appendChild(document.createTextNode(rowName + " (" + nodeCount + ")" + LABEL_SEPARATOR));
					
				// Put it in the DOM
				rowWrapper.appendChild(linkNodes);
				
				// Update our row counters
				isEmpty = false;
				isFirst = !isFirst;
				totalNodes += nodeCount;
			}
		};
		
		isEmpty = true;
			
		for (var linkRow in values) {
			// Append both directions of link
			rowValues   = values[linkRow];

			// First check and create the "subject to object" links of this type
			if (rowValues.a2bLinks) {
				createLinkNode(rowValues.a2bLinks, rowValues.a2bName);
			}
			
			// Then check and create the "object to subject" links of this type
			if (rowValues.b2aLinks) {
				createLinkNode(rowValues.b2aLinks, rowValues.b2aName);
			}
		}
		
		if (totalNodes === 0) {
			isEmpty = true;
		}
		
		return {
			node: overallWrapper,
			hasChanged: hasChanged,
			isEmpty: isEmpty,
			count: totalNodes
		};
	},
	
	_getLinks: function(displayType, targetConfiguration, sourceConfiguration, resultCollector) {
		// summary:
		//		Gets the links from the server and calculates their differences locally.
		// displayType:
		//		Indicates of the type of change under consideration.
		// targetConfiguration, sourceConfiguration:
		//		Configurations to be compared.  May be undefined for non-VVC comparison.
		// resultCollector:
		//		Contains the resources to be compared, and the associated type and attribute definitions.
		// returns:
		//		A promise that resolves with resultCollector once the differences have been loaded and added
		//		to it, or the supplied resultCollector if we have no links to show.  
		
		var targetResourceURI;
		var sourceResourceURI;
		var fromAddress;
		var toAddress;
		var linkPromise;
		var promises = [];
		
		resultCollector.targetResourceLinks = {};
		resultCollector.sourceResourceLinks   = {};
		
		// Get links attached to the targetResource.
		if (resultCollector.targetResource) {

			// Set our From addresses, and make it configuration aware
			targetResourceURI = resultCollector.targetResource.getResourceURI();
			fromAddress = VVCUtils.getConfigurationAwareAddress(targetResourceURI, targetConfiguration);
			
			// TODO: evaluate whether skipCache is needed
			linkPromise = ServerLinkUtils.getLinksForResource({ 
				sourceAndTargetURI: targetResourceURI,
				skipCache: true,
				vvcConfiguration: targetConfiguration
			})
			.then(dojo.hitch(this, 
							this._createLinksObject, 
							targetResourceURI, 
							true, 
							targetConfiguration,
							resultCollector.targetResource.getComponentURI()));
			
			promises.push(linkPromise);
		}
		
		// Get links for the sourceResource
		if (resultCollector.sourceResource && (resultCollector.sourceResource !== resultCollector.targetResource)) {
		
			// Set our From addresses, and make it configuration aware
			sourceResourceURI = resultCollector.sourceResource.getResourceURI();
			toAddress = VVCUtils.getConfigurationAwareAddress(sourceResourceURI, sourceConfiguration);
			
			// Fire two requests - one for the links which this is the "head" of
			linkPromise = ServerLinkUtils.getLinksForResource({ 
				sourceAndTargetURI: sourceResourceURI,
				skipCache: true,
				vvcConfiguration: sourceConfiguration
			})
			.then(dojo.hitch(this, 
							this._createLinksObject, 
							sourceResourceURI, 
							false, 
							sourceConfiguration,
							resultCollector.sourceResource.getComponentURI()));

			promises.push(linkPromise);
		}
		
		// Set the primary resource's links to be either the sourceResource or targetResource's links
		switch (displayType) {
			case CompareConstants.DISPLAY_AS_ADDED:
				resultCollector.primaryResourceLinks = resultCollector.sourceResourceLinks;
				break;
			case CompareConstants.DISPLAY_AS_DELETED:
				resultCollector.primaryResourceLinks = resultCollector.targetResourceLinks;
				break;
			case CompareConstants.DISPLAY_AS_CHANGED:
				// For consistency, just default it to the sourceResource's links
				resultCollector.primaryResourceLinks = resultCollector.sourceResourceLinks;
				break;
			default:
				break;
		}
		
		// Return the promise that will resolve when all our promises have
		return DeferredUtils.all(promises)
			.then(function(promiseResults) {
			
				dojo.forEach(promiseResults, function(result) {
					// Put our final formatted link objects into their proper
					// places in the resultCollector
					if (result) {	
						if (result.targetResource) {
							resultCollector.targetResourceLinks = result.linkObject;
						} else {
							resultCollector.sourceResourceLinks = result.linkObject;
						}
					}
				});
				
				return resultCollector;
			});
	},
	
	_createLinksObject: function(resourceURI, targetResource, configuration, componentURI, links) {
		// summary:
		//		Obtains the links and their types for the provided resource, and 
		//		processes them into a returned object
		// resourceURI:
		//		Uri of the resource we are to obtain the links for
		// targetResource:
		//		Boolean, true if the resource we are using is the "from" resource.
		//		False otherwise.
		// configuration:
		//		URI of the configuration we are working in
		// componentURI:
		//		URI of the project in which the resourceURI lives
		// links:
		//		Array of link resource uris which we are to obtain link objects for
		// returns:
		//		An object with two elements, a "linkObject" and a boolean "targetResource".
		//		
		//		linkObject: An object formatted as such
		//		{
		//			linkType1URI : 
		//			{
		//				"a2bLinks" : 
		//				{
		//					linkAURI : {"id" : id,  "title" : title,  "type" : type},
		//					linkBURI : {"id" : id2, "title" : title2, "type" : type}
		//				},
		//				"a2bName"  : "Link To",
		//				"b2aLinks" : 
		//				{
		//					linkCURI: {"id" : id3, "title" : title2, "type" : type}
		//				},
		//				"b2aName" : "Link From"
		//			},
		//			linkType2URI : 
		//			{
		//				"a2bLinks" : 
		//				{
		//					linkDURI : {"id" : id,  "title" : title,  "type" : type}
		//				},
		//				"a2bName"  : "Embeds",
		//				"b2aLinks" : 
		//				{
		//					linkCURI: {"id" : id3, "title" : title2, "type" : type}
		//				},
		//				"b2aName" : "Embedded From"
		//			}
		//		}
		
		var linkTypesArray    = [];
		var resourcesToGet    = [];
		var resourceName;
		var nameInfo;
		var linkName;
		var formattedLinks    = {};
		var linkNames         = [];
		var linkTypeURIMap    = {};
		var promises          = [];
		var returnObject      = {};
		
		// necessary to use a fakeDeferred like this, because of how ResourceManager.getMultiple doesn't
		// wait for all its calls to resolve before resolving its promise.
		var deferred = new dojo.Deferred();
		
		// Put each link into the resourceLinks object
		dojo.forEach(links.resourceInfos, function(link) {
			linkTypesArray.push(link._linkType);
			linkTypeURIMap[link._linkType] = {};
		});
		
		var addTitles = function(resources) {
			
			// Go through each resource and put it in its proper place
			dojo.forEach(resources, function(resource) {
				if (!resource.getID) {
					// There's been some failure reading this resource - it is an object
					// containing an error message.
					return;
				}
			
				// Grab the names and IDs from the resource
				nameInfo = {"id" : resource.getID(), "title" : resource.getTitle()};
			
				// Find where the resource information belongs
				for (var linkType in linkTypeURIMap) {
					// And put them into their correct place in our linkTypeURI object
					if (linkTypeURIMap[linkType]["a2bLinks"] && 
						linkTypeURIMap[linkType]["a2bLinks"][resource.resourceURI]) {
						// groan...
						nameInfo.linkConcept = linkTypeURIMap[linkType]["a2bLinks"][resource.resourceURI]["linkConcept"];
						linkTypeURIMap[linkType]["a2bLinks"][resource.resourceURI] = nameInfo;
					}
					
					if (linkTypeURIMap[linkType]["b2aLinks"] &&
						linkTypeURIMap[linkType]["b2aLinks"][resource.resourceURI]) {
						nameInfo.linkConcept = linkTypeURIMap[linkType]["b2aLinks"][resource.resourceURI]["linkConcept"];
						linkTypeURIMap[linkType]["b2aLinks"][resource.resourceURI] = nameInfo;
					}
				}
			});

			deferred.callback({"linkObject" : linkTypeURIMap, "targetResource" : targetResource});
		};
		
		promises.push(
				// Load the link types from both configuration incase one has been
				// removed from source. (WI 96202)
				DeferredUtils.all([LinkTypeUtils.getLinkTypes(linkTypesArray),
								   LinkTypeUtils.getLinkTypes(linkTypesArray, configuration)])
				.then(function(results) { return results[0].concat(results[1]); })
				.then(dojo.hitch(this, function(types) {
			
					// Process the types in order to determine the link names, and their direction
					var linkName;
					var tempType;
					
					// Process the types in order to determine the link names, and their direction
					dojo.forEach(links.resourceInfos, function(link) {
		
						// Associate the type object relating to the link we're dealing with
						tempType = dojo.filter(types, function (type) {
							return type.resourceURI === link._linkType;
						});
						// Since that filter returned an array, extract the object
						tempType = tempType[0];
						
						// This is the subject, so add it into the "a2bLinks" area and populate "a2bName"
						if (tempType && resourceURI === link._subject) {
							// Create our object if we don't have it yet, but this "a2bLinks"
							// object is only being created because we plan to put a resource
							// into it.  Otherwise we don't want this field.  Same applies to
							// "b2aLinks"
							if (!linkTypeURIMap[tempType.resourceURI]["a2bLinks"]) {
								linkTypeURIMap[tempType.resourceURI]["a2bLinks"] = {};
							}
							
							// We want the resource which this link is pointed TO, not the resource
							// we already have.
							resourceName = link._object;
							
							linkTypeURIMap[tempType.resourceURI]["a2bLinks"][resourceName] = {};
							linkTypeURIMap[tempType.resourceURI]["a2bName"] = tempType.subjectToObjectLabel;
							linkTypeURIMap[tempType.resourceURI]["a2bLinks"][resourceName]["linkConcept"] = link.getResourceURI();
						} else if (tempType) {
						// Not the subject, so it's the object.  Otherwise do the same things.
						
							if (!linkTypeURIMap[tempType.resourceURI]["b2aLinks"]) {
								linkTypeURIMap[tempType.resourceURI]["b2aLinks"] = {};
							}
							
							resourceName = link._subject;
							
							linkTypeURIMap[tempType.resourceURI]["b2aLinks"][resourceName] = {};
							linkTypeURIMap[tempType.resourceURI]["b2aName"] = tempType.objectToSubjectLabel;
							linkTypeURIMap[tempType.resourceURI]["b2aLinks"][resourceName]["linkConcept"] = link.getResourceURI();
						}
						
						// RM internal link, so add it to the queue of resources to be processed
						if (LinkUtils.getURLInfo(resourceName).sameLocation) {
							// Build our object to send to the ResourceManager
							resourcesToGet.push(
								{
								"resourceURI" : resourceName,
								"vvcConfiguration" : configuration
								}
							);
						}  else if (tempType) {
						// External link, so put it into our big mapped object, but don't send it off to the server.
							if (resourceName === link._object) {
								linkTypeURIMap[tempType.resourceURI]["a2bLinks"][resourceName] = ({"title" : resourceName, "id" : link._title});
							} else {
								linkTypeURIMap[tempType.resourceURI]["b2aLinks"][resourceName] = ({"title" : resourceName, "id" : link._title});
							}
						}
					});
		
					// Get our resources if we have any to get
					if(resourcesToGet.length) {
						ResourceManager.getMultiple(resourcesToGet, addTitles);
					} else {
					// Otherwise, just fire our callback with the URI map.  
					//		It may or may not be empty, depending on whether there are
					// 		external resources in it or not.
						deferred.callback({"linkObject" : linkTypeURIMap, "targetResource" : targetResource});
					}
				})
			));
		
		promises.push(deferred);
		
		return DeferredUtils.all(promises).then(function(results) {
			
			// One of our results will be empty from the function above, 
			// and the other will have our complete return object from
			// the callback addTitles() from ResourceManager.getMultiple.
			//
			// We want that one.
			for (var i = 0; i < results.length; i+=1) {

				if (results[i] && results[i].linkObject) {
					return results[i];
				}
			}
		});
	},
	
	_addLinksDifference: function(displayType, fromLinks, toLinks) {
		// summary:
		//		Takes two arrays of links objects, and reorganizes them into the format
		//		described below depending on the displayType.
		// fromLinks:
		//		Object with each key being the link type uri, and the data in each key being
		//		an array of link objects in the "from" configuration 
		// toLinks:
		//		Object with each key being the link type uri, and the data in each key being
		//		an array of link objects in the "to" configuration 
		// returns:
		//		An array of objects with the following format, each corresponding
		//		to a row to be displayed for the links type
		//		
		//		{
		//			linkType1URI : 
		//			{
		//				"a2bLinks" : 
		//				{
		//					linkAURI : {"id" : id,  "title" : title,  "type" : type},
		//					linkBURI : {"id" : id2, "title" : title2, "type" : type}
		//				},
		//				"a2bName"  : "Link To",
		//				"b2aLinks" : 
		//				{
		//					linkCURI: {"id" : id3, "title" : title2, "type" : type}
		//				},
		//				"b2aName" : "Link From"
		//			}
		//		}
	
		var masterLinksList = {};
		var wrapperNode;
		var tempLinkTitle;
		var tempLinkId;
		
		switch (displayType) {
			case CompareConstants.DISPLAY_AS_ADDED:
				// Only a "to" for an add delta
				this._setupMasterList(masterLinksList, toLinks, "BOTH");
				break;
			case CompareConstants.DISPLAY_AS_DELETED:
				// Only a "from" for a delete delta
				this._setupMasterList(masterLinksList, fromLinks, "BOTH");
				break;
			case CompareConstants.DISPLAY_AS_CHANGED:
			// Now that we have the labels, we can iterate through them
			// to get a diff for each.  But first we must find the items which
			// have changed, and which have not.

				// Initially, copy all the values from our fromLinks into the masterLinksList
				this._setupMasterList(masterLinksList, fromLinks, "FROM");
				
				// Now copy the new values over, changing our masterLinksList values to "BOTH" if
				// we find an existing value in the same row.  Otherwise the new values will take on
				// the "TO" value;
				for (var linksRow in toLinks) {
					var toLinksRowObject = toLinks[linksRow];
					var masterRowLinksObject;
					
					// This type isn't in the master list yet!  Calling setupChangeList
					// later on will make sure everything is copied over correctly, including
					// a "TO" label
					if (!masterLinksList[linksRow]) {	
						masterLinksList[linksRow] = {};
					}
					
					masterRowLinksObject = masterLinksList[linksRow];

					// Merge the new data into the existing "from" values
					if (toLinksRowObject.a2bLinks) {
					
						// Give these fields proper instantiation
						if (!masterRowLinksObject.a2bLinks) {
							masterRowLinksObject.a2bLinks = {};
						}
					
						this._setupChangeList(	masterRowLinksObject.a2bLinks,
												toLinksRowObject.a2bLinks);
												
						// Give our row (link type) a name, but only if we've got links
						// we're showing.  It's possible that all links are "BOTH" in a row
						// and hence removed from being shown in this view.
						if (!VVCCompareUtils.isEmptyObject(masterRowLinksObject.a2bLinks)) {
							masterRowLinksObject.a2bName = toLinksRowObject.a2bName;
						}
						else {
							delete masterRowLinksObject.a2bLinks;
						}
					}
			
					if (toLinksRowObject.b2aLinks) {	
					
						// Give these fields proper instantiation
						if (!masterRowLinksObject.b2aLinks) {
							masterRowLinksObject.b2aLinks = {};
						}
						
						this._setupChangeList(	masterRowLinksObject.b2aLinks,
												toLinksRowObject.b2aLinks);
												
						// Give our row (link type) a name
						if (!VVCCompareUtils.isEmptyObject(masterRowLinksObject.b2aLinks)) {
							masterRowLinksObject.b2aName = toLinksRowObject.b2aName;
						}
						else {
							delete masterRowLinksObject.b2aLinks;
						}
					}
				}
				
				break;
			default:
				break;
		}
		
		// Calculate the diff DOM and return
		return masterLinksList;
	},
	
	_setupChangeList: function(masterRowLinks, toRowLinks) {
		// summary:
		//		Copies resources from toRowLinks to masterRowLinks and assigns them
		//		a type.
		// masterRowLinks:
		//		A partial Links object (see _addLinksDifference for layout) for our master list,
		//		at attribute for a particular row corresponding to a Link Type and direction.
		// masterLinksName:
		//		The attribute for the "name" field in the row corresponding to masterRowLinks.
		// toRowLinks:
		//		A partial Links object for our "to" object, which we will be copying links from
		//		to place into masterRowLinks.
		// toLinksName:
		//		The attribute for the "name" field in the row corresponding to toRowLinks
		// returns:
		//		No return, but the resources in toRowLinks are copied over to masterRowLinks
		//		and their "type" attributes are modified whether they're in both toRowLinks
		//		and masterRowLinks, or only in toRowLinks.

		for (var resource in toRowLinks) {
			// Found a match
			if (masterRowLinks && masterRowLinks[resource]) {
				// We're in a changed artifact and the resource is the same,
				// so don't show it in the results
				masterRowLinks[resource].type = "BOTH";
			}
			else {
			// Value, only exists in "TO" resource
				// Careful, it might be empty
				if (!masterRowLinks) {
					masterRowLinks = {};
				}
				
				// Deep copy the resource object
				masterRowLinks[resource] = dojo.clone(toRowLinks[resource]);
				
				// Add in the type
				masterRowLinks[resource].type = "TO";
			}
		}
	},
	
	_setupMasterList: function(masterLinksList, originalLinks, initType) {
		// summary: 
		//		Populates all fields of masterLinksList that are in
		//		originalLinks with the data from originalLinks.
		// masterLinksList:
		//		A Links object (see _addLinksDifference for layout) for our master
		//		list which we are copying information to
		// originalLinks:
		//		A Links object which we are copying information from
		// initType:
		//		A string to describe all resources copied into masterLinksList
		// returns:
		//		No return, but masterLinksList is populated with information from
		//		originalLinks, and initType is copied into its resources
		
		var resource;
		var masterRow;
		var masterA2BLinks;
		var masterB2ALinks;
		
		// Modify the resources for each link type in originalLinks
		for (var linkType in originalLinks) {
			masterLinksList[linkType] = dojo.clone(originalLinks[linkType]);
			masterRow = masterLinksList[linkType];
			
			// Grab the "a2b" direction links
			if (masterRow.a2bLinks) {
				
				masterA2BLinks = masterRow.a2bLinks;
				// Add the initType type to all the resources in our "a2b" links
				for (resource in masterA2BLinks) {
						masterA2BLinks[resource].type = initType;
				}
			}
		
			// Grab the "a2b" direction links
			if (masterRow.b2aLinks) {
				
				masterB2ALinks = masterRow.b2aLinks;
				// Add the initType type to all the resources in our "a2b" links
				for (resource in masterB2ALinks) {
					masterB2ALinks[resource].type = initType;
				}
			
			}
		}
	},
	
	
	findLinkedArtifacts: function(targetConfigurationURI, sourceConfigurationURI, resourceDeltas) {
		// summary:
		//		Add linked artifacts into a set of resourceDeltas.  For any links in resourceDeltas the
		//		endpoints of the links are found and if necessary, added to resourceDeltas.  The links
		//		are removed and the endpoints (whether existing or newly added) are marked as having a
		//		DELTA_TYPE_LINKS_CHANGED deltaType flag.
		// resourceDeltas:
		//		A hash of resourceDelta entries - keyed by conceptURI.
		// returns:
		//		A promise that resolves with the updated resourceDeltas.
		
		var sourceLinkURIs   = [];
		var targetLinkURIs     = [];
		var sourceLinkGetArgs    = [];
		var targetLinkGetArgs    = [];
		var promiseGetList = [];
		var configHeader   = Headers.VVC_CONFIGURATION;
		var concept, delta;
		
		for (concept in resourceDeltas) {
			if (concept.lastIndexOf(com.ibm.rdm.web.FRONTING_SERVICE_LINKS, 0) === 0) {  // more efficient way to test for match at start than string1.indexOf(string2)===0
				// it is a link - see if it has changed
				delta = resourceDeltas[concept];
				if (VVCCompareUtils.hasDeltaType(delta, [CompareConstants.DELTA_TYPE_TARGET_ADD, 
														CompareConstants.DELTA_TYPE_SOURCE_DELETE, 
														CompareConstants.DELTA_TYPE_SOURCE_CHANGE])) {
					targetLinkURIs.push(delta.concept);
				} else if (VVCCompareUtils.hasDeltaType(delta, [CompareConstants.DELTA_TYPE_SOURCE_ADD, 
																CompareConstants.DELTA_TYPE_TARGET_DELETE, 
																CompareConstants.DELTA_TYPE_TARGET_CHANGE])) {
					sourceLinkURIs.push(delta.concept);
				}
				
			}
		}
		
		// Build our request object to match the following format
		//
		// getLinkArgs 
		// { 
		// 		uri: linkURI,
		// 		xhrArgs.headers["vvc.configuration"]: vvc_configuration_uri
		// }
		dojo.forEach(sourceLinkURIs, function(link) {
			// Create our arguments header
			var requestArgs = {
				uri: link,
				xhrArgs: {
					headers: {}
				}
			};
			requestArgs.xhrArgs.headers[configHeader] = sourceConfigurationURI;
			
			sourceLinkGetArgs.push(requestArgs);
		}, this);
		
		dojo.forEach(targetLinkURIs, function(link) {
			// Create our arguments header 
			var requestArgs = {
				uri: link,
				xhrArgs: {
					headers: {}
				}
			};
			requestArgs.xhrArgs.headers[configHeader] = targetConfigurationURI;
			
			targetLinkGetArgs.push(requestArgs);
		}, this);
		
		// Make the requests to get our link information - the direction and its linked artifacts
		
		// Get links that are in the source configuration and
		// parse them to include their endpoints in resourceDeltas
		if (sourceLinkGetArgs.length) {
			promiseGetList.push(RDMClient.doMultiGet(sourceLinkGetArgs)
				.then(dojo.hitch(this, function(result) {
					// Get result into the right shape for calling parseLinkEndpoints
					return result.map(function(entry, index){
						return{
							xml: entry,
							linkURI: sourceLinkGetArgs[index].uri
						};
					});
			})));
		}
		// Get links that are in the target configuration and
		// parse them to include their endpoints in resourceDeltas
		if (targetLinkGetArgs.length) {
			promiseGetList.push(RDMClient.doMultiGet(targetLinkGetArgs)
				.then(dojo.hitch(this, function(result) {
					// Get result into the right shape for calling parseLinkEndpoints
					return result.map(function(entry, index){
						return{
							xml: entry,
							linkURI: targetLinkGetArgs[index].uri
						};
					});
			})));
		}
		
		// we now have promiseGetList containing 0, 1 or 2 promises, depending on whether we had links to
		// fetch from source and target configuration.
		if (!promiseGetList.length) {
			// There were no links to fetch.  Return resourceDeltas directly, which will be passed down the
			// callback chain.
			return resourceDeltas;
		}
		// Wait for the promises to complete.  Once they are complete they have fetched the
		// links.  Make up a single array that contains all the fetched links, and parse them
		// to add entries for their endpoints into resourceDeltas if they are not already there. 
		return DeferredUtils.all(promiseGetList)
			.then(function(linkPromiseResults){
				var allLinks = linkPromiseResults[0];
				if (linkPromiseResults.length > 1) {
					allLinks = allLinks.concat(linkPromiseResults[1]);
				}
				return parseLinkEndpoints(resourceDeltas, allLinks);
			});
	},
	
	
	mixinResourceDeltaInfoWithLinksDiff: function(linksDiff, resourceDeltas) {
		// Summary:
		//   Given a 'linksDiff' (the output of the above getLinksDiff function
		//   and a set of resource deltas, notate objects which are different between
		//   the two configurations, but are not in the given resource deltas.
		//
		//   This function is required because the links diff works by comparing the links
		//   on the "head" version of an artifact in the source, with the head version in
		//   the target, thus it includes links which are in one configuration and not the other
		//   but that may not be getting delivered.  This function is used to note which
		//   link diffs are not in the set of resource deltas, so although they differ 
		//   between the two configurations, they are not going to get delivered.
		Object.getOwnPropertyNames(linksDiff)
		   .map(function(type) { return linksDiff[type]; }, this)
		   .forEach(function(linkDiff) {
		         [linkDiff.a2bLinks, linkDiff.b2aLinks].forEach(function(diffObj) {
		         	if (diffObj) {
		            	Object.getOwnPropertyNames(diffObj)
		            	  .forEach(function(resourceUri) { 
		            	  	  // if this concept is NOT in resourceDeltas, remove the link from a2blinks...
		            	  	  var resource = diffObj[resourceUri];
		            	  	  if (!resourceDeltas[resource.linkConcept]) {
		            	  	       diffObj[resourceUri].notInDeltas = true;
		            	  	  } 
		            	  });
		            	}
		         });
		   });
	}
	
};

})();
