/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2016. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("com.ibm.rdm.web.model.attributes.ui.PopupAttributeInputWidget");

dojo.require("dojo.i18n");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.IntAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.FloatAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.StringAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.BooleanAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.DateAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.DateTimeAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.DurationAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.EnumFilteredListAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.EnumListAttributeWidget");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.EnumListMultiValueAttributeWidget2");
// dojo.require("com.ibm.rdm.web.model.attributes.ui.WorkflowFilteredListAttributeWidget");
dojo.require("com.ibm.rdm.web.model.attributes.ui.EditorPopup");
dojo.require("com.ibm.rdm.web.model.attributes.ui.EditorForm");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.common.Constants");
dojo.require("com.ibm.rdm.web.util.domutils");

dojo.requireLocalization("com.ibm.rdm.web.model.attributes.ui", "AttributeInputWidget");


var EditorForm = com.ibm.rdm.web.model.attributes.ui.EditorForm;
var EditorPopup = com.ibm.rdm.web.model.attributes.ui.EditorPopup;
var RRMTypes = com.ibm.rdm.web.model.types.RRMTypes;
var domutils = com.ibm.rdm.web.util.domutils;
var isPopupEditorCurrentlyOpened = false;


dojo.declare("com.ibm.rdm.web.model.attributes.ui.PopupAttributeInputWidget",
	[dijit._Widget, dijit._Templated],
{
	// summary:
	//		Show a popup editor for editing a particular attribute value within the grid.
	// description:
	//		This widget is responsible for displaying a popup window with an attribute
	//		editor within it and it's used whenever you edit a cell (other than Contents/-
	//		PrimaryText or Title) within the grid (eventually the right hand sidebar's 
	//		Overview sidebar section will ALSO use this class to enable a popup based
	//		approach to editing an opened artifact's attributes).
	//		
	//		This widget creates a "read-only" instance of an AttributeInputWidget (e.g.
	//		StringAttributeWidget) and places it just before this widget's domNode (i.e.
	//		within the grid) and then, on demand, creates a new "write" instance of the
	//		same AttributeInputWidget (e.g. StringAttributeWidget) to be housed within
	//		a popup editor form whenever the user elects to edit the cell associated with
	//		this widget.  Obviously the end result of performing an edit is that a new
	//		value is set within the read-only version and then that read-only version is
	//		asked to update its display to show this new value appropriately.


	templateString: '<span data-dojo-attach-point="_rootNode" class="com-ibm-rdm-web-model-attributes-ui-AttributeWidget"></span>',
	widgetsInTemplate: false,

	value: null,
	required: null,
	dataType: null,
	componentURI: null,
	processAreaURI: null,
	enumEntries: null,
	attributeTypeURI: null,
	focusNode: null,
	minValue: null,
	maxValue: null,
	width: null,
	isMultiValued: false,
	isSystemDefinedAttribute: false,
	workflowViewType: null,
	artifactTypeURI: null,

	_messages: null,
	_rootNode: null,
	_inputWidgetClass: null,
	_inputWidget: null,
	_readOnlyInputWidget: null,
	_editorPopup: null,
	_keypressHandle: null,
	_editorForm: null,
	_doCloseOnModuleScroll: null,


	///////////////////////////////////////////////////////////////////////////
	// Public API

	postCreate: function() {
		var inputWidgetClass = null;
		if (this.enumEntries) {
			if (this.workflowViewType) {
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.WorkflowFilteredListAttributeWidget;
			} else {
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.EnumFilteredListAttributeWidget;
			}
		} else {
			switch(this.dataType) {
			case RRMTypes.getWellKnownAttributeTypes()["int"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.IntAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["float"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.FloatAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["string"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.StringAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["boolean"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.BooleanAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["date"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.DateAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["time"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.TimeAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["dateTime"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.DateTimeAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["duration"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.DurationAttributeWidget;
				break;
			case RRMTypes.getWellKnownAttributeTypes()["user"]:
				inputWidgetClass = com.ibm.rdm.web.model.attributes.ui.EnumFilteredListAttributeWidget;
				break;
			}
		}
		this._readOnlyInputWidget = new inputWidgetClass(
			this._getAttributeInputWidgetConstructorArgs(true, false)
		);

		this._inputWidgetClass = inputWidgetClass;
		dojo.place(this._readOnlyInputWidget.domNode, this.domNode, "before");
	},

	uninitialize: function() {
		if (this._keypressHandle) {
			dojo.disconnect(this._keypressHandle);
		}
		if (this._readOnlyInputWidget) {
			this._readOnlyInputWidget.destroyRecursive();
			this._readOnlyInputWidget = null;
		}
		this._destroyWidgetsUsedInPopups();
		this.inherited(arguments);
	},

	postMixInProperties: function() {
		this._messages = dojo.i18n.getLocalization("com.ibm.rdm.web.model.attributes.ui", "AttributeInputWidget");
	},

	onChange: function(value) {
		// summary:
		//		This callback is invoked whenever a new "legal value" has been
		//		entered in the form that this widget manages (e.g. perhaps a
		//		Priority based value has changed from high to low).
		// tags:
		//		Callback
	},

	onOpen: function() {
		// summary:
		//		This callback is invoked whenever the EditorPopup that this 
		//		widget manages has been opened.
		// tags:
		//		Callback
	},

	onClose: function(doNeedToTakeFocus) {
		// summary:
		//		This callback is invoked whenever the EditorPopup that this
		//		widget manages is about to close.
		// doNeedToTakeFocus:
		//		Set to true iff the user last had focus somewhere within the EditorForm
		//		associated with the hover popup that this widget shows (meaning the caller
		//		now needs to reset focus to somewhere reasonable).
		// tags:
		//		Callback
	},

	onSave: function() {
		// summary:
		//		This callback is invoked whenever the user enters Ctrl-s
		//		(or Command-s) from within the form that this widget manages.
		// tags:
		//		Callback
	},

	refreshInputWidgets: function() {
		// try to re-add sub-domNodes that may have gotten removed during a row refresh
		if (this._rootNode && this._readOnlyInputWidget && this._readOnlyInputWidget.domNode) {
			this._rootNode.appendChild(this._readOnlyInputWidget.domNode);
			if (this._readOnlyInputWidget._inputWidget && this._readOnlyInputWidget._inputWidget.domNode) {
				this._readOnlyInputWidget.domNode.appendChild(this._readOnlyInputWidget._inputWidget.domNode);
			}
		}
	},
	
	getSubWidget: function(){
		return this._readOnlyInputWidget.subWidget;
	},
	
	getFocusNode: function() {
		return this._readOnlyInputWidget.focusNode;
	},

	isValid: function(){
		return (this._inputWidget && this._inputWidget.isValid) ? this._inputWidget.isValid() : true;
	},

	getValue: function() {
		if (this._inputWidget) {
			return this._inputWidget.value;
		} else {
			return this._readOnlyInputWidget.value;
		}
	},
	
	focus: function(event) {
		if (this._editorPopup) {
			this._inputWidget.focus();
		}
		else {
			this.popupEditableNode.focus();
		}
	},

	open: function(/*String?*/title) {
		this._openEditorPopup(title);
		this.onOpen();
		this._inputWidget.focus();
		
		// Select the text in the text input control
		if (this._inputWidget.focusNode && 
			(this._inputWidget.focusNode.type === "text" || this._inputWidget.focusNode.type === "textarea")) {
			this._inputWidget.focusNode.select();
		}
	},

	render: function() {
		this.inherited(arguments);
		dojo.place(this._readOnlyInputWidget.domNode, this.domNode, "before");
	},

	///////////////////////////////////////////////////////////////////////////
	// Private API

	_destroyWidgetsUsedInPopups: function() {
		var destroyWidget = dojo.hitch(this, function(fieldName){
			var widget = this[fieldName];
			if (widget) {
				widget.destroyRecursive();
				this[fieldName] = null;
			}
		});
		destroyWidget("_inputWidget");
		destroyWidget("_editorPopup");
		destroyWidget("_editorForm");
	},

	_createInputWidgetToBeUsedWithinPopup: function() {
		var inputWidgetClass = this._inputWidgetClass,
			inputWidget = new inputWidgetClass(
				this._getAttributeInputWidgetConstructorArgs(false, true)
			);
		dojo.addClass(
			inputWidget.domNode,
			com.ibm.rdm.web.common.Constants.POPUP_ATTRIBUTE_EDITOR_CSS_CLASSNAME
		);
		if (this.width) {
			dojo.style(inputWidget.domNode, "width", this.width);
		}

		// Note: We need to do some "special styling" for String based attributes to
		// enable dealing with long strings ... that's why we make the following
		// dojo.addClass call.
		if (inputWidgetClass === com.ibm.rdm.web.model.attributes.ui.StringAttributeWidget) {
			dojo.addClass(
				inputWidget.domNode,
				"string-attribute-widget"
			);
		}

		return inputWidget;
	},

	_updateCellTextValue: function(value) {
		this._readOnlyInputWidget.set("value", value);
		this._readOnlyInputWidget.updateReadOnlyAttribute();
	},

	_getAttributeInputWidgetConstructorArgs: function(isDisabled, isContextPopupEditor) {
		return {
			value: this.value,
			required: this.required,
			dataType: this.dataType,
			componentURI: this.componentURI,
			processAreaURI: this.processAreaURI,
			onChange: this._onChange,
			onKeyPress: this.onKeyPress,
			enumEntries: this.enumEntries,
			isForSelection: this.isForSelection,
			moduleURI: this.moduleURI,
			disabled: isDisabled,
			parentWidget: this,
			focusNode: this.focusNode,
			attributeTypeURI: this.attributeTypeURI,
			attributeDefinitionURI: this.attributeDefinitionURI,
			messages: this._messages,
			minValue: this.minValue,
			maxValue: this.maxValue,
			viewNode: this.viewNode,
			defaultValue: this.defaultValue,
			resourceURI: this.resourceURI,
			resource: this.resource,
			artifactTypeURI: this.artifactTypeURI,
			adjustInputForMVA: this.adjustInputForMVA,
			allSet: this.allSet,
			allMixed: this.allMixed,
			isMultiValued: this.isMultiValued,
			isUserAllowedToModify: true,
			isSystemDefinedAttribute: this.isSystemDefinedAttribute,
			isContextPopupEditor: isContextPopupEditor,
			isContextSidebar: false
		};
	},

	_onChange: function() {
	},

	_shouldShowWithOkAndCancel: function() {
		var isTypeBoolean = this.dataType === RRMTypes.getWellKnownAttributeTypes()["boolean"],
			isTypeUser	  = this.dataType === RRMTypes.getWellKnownAttributeTypes()["user"];
		if ((isTypeBoolean || isTypeUser || this.enumEntries) && !this.isMultiValued)
		{
			return false;
		} else {
			return true;
		}
	},

	_allowForInitialPopupRepositioning: function(moduleContainerNode) {
		// summary:
		//		For the first 1.5 seconds allow for the popup to be repositioned.
		// description:
		//		As a side-effect, this method will set _doCloseOnModuleScroll to 
		//		true when it's done allowing for repositioning.

		// We're using a setTimeout based re-positioning workaround here.
		// Here's the scenario we're trying to account for:
		// 1. The Name field is present and it's so narrow that most of the
		//    displayed names are wrapped.
		// 2. The user has a row in edit mode.
		// 3. The user invokes a popup editor for a row that appears below this
		//    edit-mode row.
		// The end result of this is that the popup editor is asked to position
		// itself on a row that is about to be "pushed down" because the 
		// edit-mode row is going back into view-mode.  More specifically the
		// Name cell, which in edit-mode is just tall enough to accommodate a
		// single line textbox, is now going back to containing multiple rows 
		// of text.  In other words this edit-mode row, in going back to view-
		// mode becomes taller at the same time that we're trying to invoke a 
		// popup editor for a row below it that is about to get pushed down.

		var editAnchorNode = this._getEditAnchorNode(),
			gridCellNode = this._getGridCellNode();
		var isPositionBetweenPositions = dojo.hitch(this, function(position, smallerPosition, largerPosition){
			return position >= smallerPosition && position <= largerPosition;
		});
		var isCellStillVisible = dojo.hitch(this, function(){
			// Returns true iff the grid cell this editor is associated with is at least
			// partially shown within the underlying viewport.
			var modulePosition = dojo.position(moduleContainerNode),
				cellPosition = dojo.position(gridCellNode),
				cellTop = cellPosition.y,
				cellBottom = cellPosition.y + cellPosition.h,
				cellLeft = cellPosition.x,
				cellRight = cellPosition.x + cellPosition.w,
				moduleTop = modulePosition.y,
				moduleBottom = modulePosition.y + modulePosition.h,
				moduleLeft = modulePosition.x,
				moduleRight = modulePosition.x + modulePosition.w;
			return (
				(
					isPositionBetweenPositions(cellTop, moduleTop, moduleBottom)
					|| isPositionBetweenPositions(cellBottom, moduleTop, moduleBottom)
					|| cellTop <= moduleTop && cellBottom >= moduleBottom
				)
				&& (
					isPositionBetweenPositions(cellLeft, moduleLeft, moduleRight)
					|| isPositionBetweenPositions(cellRight, moduleLeft, moduleRight)
					|| cellLeft <= moduleLeft && cellRight >= moduleRight
				)
			);
		});
		var repositionPopupEditorAfterDelay = dojo.hitch(this, function(delayInMilliseconds, numRetries){
			setTimeout(
				dojo.hitch(this, function(){
					if (isPopupEditorCurrentlyOpened && this._editorPopup && this._editorForm) {
						// Tell the popup editor to reposition itself again.  In most
						// cases this is a no-op.  But occasionally the row associated
						// with the cell gets "pushed down" and this 'position' call
						// will result in the popup editor getting pushed down as well
						// (as it would need to do as its associated cell has moved).
						if (isCellStillVisible())
						{
							// The pencil icon is still visible, reposition the editor popup.
							this._editorPopup.position();
							// Reposition again if needed ..
							if (numRetries > 0) {
								repositionPopupEditorAfterDelay(delayInMilliseconds, --numRetries);
							} else {
								this._doCloseOnModuleScroll = true;
							}
						} else {
							// The cell from which this editor popup was originally launched
							// is no longer visible.  It's likely that the user has used the
							// mouse wheel to scroll the module up, down, left or right so 
							// much so that it now no longer makes sense to show a popup 
							// editor.  Therefore we close the editor popup that is currently
							// shown.
							this._editorPopup.close();
						}
					}
				}),
				delayInMilliseconds
			);
		});
		this._doCloseOnModuleScroll = false;
		repositionPopupEditorAfterDelay(500, 2);
	},

	_getEditAnchorNode: function() {
		// summary:
		//		Returns the "edit icon" (actually an HTML anchor element) associated
		//		with the grid cell that this popup editor is associated with.

		var parentNode = this.domNode.parentNode,
			nodes = dojo.query("a.editButton", parentNode);
		return (
			nodes && nodes.length > 0 ?
			nodes[0] :
			null
		);
	},

	_getModuleContainerNode: function() {
		return domutils.findAncestorByClass(
			this.domNode, "module-container"
		);
	},

	_getGridCellNode: function() {
		return domutils.findAncestorByClass(
			this.domNode, "gridCellWrapper"
		);
	},

	_updateEditIconAppearanceBasedOnEditPopupOpenState: function(isOpened) {
		// summary:
		//		While the editor popup is open, we make sure to always show
		//		the edit icon within the associated grid cell as appearing
		//		with the same UI treatment it would normally get when 
		//		hovered over.

		var anchorNode = this._getEditAnchorNode();
		if (anchorNode) {
			if (isOpened) {
				dojo.addClass(anchorNode, "editPopupOpen");
			} else {
				dojo.removeClass(anchorNode, "editPopupOpen");
			}
		}
	},

	_openEditorPopup: function(/*String?*/title) {
		var isCancel = false,
			focusWasLastWithinEditorForm = false,
			moduleContainerNode = this._getModuleContainerNode();
		var onPopupClosed = dojo.hitch(this, function() {
			// The EditorPopop has closed, so clean things up and if needed
			// make appropriate updates.
			if (!isCancel) {
				if (this.isValid() && this._inputWidget) {
					this.value = this._inputWidget.get("value");
					// Tell the underlying grid about the new value (so that this
					// new value will be used on the next row save).
					this.onChange(this.value);
					// Update the displayed value within the grid.
					this._updateCellTextValue(this.value);
				}
			}
			// Tell the underlying grid that we're done editing.
			this.onClose(focusWasLastWithinEditorForm);
		});
		var onOk = dojo.hitch(this, function(){
			focusWasLastWithinEditorForm = true;
			if (this._editorPopup) {
				this._editorPopup.close();
			}
		});
		var onCancel = dojo.hitch(this, function(){
			focusWasLastWithinEditorForm = true;
			isCancel = true;
			this._editorPopup.close();
		});
		var onSave = dojo.hitch(this, function(){
			onOk();
			// Tell the underlying grid that a save has been requested.
			this.onSave();
		});
		var autoCloseWhenModuleScrolls = dojo.hitch(this, function(){
			if (moduleContainerNode) {
				this.connect(moduleContainerNode, "onscroll", function(event){
					if (this._doCloseOnModuleScroll) {
						this._editorPopup.close();
					}
				});
			}
		});
		var autoCloseWhenItemSelected = dojo.hitch(this, function(){
			this.connect(this._inputWidget, "onChange", dojo.hitch(this, function(){
				onOk();
			}));
		});
		var autoCloseWhenEnterKeyIsPressed = dojo.hitch(this, function(){
			this._keypressHandle = dojo.connect(
				this._inputWidget.focusNode, 
				"onkeypress",
				dojo.hitch(this, function(event){
					if (event.keyCode === dojo.keys.ENTER && this._inputWidget.isValid()) {
						focusWasLastWithinEditorForm = true;
						dojo.disconnect(this._keypressHandle);
						this._keypressHandle = null;
						if (this._editorPopup) {
							this._editorPopup.close();
						}
						dojo.stopEvent(event);
					}
				})
			);
		});
		var createContents = dojo.hitch(this, function() {
			this._inputWidget = this._createInputWidgetToBeUsedWithinPopup();
			this._inputWidget.set("onOk", onOk);
			this._editorForm = new com.ibm.rdm.web.model.attributes.ui.EditorForm({
				title: title,
				onOk: onOk,
				onCancel: onCancel,
				onSave: onSave,
				inputWidget: this._inputWidget,
				shouldShowWithOkAndCancel: this._shouldShowWithOkAndCancel()
			});
			this.domNode.appendChild(this._editorForm.domNode);
			return this._editorForm.domNode;
		});

		// Note: This _destroyWidgetsUsedInPopups call was originally found in 
		// onPopupClosed, but calling it from there resulted in EditManager's
		// "onclick" handler getting invoked as soon as you, for example, click
		// on a value in a BooleanAttributeWidget (I suspect the EditForm it's
		// associated with needs to stick around so that it can recieve
		// [and squelch] the mouseup event?).  We don't want this to occur as it
		// results in the row being saved when it should instead remain open.
		this._destroyWidgetsUsedInPopups();

		// Create a new popup, open it and position it ...
		this._editorPopup = new EditorPopup({
			around: [this._readOnlyInputWidget.domNode],
			createContents: createContents,
			destroyContents: false,
			maxWidth: 350,
			maxHeight: 500,
			programmatic: true,
			orientation: "V",
			moduleContainerNode: moduleContainerNode,
			onOpen: dojo.hitch(this, function() {
				isPopupEditorCurrentlyOpened = true;
				this._updateEditIconAppearanceBasedOnEditPopupOpenState(true);
			}),
			onClose: dojo.hitch(this, function() {
				isPopupEditorCurrentlyOpened = false;
				this._updateEditIconAppearanceBasedOnEditPopupOpenState(false);
				onPopupClosed();
			})
		});
		this._editorPopup.open(this._getEditAnchorNode());
		this._editorPopup.position();
		this._allowForInitialPopupRepositioning(moduleContainerNode);

		// If the user scrolls the underlying module, close this popup.
		autoCloseWhenModuleScrolls();

		// Set up auto-closing for those base attribute types where auto-closing based
		// on single item selection makes sense.
		var isTypeBoolean = this.dataType === RRMTypes.getWellKnownAttributeTypes()["boolean"],
			isTypeUser	  = this.dataType === RRMTypes.getWellKnownAttributeTypes()["user"];
		if ((isTypeBoolean || isTypeUser || this.enumEntries) && !this.isMultiValued)
		{
			autoCloseWhenItemSelected();
		}

		// Set up auto-closing for those base attribute types where auto-closing based
		// on entering Return makes sense.
		else if (
			this.dataType === RRMTypes.getWellKnownAttributeTypes()["int"]
			|| this.dataType === RRMTypes.getWellKnownAttributeTypes()["float"]
			|| this.dataType ===  RRMTypes.getWellKnownAttributeTypes()["duration"]
			|| this.dataType ===  RRMTypes.getWellKnownAttributeTypes()["date"]
			|| this.dataType ===  RRMTypes.getWellKnownAttributeTypes()["dateTime"]
			|| this.dataType ===  RRMTypes.getWellKnownAttributeTypes()["time"]
		) {
			autoCloseWhenEnterKeyIsPressed();
		}
	}
});

com.ibm.rdm.web.model.attributes.ui.PopupAttributeInputWidget.isCurrentlyOpened = function(){
	return isPopupEditorCurrentlyOpened;
};

})();
