/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012, 2016.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */

(function(){
dojo.provide("com.ibm.rdm.web.review.ReviewArtifactWidget");

// Dojo Requires
dojo.require("dijit.layout.ContentPane");

// RM Requires
dojo.require("com.ibm.rdm.web.artifact.ArtifactWidgetInterface");
dojo.require("com.ibm.rdm.web.commenting.CommentingWidget");
dojo.require("com.ibm.rdm.web.common.MessageForm");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.common.PermalinkDialog");
dojo.require("com.ibm.rdm.web.common.ProjectCache");
dojo.require("com.ibm.rdm.web.common.UpdateContext");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.notifications.ChangeNotifier");
dojo.require("com.ibm.rdm.web.notifications.EventTopics");
dojo.require("com.ibm.rdm.web.review.ReviewArtifactsCommentWidget");
dojo.require("com.ibm.rdm.web.review.ReviewContainer");
dojo.require("com.ibm.rdm.web.review.ReviewParser");
dojo.require("com.ibm.rdm.web.review.reviewUpdateMediator");
dojo.require("com.ibm.rdm.web.review.ReviewOverviewSection");
dojo.require("com.ibm.rdm.web.reporting.ReportWizard");
dojo.require("com.ibm.rdm.web.review.reviewChangeNotifier");
dojo.require("com.ibm.rdm.web.review.reviewTypes");
dojo.require("com.ibm.rdm.web.review.reviewutils");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.server.client.utils.ResourceUtil");
dojo.require("com.ibm.rdm.web.util.queryutils");
dojo.require("com.ibm.rdm.web.util.vvcutils");
dojo.require("com.ibm.rdm.web.server.XhrUtils");
dojo.require("com.ibm.rdm.web.pages.WebUIinitializer");
dojo.require("com.ibm.rdm.web.pages.ActionViewCache");
dojo.require("com.ibm.rdm.web.parsers.xml2json");
dojo.require("com.ibm.rdm.web.common.TaskTracker");
dojo.require("com.ibm.rdm.web.server.client.utils.DashboardUtil");

// jazz.ui
dojo.require("jazz.ui.Toolbar");
dojo.require("jazz.ui.toolbar.Button");
dojo.require("jazz.ui.toolbar.Separator");
dojo.require("jazz.util.html");

dojo.require("com.ibm.rdm.web.util.bidiutils"); //bidi

// NLS
dojo.requireLocalization("com.ibm.rdm.web.common", "common");
dojo.requireLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");


var ResourceManager = com.ibm.rdm.web.model.resources.ResourceManager;
var MessageForm = com.ibm.rdm.web.common.MessageForm;
var ProjectCache = com.ibm.rdm.web.common.ProjectCache;
var ChangeNotifier = com.ibm.rdm.web.notifications.ChangeNotifier;
var EventTopics = com.ibm.rdm.web.notifications.EventTopics;
var XhrUtils = com.ibm.rdm.web.server.XhrUtils;
var RDMClient = com.ibm.rdm.web.server.client.RDMClient;
var reviewChangeNotifier = com.ibm.rdm.web.review.reviewChangeNotifier;
var BidiUtils = com.ibm.rdm.web.util.bidiutils; //bidi
var VvcUtils = com.ibm.rdm.web.util.vvcutils;
var ActionViewCache = com.ibm.rdm.web.pages.ActionViewCache;
var TaskTracker = com.ibm.rdm.web.common.TaskTracker;
var DashboardUtil = com.ibm.rdm.web.server.client.utils.DashboardUtil;


dojo.declare(
	"com.ibm.rdm.web.review.ReviewArtifactWidget",
	[com.ibm.rdm.web.artifact.ArtifactWidgetInterface],
{
	// summary:
	//		Presents an editor used for review artifacts.
	// description:
	//		This editor is used for creation of a review, managing its
	//		lifecycle (e.g. Draft -> In Progress -> Reviewed -> Finalized),
	//		and enabling participants to cast votes.
	//		Historical note: Much of the code in this class originally was found
	//		in [now obsolete] ApprovalEditor.js.

	templatePath: dojo.moduleUrl("com.ibm.rdm.web",
		"review/templates/ReviewArtifactWidget.html"),

	///////////////////////////////////////////////////////////////////////////
	// Attach points

	_contentPane: null,
	
	///////////////////////////////////////////////////////////////////////////
	// Public fields

	containerNode: null,

	///////////////////////////////////////////////////////////////////////////
	// Private fields

	// Private Members
	_messages: null,
	_commonMessages: null,
	_actionsToolbar: null,
	_titleNode: null,
	_generateReportButton: null,
	_deleteButton: null,
	_refreshButton: null,
	_titleElement: null,
	_dirtyMarker: null,
	_isDirty: null,
	_breadcrumbsWidget: null,
	_refreshInProgress: null,
	_reviewOverviewSection: null,
	_breadcrumbsSpan: null,
	
	_reviewParser: null,
	_reviewContainer: null,
	
	_onBeforeUnloadHandler: null,
	_bodyClickHandler: null,

	_width: null,
	_height: null,

	_onReviewChanged: null,

	///////////////////////////////////////////////////////////////////////////
	// API

	constructor: function() {
		this.inherited(arguments);
		this._refreshInProgress = false;
	},
	
	postCreate: function() {
		this._contentPane = new dijit.layout.ContentPane({
			
		}, this._contentPane);
		this._toDestroy.push(this._contentPane);
	},
	
	postMixInProperties: function() {
		this.inherited("postMixInProperties", arguments);
		this._messages =
			dojo.i18n.getLocalization("com.ibm.rdm.web.review", "reviewAndApprovalCommon");
		this._commonMessages =
			dojo.i18n.getLocalization("com.ibm.rdm.web.common", "common");
	},

	startup: function() {
		this.inherited(arguments);
		var startupDeferred = new dojo.Deferred();

		DashboardUtil.updateBannerProjectDashboardLink(this.componentURI);

		// Load review 
		this.loadResource().addCallback(this, function(loadResponse) {			
			startupDeferred.callback();
			return loadResponse;
		});

		return startupDeferred;
	},

	uninitialize: function() {
		this._unregisterNavigateHandlers();
		this._removeReviewListeners();
		this.inherited(arguments);		
	},

	onClose: function() {
		if (this.isDirty()) {
			this._revertDirtiedReviewParserBackToCleanState();
		}
	},

	onSave: function() {
		return this._reviewContainer.onSave();
	},

	///////////////////////////////////////////////////////////////////////
	// Artifact Widget API
	
	// summary:
	//		Subclasses must override this. 
	//		This should provide a js array of 'subviews'. Each subview should be a proper dijit._widget
	getSubViews: function() {
		var subViews = [];
		var baseArgs = {
			viewer: this, 
			baselineURI: this.baselineURI,
			resourceURI: this.resourceURI,
			resourceContentType: this.getContentType(),
			selectLinkURI: this._selectLinkURI,
			revisionDate: (this._successorRevision)?this._successorRevision.revisionDate:null,
			resource: this.resource,
			isModule: false
		};
		
		// Create the overview section
		this._reviewOverviewSection =
			new com.ibm.rdm.web.review.ReviewOverviewSection(baseArgs);
		this._toDestroy.push(this._reviewOverviewSection);
		subViews.push(this._reviewOverviewSection);

		// Create the commenting section
		this._commentingSection =
			new com.ibm.rdm.web.commenting.CommentingWidget(baseArgs);
			
		this._commentingSection.setResourceInfo( {
			componentURI: this.componentURI,
			displayName: this.resource.getTitle()
			// TODO - Ideally we would also like to add these attributes, so that commenting widget can add them when
			// creating comments, but getAttributes on Review object does not make them available.
			//createdByUserUri: this.resource.getAttribute(RRMTypes.WELL_KNOWN_SAME_AS_ATTRIBUTE_NAMES.Creator),
			//lastModifiedByUserUri: this.resource.getAttribute(RRMTypes.WELL_KNOWN_SAME_AS_ATTRIBUTE_NAMES.Contributor)
		} );		
			
		this._toDestroy.push(this._commentingSection);
		subViews.push(this._commentingSection);
		
		// Create the review artifact commenting section
		this._reviewArtifactsCommentingSection =
			new com.ibm.rdm.web.review
				.ReviewArtifactsCommentWidget(baseArgs);
		this._toDestroy.push(this._reviewArtifactsCommentingSection);
		subViews.push(this._reviewArtifactsCommentingSection);

		return subViews;
	},
	
	getToolbarItems: function() {
		// summary:
		//		Return a DOM Node containing the toolbar for Reviews
		this._actionsToolbar = new jazz.ui.Toolbar();
		// Add Report button
		this._generateReportButton = this._addGenerateReportAction();
		// Add Delete button
		this._deleteButton = this._addDeleteReviewAction();
		// Add Refresh button
		this._addRefreshArtifactAction();
		this._toDestroy.push(this._actionsToolbar, this._generateReportButton, this._deleteButton);
		
		// TODO: There is a better place for this, find it. --Christian
		this._updateActionsForPermissions();

		return this._actionsToolbar.domNode;
	},
	
	
	getLifecycleItems: function() {
		// summary:
		//		Review and approval does this differently, so we will just
		//		return null for now. Eventually, we would like to get
		//		reviews to behave the same way that "real" artifacts do.
		return null; 
	},
	
	// summary:
	//		Subclasses must override this. 
	//		This should take action to refresh the artifact being displayed 
	handleRefresh: function(){
		if (this.isDirty()) {
			if (confirm(this._commonMessages.unsavedNavigateConfirm)) {
				this._doRefresh();
			}
		} else {
			this._doRefresh();
		}
	},
	
	// summary:
	//		Subclasses must override this. 
	//		This should take action cancel out of the widget, and listen to a 'force' request
	onCancel: function(/*Bool*/force) {
	},
	
	// summary:
	//		Subclasses must override this. 
	//		This should return a widget for placement in the breadcrumb space
	getBreadcrumbs: function(){
		if (!this._breadcrumbsWidget){
			this._breadcrumbsWidget = new jazz.ui.Breadcrumbs({
				segments: [],
				_truncate: function(){return;}
			});
			this._toDestroy.push(this._breadcrumbsWidget);
		}
		if (!this._breadcrumbsSpan){
			this._breadcrumbsSpan = dojo.create("span");
			
			// Add the '|' icon that splits the back navigation button with the breadcrumbs.
			var splitterUri = dojo.moduleUrl("com.ibm.rdm.web", "common/images/icons/obj16/breadcrumbs_splitter.png");
			var splitterImg = dojo.create("img", {
					src: splitterUri,
					className: "breadcrumbs-splitter"
			});
			dojo.place(splitterImg, this._breadcrumbsSpan, "only");
			
			dojo.place(this._breadcrumbsWidget.domNode, splitterImg, "after");
		}

		var projectDeferred = ProjectCache.getComponent(this.componentURI);
		
		projectDeferred.addCallback(this, function(project) {
			var actionURI = null, breadcrumbPath = [];
			if (this.baselineURI){
				actionURI = "#action=com.ibm.rdm.web.pages.showProjectHistory&componentURI=" + encodeURIComponent(this.componentURI) +
				"&baselineURI=" + encodeURIComponent(this.baselineURI);
			}
			else{
				actionURI = VvcUtils.addCurrentConfigurationToActionFragmentIfNeeded("#action=com.ibm.rdm.web.pages.showProjectDashboard&componentURI=" + encodeURIComponent(this.componentURI));
			}
			
			breadcrumbPath.push({
				label: project.name,
				href: actionURI
			});
			
			this._breadcrumbsWidget.segments = breadcrumbPath;
			this._breadcrumbsWidget.refresh();
			
			// Removing the '>' text node after the link, because we manually add it later in this function
			var breadcrumbsWidgetSpan = this._breadcrumbsWidget.domNode.firstChild;
			if (breadcrumbsWidgetSpan) {
				breadcrumbsWidgetSpan.removeChild(breadcrumbsWidgetSpan.lastChild);
			}
			
			// Set the breadcrumb widget's text and title, and include an arrow.
			var breadcrumbsJazzLink = this._breadcrumbsWidget.domNode.firstChild.firstChild;
			var breadcrumbsLabel = this._commonMessages.breadcrumbsLabelSingleFolder;
			breadcrumbsLabel = dojo.string.substitute(breadcrumbsLabel, [breadcrumbsJazzLink.text]);
			breadcrumbsJazzLink.text = breadcrumbsLabel;
			breadcrumbsJazzLink.title = breadcrumbsLabel;
			
			return project;
		});

		return this._breadcrumbsSpan;
	},
	
	getTitleNode: function(){
		// summary:
		//		Reviews are not renamable, so this simply returns a div
		//		containing the title and the Context Sensitive Help
		//		bits.

		this._titleNode = dojo.create("div", {"class": "resource-title-cell"});
		this._titleElement =
			dojo.create(
				"span",
				{
					"class": "resource-title",
					innerHTML: jazz.util.html.escape(this.getTitle())
				},
				this._titleNode
			);
		BidiUtils.applyTextDirSetting(this._titleElement, this._titleElement.innerHTML);			
		this._dirtyMarker = dojo.create("span", {"class": "dirtyMarker", innerHTML: "*"}, this._titleNode);
		
		//Setup help node bits
		this._cshNode = dojo.create("span", {"class": "contextSensitiveHelp"}, this._titleNode);
		var pageHoverHelp = new jazz.app.ua.PageHoverHelp({uaKey: com.ibm.rdm.web.help.context.review});
		this._toDestroy.push(pageHoverHelp);
		this._cshNode.appendChild(pageHoverHelp.domNode);
		dojo.create("br", null, this._titleNode);
		
		return this._titleNode;
	},
	
	getTitle: function(){
		// summary:
		//		Return the review name as the title for the viewer.
		
		// Only grab the name if we have our review parser.
		//
		// If the editor has been instantiated but the review data has not yet been loaded,
		// we can use the resource to get the reviewName.
		// See SidebarArtifactLayout ArtifactWidgetFactory.get call.
		if (this._reviewParser) {
			return this._reviewParser.getReviewName();
		}
		else if (this.resource) {
			return this.resource._reviewParser.getReviewName();
		}
	},
	
	
	//This should be passed upon instantiation by the layout widget.
	refreshHook: function(){
		throw new Error('Unimplemented API: refreshHook undefined. Must be passed at instantiation');
	},

	///////////////////////////////////////////////////////////////////////
	// Review and Approval API

	start: function() {
		// summary:
		//		Start and save the review
		
		var startDeferred = new dojo.Deferred();
		
		// When user acknowledges the child configuration alert for starting a formal review, reload review with that configuration.
		// This requires GET of the review to learn its configuration then reloading the window with that configuration.
		var onChildConfigurationAlertOk = dojo.hitch(this, function() {
			RDMClient.doGet({
				uri:this._reviewParser.getSelfUri(),
				skipCache: true
			}).then(
					dojo.hitch(this, function(response) {
						// Get URI of child workspace in the newly saved review
						var xmlParser = new com.ibm.rdm.web.parsers.xml2json(response.doc);
						var reply = xmlParser.parse();
						var reviewConfigurationURI = reply.RDF.Review.configuration.value;
						
						// Invalidate cached page and switch to review in new workspace
						ActionViewCache.removeCachedActionView("showArtifact", this._reviewParser.getSelfUri());
						var uri = VvcUtils.replaceConfigurationInActionUri(window.location.href, reviewConfigurationURI);			
						window.location.href = uri;
					})
			);
		});
		
		var onSaveComplete = dojo.hitch(this, function() {
			this._reviewContainer.showAsBusy(false);
			this.setDirty(false);
			
			// If a formal review started, need to determine what CM configuration it created, then reload review in that workspace;
			// alert user before doing so
			if (this._reviewParser.isFormal()) {
				var confirmContent = this._messages["changeConfigurationAfterStartConfirmation"];
				var form = new MessageForm({
					confirmContent: confirmContent,
					dialogTitle: this._messages["confirmSwitchConfigurationAfterStartDialogTitle"],
					dialogStyle: MessageForm.DIALOG_TYPES.Ok,
					dialogWidth: "600px",
					onOk: onChildConfigurationAlertOk,
					closable: false
				});
				form.show();
			}			
		});
		var onSaveError = dojo.hitch(this, function(errObj) {
			this._reviewContainer.showAsBusy(false);
		});
		
		startDeferred.addCallback(onSaveComplete);
		startDeferred.addErrback(onSaveError);

		// Save the review resource.
		// TODO: Declare these functions in the scope of the
		// ReviewArtifactWidget. Initial API adoption time constraints
		// prevented me from doing so, but they will need to move
		// before the refactor is complete. The same function appears
		// repeated in ReviewArtifactWidget.resume().
		// --Christian
		var changeStateToStartedAndSaveReview = dojo.hitch(this, function() {

			/*
			 * Review resource completed saving; if it was a formal review that was started,
			 * the page will need to be reloaded with the new configuration. Otherwise,
			 * if any participant results changed, save them after the review is saved.
			 * (Participant results won't need saving if the review was started for the first time.)
			 */
			var onReviewSaveCompleted = dojo.hitch(this, function(response) {
				// Task tracker returned for starting a formal review for the first time; reload page
				if (response && response.taskTracker) {
					startDeferred.callback();
				}

				// No new stream created, reload page not needed
				else {
					var userURIs =
						this._reviewParser.getUserURIsThatHaveDirtyParticipantResults();
	
					var onParticipantResultsSaveCompleted = dojo.hitch(this, 
						function() {
	
						this._performRefresh(
							userURIs,
							function() { startDeferred.callback(); }
						);
					});
					if (userURIs.length > 0) {
						this._saveParticipantResults(
							userURIs, onParticipantResultsSaveCompleted, onSaveError);
					} else {
						startDeferred.callback();
					}
			    }
			});
			this._reviewParser.setReviewState(
					com.ibm.rdm.web.review.reviewTypes.REVIEW_STATE_TYPES.Started);
			
			// TODO: Eventually, we will need to refactor the
			// reviewUpdateMediator so that the last two arguments
			// are no longer required. --Christian
			return com.ibm.rdm.web.review.reviewUpdateMediator.saveReview(
				this._reviewParser, 
				onReviewSaveCompleted,
				dojo.hitch(function(errObj) {
					startDeferred.errback(errObj);
					return errObj;
				}));
		});
		
		/*
		 * When a formal review is started for the first time, it creates a child stream;
		 * this may take a long time so alert user with a different message.
		 */
		if (this._reviewParser.isFormal() && this._reviewParser.getReviewState() === 	com.ibm.rdm.web.review.reviewTypes.REVIEW_STATE_TYPES.Draft) {
			this._reviewContainer.showAsBusy(true, this._messages.startingFormalReview);
		}
		else {
			this._reviewContainer.showAsBusy(true, this._messages.savingChangesToReview);
		}

		// Only for formal reviews
		var onUpdatedArtifactInformationComplete = dojo.hitch(this,
			function(resourceDescriptions, startTimestampToUse) {

			this._reviewParser.beginBatchChanges();
			this._reviewParser.updateArtifactInfos(resourceDescriptions);
			this._reviewParser.changeStartTimestamp(startTimestampToUse);
			this._reviewParser.endBatchChanges();
			changeStateToStartedAndSaveReview();
		});
		
		// Start the chain of deferreds to save the review here
		if (this._reviewParser.isFormal()) {
			com.ibm.rdm.web.review.reviewutils.obtainMostRecentArtifactInformation(
				this._reviewParser, onUpdatedArtifactInformationComplete, true);
		} 
		else {
			// Using this get request to get the start date for a informal review.
			// Because we are in the review this will hit cache, and shouldn't be a 
			// terrible performance hit.
			RDMClient.doGet({
				uri:this._reviewParser.getSelfUri()
			}).then(
					dojo.hitch(this, function(response) {
						var startTimestampToUse = XhrUtils.getResponseLastModified(response);
						this._reviewParser.changeStartTimestamp(startTimestampToUse);
						changeStateToStartedAndSaveReview();
					})
				);
		}
		
		return startDeferred;
	},
	
	pause: function() {
		// summary:
		//		Pause and save the review
		
		this._reviewContainer.showAsBusy(true, this._messages.savingChangesToReview);
		this._reviewParser.setReviewState(
			com.ibm.rdm.web.review.reviewTypes.REVIEW_STATE_TYPES.Paused);
		return com.ibm.rdm.web.review.reviewUpdateMediator.saveReview(
			this._reviewParser, dojo.hitch(this, function(res) {
				
				this._reviewContainer.showAsBusy(false);
				this.setDirty(false);
				// We refresh here as we need to make sure to we've got the most 
				// recent participant results for all participants.
				this._doRefresh();
				return res;
			}), dojo.hitch(this, function(res) {
				this._reviewContainer.showAsBusy(false);
				return res;
			}));
	},

	save: function() {
		return this._reviewContainer.onSave();
	},
	
	resume: function() {
		// summary:
		//		Continue and save the review

		this._reviewContainer.showAsBusy(true,
			this._messages.savingChangesToReview);
		
		var resumeDeferred = new dojo.Deferred();

		var onSaveComplete = dojo.hitch(this, function() {
			this._reviewContainer.showAsBusy(false);
			this.setDirty(false);
		});
		var onSaveError = dojo.hitch(this, function(errObj) {
			this._reviewContainer.showAsBusy(false);
		});
		
		resumeDeferred.addCallback(onSaveComplete);
		resumeDeferred.addErrback(onSaveError);
		var changeStateToStartedAndSaveReview = dojo.hitch(this, function() {

			var onReviewSaveCompleted = dojo.hitch(this, function() {
				var userURIs =
					this._reviewParser.getUserURIsThatHaveDirtyParticipantResults();

				var onParticipantResultsSaveCompleted = dojo.hitch(this, function() {
					this._performRefresh(
						userURIs,
						function() { resumeDeferred.callback(); }
					);
				});
				if (userURIs.length > 0) {
					this._saveParticipantResults(
						userURIs, onParticipantResultsSaveCompleted, onSaveError);
				} else {
					resumeDeferred.callback();
				}
			});
			this._reviewParser.setReviewState(
				com.ibm.rdm.web.review.reviewTypes.REVIEW_STATE_TYPES.Started);
			
			// TODO: Eventually, we will need to refactor the
			// reviewUpdateMediator so that the last two arguments
			// are no longer required. --Christian
			return com.ibm.rdm.web.review.reviewUpdateMediator.saveReview(
				this._reviewParser, onReviewSaveCompleted,
				resumeDeferred.errback);
		});
	
		var onUpdatedArtifactInformationComplete = dojo.hitch(this,
		    function(resourceDescriptions, startTimestampToUse) {

			this._reviewContainer.showAsBusy(true, this._messages.savingChangesToReview);
			this._reviewParser.beginBatchChanges();
			var anythingUpdated = this._reviewParser.updateArtifactInfos(resourceDescriptions);
			if (anythingUpdated) {
				// Avoid unnecessarily updating the review's start time if nothing changed during the pause
				this._reviewParser.changeStartTimestamp(startTimestampToUse);
			}
			this._reviewParser.endBatchChanges();
			changeStateToStartedAndSaveReview();
			
			this._reviewContainer.showAsBusy(false);
		}); // End onUpdatedArtifactInformationComplete

		if (this._reviewParser.isFormal()) {
			com.ibm.rdm.web.review.reviewutils
				.obtainMostRecentArtifactInformation(
					this._reviewParser, onUpdatedArtifactInformationComplete);
		} else {
			this._reviewContainer.showAsBusy(true,
				this._messages.savingChangesToReview);
			changeStateToStartedAndSaveReview();
		}
		
		return resumeDeferred;
	},
	
	complete: function() {
		// summary:
		//		Finalize and save the review.
		// description:
		//		This correlates to "Finalize" in the Review/Approval UI,
		//		but we couldn't call it that without breaking dojo.
		
		var completeDeferred = new dojo.Deferred();
		
		var headDeferred = RDMClient.doHead({
			uri: this._reviewParser.getSelfUri(),
			skipCache: true,
			skipErrors: true
		});

		var onSaveComplete = dojo.hitch(this, function(response) {
			this._reviewContainer.showAsBusy(false);
			this.setDirty(false);
			completeDeferred.callback(response);

			// Retrieve the review, and switch to the new baseline it created
			RDMClient.doGet({
				uri:this._reviewParser.getSelfUri(),
				skipCache: true
			}).then(
				dojo.hitch(this, function(response) {
					// Get URI of baseline in the newly finalized review
					var xmlParser = new com.ibm.rdm.web.parsers.xml2json(response.doc);
					var reply = xmlParser.parse();
					var reviewConfigurationURI = reply.RDF.Review.configuration.value;
					
					// Invalidate cached page and switch to review in new workspace
					ActionViewCache.removeCachedActionView("showArtifact", this._reviewParser.getSelfUri());
					var uri = VvcUtils.replaceConfigurationInActionUri(window.location.href, reviewConfigurationURI);			
					window.location.href = uri;
				})
			);
		});

		var onSaveError = dojo.hitch(this, function(errObj) {
			this._reviewContainer.showAsBusy(false);
		});
		
		headDeferred.addCallback(this, function(headResponse) {
			var etag = XhrUtils.getResponseEtag(headResponse);
			var revision = XhrUtils.getResponseRevision(headResponse);
			this._reviewParser.setRevisionAndETag(revision, etag);
			
			/*
			 * When a formal review is finalized, it creates a baseline of the review stream;
			 * this may take a long time so alert user with a different message.
			 */
			if (this._reviewParser.isFormal()) {
				this._reviewContainer.showAsBusy(true, this._messages.finalizingFormalReview);
			}
			else {
				this._reviewContainer.showAsBusy(true, this._messages.savingChangesToReview);
			}

			this._reviewParser.setReviewState(com.ibm.rdm.web.review
				.reviewTypes.REVIEW_STATE_TYPES.Finalized);
			
			var saveDeferred =
				com.ibm.rdm.web.review.reviewUpdateMediator.saveReview(
						this._reviewParser, onSaveComplete, onSaveError);
			
			return headResponse;
		});
		
		headDeferred.addErrback(this, function(errObj) {
			console.error("Error doing HEAD on review:", errObj);
		});
		
		return completeDeferred;
	},
	
	getReviewParser: function() {
		return this._reviewParser;
	},
	
	isInReviewMode: function() {
		return false;
	},
	
	isInHistoryMode: function() {
		return false;
	},

	isStoryboardViewer: function(){
		return false;
	},
	
	//	summary:
	//		Comments can only be made on the review artifact.
	supportsElementCommenting: function() {
		return false;
	},
	
	isMerge: function(){
		return false;
	},
	
	isReviewArtifact: function() {
		return true;
	},
	
	getContentType: function() {
		return com.ibm.rdm.web.common.MimeTypes.CONTENT_TYPE_REVIEW;
	},
	
	isDirty: function(){
		return this._isDirty;
	},

	setDirty: function(dirty){
		this.setWindowTitleDirty(dirty);
		if (dirty){
			dojo.style(this._dirtyMarker, "display", "inline");
			this._registerNavigateHandlers();
		}
		else{
			dojo.style(this._dirtyMarker, "display", "none");
			this._unregisterNavigateHandlers();
		}
		this._isDirty = dirty;
	},
	
	setDimensions: function(width, height) {
		this.inherited(arguments);
		this._width = width;
		this._height = height;
		if (this._reviewContainer) {
			this._reviewContainer.setDimensions(width, height);
		}
	},
	
	loadResource: function(revision, skipCache) {
		if(!revision || revision != this.revision) {
			skipCache = true;
		}
		this.revision = revision;
		return ResourceManager.get({
				resourceURI: this.resourceURI,
				projectSnapshotURI: this.baselineURI,
				skipCache: skipCache
			}, dojo.hitch(this, this._handleLoadResponse));
	},
	
	///////////////////////////////////////////////////////////////////////
	// Private API
	
	_addReviewListeners: function() {
		if (!this._onReviewChanged) {
			var onReviewChanged = dojo.hitch(this, function(reviewParser) {
				this._reviewParser = reviewParser;
			});
			reviewChangeNotifier.addReviewListener(
				this._reviewParser.getSelfUri(),
				onReviewChanged
			);
			this._onReviewChanged = onReviewChanged;
		}
	},

	_removeReviewListeners: function() {
		if (this._onReviewChanged) {
			reviewChangeNotifier.removeReviewListener(
				this._reviewParser.getSelfUri(),
				this._onReviewChanged
			);
		}
	},
	
	_getPrettyURI: function() {
		// summary:
		// 		Compute and return the pretty URI of the artifact.
		
		return this.getResourceURI() + "?vvc.configuration=" + encodeURIComponent(VvcUtils.getCurrentConfigurationUri());
	},
	
	// Creates a pop-up message form that displays the pretty URI
	// A modified version of this function is found in AbstractResourceArtifactWidget.js
	_openPrettyLinkMessage: function() {
		var displayTitle = this.resource.getTitle();
		displayTitle = jazz.util.html.escape(displayTitle);
		var linkURI = this._getPrettyURI();
		var permalinkDialog = new com.ibm.rdm.web.common.PermalinkDialog({
			initLinkURI: linkURI, 
			initDisplayText: displayTitle, 
			initDialogWidth: "450px"
		});
		this._toDestroy.push(permalinkDialog);
		permalinkDialog.show();
	}, 
	
	_saveParticipantResults: function(userURIs, onSaveComplete, onSaveError) {
		// Note: we take care here to serialize updates for participant results
		// in order to work around the issue caused be server defect 23384.
		var userIndex = 0;
		var reviewParser = this._reviewParser;
		var participantResultInfos = reviewParser.getParticipantResultInfos();
		var saveSingleUserParticipantResults;
		var onParticipantResultSaveComplete = function() {
			userIndex++;
			if (userIndex < userURIs.length) {
				saveSingleUserParticipantResults(userURIs[userIndex]);
			} else {
				onSaveComplete();
			}
		};
		saveSingleUserParticipantResults = function(userURI) {
			var isDone = (participantResultInfos[userURI].artifactStatusType !== null);
			com.ibm.rdm.web.review.reviewUpdateMediator.saveParticipantVotes(
				reviewParser,
				userURI,
				isDone,
				onParticipantResultSaveComplete,
				onSaveError
			);
		};
		saveSingleUserParticipantResults(userURIs[userIndex]);
	},
	
	_confirmWhetherToIncludeNewerVersions: function(artifactInfosThatCouldBeUpdated, onConfirmYes, onConfirmNo) {

		var confirmContent =
			this._messages["confirmIncludeNewerArtifactVersions"] + "<br/>";
		for (var i = 0; i < artifactInfosThatCouldBeUpdated.length; i++) {
			confirmContent += "<br/>" + artifactInfosThatCouldBeUpdated[i].artifactName;
		}
		var form = new MessageForm({
			confirmContent: confirmContent,
			dialogTitle: this._messages["confirmNewerVersionsDialogTitle"],
			dialogStyle: MessageForm.DIALOG_TYPES.YesNo,
			dialogWidth: "300px",
			onOk: onConfirmYes,
			onCancel: onConfirmNo
		});
		this._toDestroy.push(form);
		
		form.show();
	},
	
	_isAtLeastOneParticpantResultMissing: function(userURIs) {
		var isAtLeastOneParticpantResultMissing = false;
		var participantInfos = this._reviewParser.getParticipantInfos();
		for (var i = 0; i < userURIs.length; i++) {
			var userURI = userURIs[i];
			var participantInfo = participantInfos[userURI];
			var participantStatusInfo = participantInfo.participantStatusInfo;
			var atLeastOneVoteIsPresent = (
				participantStatusInfo.numApproved > 0
				|| participantStatusInfo.numDisapproved > 0
				|| participantStatusInfo.numAbstained > 0
				|| participantStatusInfo.numReviewed > 0
				|| participantStatusInfo.numInProgress > 0
			);
			if (!atLeastOneVoteIsPresent) {
				isAtLeastOneParticpantResultMissing = true;
				break;
			}
		}
		return isAtLeastOneParticpantResultMissing;
	},
	
	_performRefresh: function(userURIs,	/*Callback?*/onRefreshComplete,
			/*Boolean?*/doAvoidWorkaround) {
		var self = this;
		com.ibm.rdm.web.common.UpdateContext.cachedArtifactHasBeenSaved();
		var onReviewRefreshComplete = function() {
			// We're using a workaround here to avoid the issue caused by defect 19081.
			if (!doAvoidWorkaround && self._isAtLeastOneParticpantResultMissing(userURIs)) {
				setTimeout(
					function() {
						// recurse
						self._performRefresh(userURIs, onRefreshComplete, true);
					},
					500
				);
			}
			else {
				if (onRefreshComplete) {
					onRefreshComplete();
				}
			}
		};
		com.ibm.rdm.web.review.reviewUpdateMediator.handleRefresh(
			this._reviewParser, onReviewRefreshComplete
		);
	},
	
	_handleLoadResponse: function(response) {
		this.resource = response;
		this._reviewParser = response.getReviewParser();
		this.revision = this._reviewParser.getRevision();
		this._addReviewListeners();
		this._populateContents(this._reviewParser);
		this._checkForReviewSaveInProgress(this._reviewParser);
		
		return response;
	},
	
	_checkForReviewSaveInProgress: function(reviewParser) {
		// summary:
		//		Check to see if the review has a task resource associated with it,
		//      alerting us that someone's still saving a review during a lengthy
		//      start/finalize request. If there is such a task, disable the review
		//      and poll the task tracker until the review completes saving (and then
		//      reload the UI with the saved review).
		
		// Only bother with this check for formal reviews that aren't finalized. Informal reviews don't use task trackers.
		// And, a finalized review can't still be saving.
		if (!reviewParser.isFormal() == this._reviewParser.getReviewState() === com.ibm.rdm.web.review.reviewTypes.REVIEW_STATE_TYPES.Finalized) {
			return;
		}
		
		// Get the review to learn its task tracker, if any
		var headDeferred = RDMClient.doGet({
				uri: reviewParser.getSelfUri(),
				skipCache: true,
				skipErrors: true
			});
		
		// Check, and wait for this review to save, if necessary
		headDeferred.addCallback(this, function(headResponse) {
			var taskProgressURI = XhrUtils.getHeader(headResponse, "Task-Tracker");
			
			// Found a task for this review
			if (taskProgressURI) {
				// Put up modal dialog to block screen and alert user
				var dialog = new jazz.ui.Dialog({
					primaryTitle: this._messages["savingLongChangesToReviewTitle"],
					contentNode: dojo.create('div', {'class': 'rdmDialogForm', innerHTML: this._messages["savingLongChangesToReview"]}),
					closable: false,
					width: '350px',
					closeOnEsc: false,
					underlayArgs: {
						opacity: '0.4',
						color: '#333'
					}
				});		
				
				var handleOnProgress = dojo.hitch(this, function(){
					// console.debug("Progress...");
				});
				
				// Task tracker completed saving the review; reload the UI with the updated review.
				var handleOnComplete = dojo.hitch(this, function(){
					// console.debug("Complete");
					
					RDMClient.doGet({
						uri:this._reviewParser.getSelfUri(),
						skipCache: true
					}).then(
							dojo.hitch(this, function(response) {
								dialog.close();
								
								// Get URI of child workspace in the newly saved review
								var xmlParser = new com.ibm.rdm.web.parsers.xml2json(response.doc);
								var reply = xmlParser.parse();
								var reviewConfigurationURI = reply.RDF.Review.configuration.value;
								
								// Invalidate cached page and switch to review in new workspace
								ActionViewCache.removeCachedActionView("showArtifact", this._reviewParser.getSelfUri());
								var uri = VvcUtils.replaceConfigurationInActionUri(window.location.href, reviewConfigurationURI);			
								window.location.href = uri;
							})
					);
					
				});

				var handleOnError = dojo.hitch(this, function(){
					console.debug("Error in task tracker");
				});

				// Poll task tracker until it is complete, which then invokes handleOnComplete callback
		    	new TaskTracker({
					location: taskProgressURI,
					onProgress: handleOnProgress,
					onComplete: handleOnComplete,
					onError: handleOnError
				});
			}
			
		});
		
		headDeferred.addErrback(this, function(errObj) {
			console.error("Error doing HEAD on review: ", errObj);
		});
		
	},
	
	_populateContents: function(reviewParser) {
		if (this._reviewContainer) {
			com.ibm.rdm.web.review.reviewUpdateMediator.handleRefresh(reviewParser, null, true);
		} else {
			var reviewContainer = new com.ibm.rdm.web.review.ReviewContainer({
				resourceViewer: this,
				reviewParser: this._reviewParser,
				width: this._width,
				height: this._height
			});
			dojo.place(reviewContainer.domNode, this._contentPane.domNode, "only");
			this._contentPane.startup();
			reviewContainer.startup();
			this._reviewContainer = reviewContainer;
			this._toDestroy.push(reviewContainer);
			com.ibm.rdm.web.review.reviewUpdateMediator.reviewHasBeenOpened(reviewParser);
		}
	},
	
	_updateActionsForPermissions: function() {
		// summary:
		//		Retrieve current user's permissions, and enable
		//		actions for this review appropriately. Note that
		//		the deferred will fail silently if there is a 
		//		problem retrieving permissions.

		var ownerUri = this._reviewParser.getOwnerUserFullURI();
		var componentURI = this._reviewParser.getComponentUri();

		var onPermissionsRetrieved = dojo.hitch(this,
			function(permissionObject) {
				if(permissionObject.item(ProjectCache.MODIFY_ANY_REVIEW)) {
					this._deleteButton.set("enabled", true);
				}
			return permissionObject;
		});

		if(ownerUri != com.ibm.rdm.web.AUTHENTICATED_USER.uri) {
			var permissionsDeferred = ProjectCache
				.getProjectPermissions(componentURI);
			permissionsDeferred.addCallback(this, onPermissionsRetrieved);
		}
		else {
			this._deleteButton.set("enabled", true);
		}
	},
	
	_addDeleteReviewAction: function(){
		// summary:
		//		Create the "Delete" toolbar button, and disable it before
		//		returning it.
		var deleteButton = new jazz.ui.toolbar.Button({
			title: this._messages.deleteResource, 
			onClick: dojo.hitch(this, this._handleDeleteResourceButtonClick), 
			imgClass: "deleteCommand", 
			enabled: false
		});
		this._actionsToolbar.addChild(deleteButton);

		return deleteButton;
	},
	
	_handleDeleteResourceButtonClick: function() {
		com.ibm.rdm.web.server.client.utils.ResourceUtil.deleteResource({
			resourceURI: this.resourceURI,
			componentURI: this.componentURI,
			title: this.getTitle(),
			isReview: true,
			invalidateEditorAfterDelete: true
		});
	},
	
	_addGenerateReportAction: function(){
		var button = new jazz.ui.toolbar.Button({
			title: this._messages.createdReport, 
			onClick: dojo.hitch(this, this._handleCreatedReportButtonClick), 
			imgClass: "create-report-toggle-menu"
		});
		this._actionsToolbar.addChild(button);
		return button;
	},
	
	_handleCreatedReportButtonClick: function(){
		var getReportArtifacts = dojo.hitch(this, function(){	
			
			var reviewToReportDeferred = new dojo.Deferred();
			
			ResourceManager.get({
				resourceURI: this.resourceURI
			}, dojo.hitch(this, function(reviewResourceResponse){
				var reviewParser = reviewResourceResponse.getReviewParser();
				var reviewName   = reviewParser.getReviewName();
				var reviewState  = reviewParser.getReviewState();
				
				var reviewInfo =[];
				reviewInfo.push({
					label: reviewName,
					reviewState: reviewState,
					resourceURI: this.resourceURI,
					componentURI: this.componentURI
				});
				reviewToReportDeferred.callback(reviewInfo);
				
				return reviewResourceResponse;
			}));
			
			return reviewToReportDeferred;
		});
		
		var wizard = new com.ibm.rdm.web.reporting.ReportWizard({
			"componentURI": this.componentURI,
			"isReviewReport": true,
			"getReportArtifactsMethodDeferred":getReportArtifacts,
			"context": com.ibm.rdm.web.reporting.ReportWizard.viewType.REVIEW
		});
		this._toDestroy.push(wizard);
		
		wizard.show();
	},
	
	_doRefresh: function() {
		if (!this._refreshInProgress) {
			// refresh the configuration selector
			ChangeNotifier.notifyListeners({eventTopic: EventTopics.CURRENT_CONFIGURATION_CHANGED});
			this._reviewContainer.showAsBusy(true,
				this._commonMessages.loading);
			var loadResourceAndLoadCommentsIfNeeded = dojo.hitch(this, function(revision) {
				var deferredLoadRequest = this.loadResource(revision, true);
				deferredLoadRequest.addCallback(this, function(result) {
					
					var userURIs =
						this._reviewParser.getUserURIsThatHaveDirtyParticipantResults();

					if (userURIs.length > 0) {
						this._performRefresh(userURIs);
					}
					
					this._refreshInProgress = false;
					this._reviewParser = result.getReviewParser();
					com.ibm.rdm.web.review.reviewUpdateMediator.handleRefresh(
						this._reviewParser, null, true
					);
					this._reviewContainer.showAsBusy(false);
					return result;
				});
				deferredLoadRequest.addErrback(this, function(e){
					console.error("Error performing head request in ResourceViewer.handleRefresh\n" + e);
					this._refreshInProgress = false;
					this._reviewContainer.showAsBusy(false);
					return e;
				});
			});
			
			var deferredHeadRequest =
				RDMClient.doHead({
					uri: this.resourceURI,
					skipCache: true
				});
			deferredHeadRequest.addCallback(this, function(response) {
				var newRevision = XhrUtils.getResponseRevision(response);
				loadResourceAndLoadCommentsIfNeeded(newRevision);
				return response;
			});
			deferredHeadRequest.addErrback(this, function(response){
				if (this._isInFormalReviewMode && response.ioArgs.xhr.status === 410) { // Note: 410 is "Gone"
					// See above Note ... we still want to continue loading (after
					// alerting the user to the fact that the artifact has been
					// deleted).
					com.ibm.rdm.web.common.MessageHandler.showError(7551);

					loadResourceAndLoadCommentsIfNeeded(this.revision);
				}
				else {
					console.error("Error performing head request in ResourceViewer.handleRefresh\n" + response);
					this._markRefreshAsDone();
				}
				return response;
			});
		}
	},
	
	_registerNavigateHandlers: function() {
		var handleOnBeforeUnload = dojo.hitch(this, function() {
			if (this.isDirty()) {
				var reviewParser = this.getReviewParser();
				var reviewName = (
					reviewParser ? reviewParser.getReviewName() : ""
				);
				return dojo.string.substitute(
					this._messages.reviewHasUnsavedChanges, [reviewName]
				);
			} else {
				return false;
			}
		});
		var handleBodyClick = dojo.hitch(this, function(urlObject) {
			var doVetoClickAwayEvent;
			if (!document.getElementById(this.id)) {
				// If this DOM node has been orphaned (e.g. the user previously
				// made use of the browser's Back button when this editor was
				// in a dirty state) we need to ignore this click callback.
				doVetoClickAwayEvent = false;
			}
			else if (urlObject.uri.indexOf("#action") === -1) {
				// We want to ignore clicks for things like the Print Button,
				// the Refresh Button, etc. ... so we ignore if we see the
				// action parameter.
				doVetoClickAwayEvent = false;
			}
			else {
				if (this.isDirty()) {
					if (confirm(this._commonMessages["unsavedNavigateConfirm"])) {
						doVetoClickAwayEvent = false;
						// We need to revert this editor back to the clean state
						// (in case it's subsequently restored via the action
						// view cache).
						this._revertDirtiedReviewParserBackToCleanState();
					} else {
						doVetoClickAwayEvent = true;
					}
				} else {
					doVetoClickAwayEvent = false;
				}
			}
			return doVetoClickAwayEvent;
		});
		if (!this._onBeforeUnloadHandler) {
			this._onBeforeUnloadHandler =
				net.jazz.ajax.ui.PlatformUI.getWorkbench().registerOnBeforeUnloadHandler(
					handleOnBeforeUnload
				);
		}
		if (!this._bodyClickHandler) {
			this._bodyClickHandler =
				net.jazz.ajax.ui.PlatformUI.getWorkbench().registerBodyClickHandler(
					handleBodyClick
				);
		}
	},
	
	_unregisterNavigateHandlers: function() {
		if (this._onBeforeUnloadHandler) {
			net.jazz.ajax.ui.PlatformUI.getWorkbench().unregisterOnBeforeUnloadHandler(
				this._onBeforeUnloadHandler
			);
			this._onBeforeUnloadHandler = null;
		}
		if (this._bodyClickHandler) {
			net.jazz.ajax.ui.PlatformUI.getWorkbench().unregisterBodyClickHandler(
				this._bodyClickHandler
			);
			this._bodyClickHandler = null;
		}
	},

	_revertDirtiedReviewParserBackToCleanState: function() {
		var reviewParser = this.getReviewParser();
		if (reviewParser) {
			this.setDirty(false);
			ResourceManager.get(
				{
					resourceURI: this.resourceURI,
					revision: reviewParser.getRevision(),
					projectSnapshotURI: this.baselineURI
				},
				dojo.hitch(this, this._handleLoadResponse)
			);
		}
	}
});

})();
