/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
(function() {
dojo.provide("com.ibm.rdm.web.server.XhrUtils");
dojo.require("com.ibm.rdm.web.util.queryutils");
dojo.require("com.ibm.rdm.web.server.Headers");
dojo.require("com.ibm.rdm.web.parsers.xml2json");

var queryutils = com.ibm.rdm.web.util.queryutils;
var Headers = com.ibm.rdm.web.server.Headers;

com.ibm.rdm.web.server.XhrUtils = {
	
	/////////////////////////////////////////////////////////////////
	// API
	
	getResponseEtag: function(response, doNotStripQuotes) {
		var etag = null;
		try {
			etag = response.ioArgs.xhr.getResponseHeader("ETag");
			//This following should no longer be necessary - as etags are no longer used to represent revision
			//and this causes issues for server to compare when using If-Match or If-None-Match headers
			//as comparisons. -ebcraig
			// if(!doNotStripQuotes) {
			// 	if (etag && etag.indexOf('"') == 0) {
			// 		etag = etag.substring(1, etag.length-1);
			// 	}
			// }
		}
		catch(e) {
			console.error(e);
		}
		return etag;
	},
	
	getResponseCode: function(response) {
		var responseCode = null;
		try {
			responseCode = response.ioArgs.xhr.status;
		}
		catch(e) {
			console.error(e);
		}
		return responseCode;
	},
		
	getResponseText: function(response) {
		var responseText = null;
		try {
			responseText = response.ioArgs.xhr.responseText;
		}
		catch(e) {
			console.error(e);
		}
		return responseText;
	},
	
	getResponseXML: function(response) {
		var responseXML = null;
		try {
			responseXML = response.ioArgs.xhr.responseXML;
		}
		catch(e) {
			console.error(e);
		}
		return responseXML;
	},
	
	getValidationErrorDescription: function (errorResponse) {	
		// summary:
		//		Return the xhtml validation error description, if it
		//		exists, for the given errorResponse.
		// errorResponse:
		//		deferred errorResponse 
		// return: 
		//		String or null
		var description = null;	
		if (this.getResponseCode(errorResponse) == "403") {
			// xhtml validation error returns 403 status
			var responseXML = this.getResponseXML(errorResponse);
			if (responseXML) {
				var parser = new com.ibm.rdm.web.parsers.xml2json(responseXML);
				var parsed = parser.parse();
				if (parsed.RDF && parsed.RDF.xhtmlValidationError) {
					description =
						parsed.RDF.xhtmlValidationError.description.value;
				} else {
					// IE is not able to parse the responseXml Doc
					// for some reason?
					var msg =  this.getResponseText(errorResponse);
					var startIdx = msg.indexOf("<rm:xhtmlValidationError>");
					if (startIdx > -1) {
						description = msg.substring(startIdx,
							msg.indexOf("</rm:xhtmlValidationError>"));
					}
				}
			}
		}
		return description;
	},
	
	getResponseURI: function(response) {
		var uri = null;
		try {
			uri = response.ioArgs.args.url;
		}
		catch(e) {
			console.error(e);
		}
		return uri;       
	},
	
	getHeader: function(response, header) {
		var headerValue = null;
		try {
			headerValue = response.ioArgs.xhr.getResponseHeader(header);
		}
		catch(e) {
			console.error(e);
		}
		return headerValue;
	},
	
	getResponseContentType: function(response) {
		var contentType = null;
		try {
			contentType =
				queryutils.getNormalizedContentType(
					this.getHeader(response, Headers.CONTENT_TYPE));
		}
		catch(e) {
			console.error(e);
		}
		return contentType;
	},
	
	getResponseContentLocation: function(response) {
		var contentLocation = null;
		try {
			contentLocation =
				this.getHeader(response, Headers.CONTENT_LOCATION);
		}
		catch (e) {
			console.error(e);
		}
		return contentLocation;
	},
	
	getResponseRevision: function(response) {
		// summary:
		//		Get the revision URI for the provided response. This
		//		is NOT the same as the ETag, as it was in past
		//		releases. This change was a result of Work Item
		//		62166.
		// response:
		//		XHR Reponse
		// returns:
		//		The URI of the revision.
		return this.getResponseContentLocation(response);
	},
	
	getResponseLocation: function(response) {
		var locationHeader = null;
		try {
			locationHeader = this.getHeader(response, Headers.LOCATION);
			if (locationHeader.indexOf("?") > 0) {
				locationHeader = locationHeader.substring(0,
					locationHeader.indexOf("?"));
			}
		}
		catch(e) {
			console.error(e);
		}
		return locationHeader;
	},
	
	getResponseLastModified: function(response) {
		var lastModified = null;
		try {
			lastModified = this.getHeader(response, Headers.LAST_MODIFIED);
			lastModified = dojo.date.stamp.fromISOString(
				dojo.date.stamp.toISOString(new Date(lastModified)));
		}
		catch(e) {
			console.error(e);
		}
		return lastModified;
	},

	getResponseCreated: function(response) {
		var created = null;
		try {
			created = this.getHeader(response, Headers.CREATED);
			created = dojo.date.stamp.fromISOString(
				dojo.date.stamp.toISOString(new Date(created)));
		}
		catch(e) {
			console.error(e);
		}
		return created;
	},
	
	getResponseCreator: function(response) {
		var creator = null;
		try {
			creator = this.getHeader(response, Headers.CREATOR);
		}
		catch(e) {
			console.error(e);
		}
		return creator;
	},
	
	getResponseDate: function(response) {
		var date = null;
		try {
			date = this.getHeader(response, Headers.DATE);
			date = dojo.date.stamp.fromISOString(
				dojo.date.stamp.toISOString(new Date(date)));
		}
		catch(e) {
			console.error(e);
		}
		return date;
	},
	
	getResponseLastModifiedBy: function(response) {
		var lastModifiedBy = null;
		try {
			lastModifiedBy = this.getHeader(response, Headers.LAST_MODIFIED_BY);
		}
		catch(e) {
			console.error(e);
		}
		return lastModifiedBy;
	},
	
	getResponseComponentURI: function(response) {
		var componentURI = null;
		try {
			componentURI = this.getHeader(response, Headers.OWNING_CONTEXT);
		}
		catch(e) {
			console.error(e);
		}
		try {
			if (!componentURI) {
				componentURI = this.getHeader(response, Headers.XOWNING_CONTEXT);
			}
		}
		catch(e) {
			console.error(e);
		}
		return componentURI;
	},
	
	getResponseParentFolderURI: function(response) {
		var parentFolderURI = null;
		try {
			parentFolderURI = this.getHeader(response, Headers.PARENT_FOLDER);
		}
		catch(e) {
			console.error(e);
		}
		return parentFolderURI;
	},
	
	getResponseXLastModifiedXSD: function(response) {
		var xModified = null;
		try {
			xModified = this.getHeader(response, Headers.X_LAST_MODIFIED_XSD);
		}
		catch(e) {
			console.error(e);
		}
		return xModified;
	},
	
	getResponseXRemainingTime: function(response) {
		var xRemainingTimeMillis = 0;
		try {
			xRemainingTimeMillis = this.getHeader(response,
				Headers.X_REMAINING_TIME);
			xRemainingTimeMillis = parseInt(xRemainingTimeMillis);
		}
		catch(e) {
			xRemainingTimeMillis = 0;
			console.error(e);
		}
		return xRemainingTimeMillis;
	}
};

})();
