    /*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("com.ibm.team.workitem.web.client.internal.WorkItemClient");

dojo.require("com.ibm.team.repository.web.transport.ServiceRequest");
dojo.require("com.ibm.team.repository.web.transport.ServiceResponseHandler");
dojo.require("com.ibm.team.repository.web.transport.TeamServerClient");
dojo.require("com.ibm.team.workitem.web.client.internal.QueryClient");
dojo.require("com.ibm.team.workitem.web.client.internal.SimpleCache");
dojo.require("com.ibm.team.rtc.foundation.web.InitializationDataTypes");
dojo.require("com.ibm.team.rtc.foundation.web.util.Initialization");
dojo.require("com.ibm.team.rtc.foundation.web.util.PersistentCache");
dojo.require("com.ibm.team.workitem.web.client.util");
dojo.require("com.ibm.team.workitem.api.common.DynaCacheConstants");
dojo.require("com.ibm.team.rtc.foundation.web.util.Compression");
dojo.require("jazz.app.context");
dojo.require("com.ibm.team.workitem.web.client.internal.ClientConstants");

(function() {

var ServiceRequest			= com.ibm.team.repository.web.transport.ServiceRequest;
var ServiceResponseHandler	= com.ibm.team.repository.web.transport.ServiceResponseHandler;
var TeamServerClient		= com.ibm.team.repository.web.transport.TeamServerClient;
var QueryClient				= com.ibm.team.workitem.web.client.internal.QueryClient;
var SimpleCache				= com.ibm.team.workitem.web.client.internal.SimpleCache;
var PersistentCache			= com.ibm.team.rtc.foundation.web.util.PersistentCache;
var Utils					= com.ibm.team.workitem.web.client.util;
var InitializationDataTypes = com.ibm.team.rtc.foundation.web.InitializationDataTypes;
var Initialization= com.ibm.team.rtc.foundation.web.util.Initialization;
var Compression				= com.ibm.team.rtc.foundation.web.util.Compression;
var DynaCacheConstants 		= com.ibm.team.workitem.api.common.DynaCacheConstants;
var ClientConstants 		= com.ibm.team.workitem.web.client.internal.ClientConstants;

function validateWorkItemParams(appArgs, method) {
	
	// validate that we have any request parameters
	if (!appArgs) {
		throw new Error("WorkItemClient#" + method + ": missing request parameters.");
	}
	
	// validate that we have only one of 'id', 'itemId', or 'newWorkItem'
	var count= 0;
	var setStr;
	dojo.forEach(["id", "itemId", "newWorkItem"],
		function(x) {
			if (appArgs[x]) { count++; setStr= x + " "; }
		});
	if (count > 1) {
		throw new Error("WorkItemClient#" + method +
			": may only set one of 'id', 'itemId', or 'newWorkItem' (saw: [" +
			setStr + "]");
	}
}

/**
 * Provides helper methods to fetch workitem-related data from the team server.
 */
com.ibm.team.workitem.web.client.internal.WorkItemClient= {

	// service URI
	SERVICE_URI: "com.ibm.team.workitem.common.internal.rest.IWorkItemRestService",
	
	// get methods
	GET_VALUES_METHOD: "getValues",
	GET_CONTRIBUTORS_METHOD: "getContributors",
	
	/**
	 * see #getAttributes. Returns the attributes that can be used in a call to QueryClient.getHistogram().
	 */
	FILTER_HISTOGRAM: "histogram",
	
	_persistentCache: new PersistentCache("com.ibm.team.workitem.web.client.internal.WorkItemClient", 2, 10),
	_sessionCache: new SimpleCache(),

	/*
	 * Invokes the work item REST service specified by 'method' with optional
	 * request parameters specified by 'appArgs'.  Calls back to the functions
	 * defined by 'responseHandler'.
	 */
	_invokeWorkItemService: function(responseHandler, method, appArgs, headerArgs) {

		// Wondering what this is about?
		// See: https://connect.microsoft.com/IE/feedback/details/836581/ie11-xmlhttprequest-ignores-cache-control-headers
		if(navigator.userAgent.toLowerCase().search("edge") >= 0){
			if(!appArgs){
				var appArgs = {};
			}
			appArgs.now = Date.now();
		}
						
		var headerObj = dojo.mixin(
			{}, 
			ClientConstants.CACHE_CONTROL_HEADERS, 
			headerArgs || {}
		);
		
		// appArgs is optional, so test whether or not it was sent
		var serviceRequest = appArgs ? new ServiceRequest(WorkItemClient.SERVICE_URI, method, appArgs) 
									 : new ServiceRequest(WorkItemClient.SERVICE_URI, method);
		
		TeamServerClient.invokeService(serviceRequest, responseHandler, {headers: headerObj});
	},
	
	/*
	 * Invokes the process REST service specified by 'method' with optional
	 * request parameters specified by 'appArgs'.  Calls back to the functions
	 * defined by 'responseHandler'.
	 */
	_invokeProcessService: function(responseHandler, method, appArgs) {
		var request= new ServiceRequest("com.ibm.team.process.internal.service.web.IProcessWebUIService", method, appArgs); //$NON-NLS-1$
		TeamServerClient.invokeService(request, responseHandler);
	},
	
	_invokeCachingWorkItemService: function(responseHandler, method, appArgs, cacheKey, ttl) {
		var pId= null;
		var pStateId= null;
		if (appArgs != null) {
			pId= appArgs.projectAreaItemId;
			pStateId= appArgs.projectAreaConfigurationStateId;
		}
		if (pStateId != null) {
			pStateId= pStateId + "-" + dojo.locale;
		}
		var self= this;
		this._invokeCachingService(responseHandler, pId, pStateId, cacheKey, ttl, function(handler) {
			self._invokeWorkItemService(handler, method, appArgs);
		});
	},
	
	_invokeCachingService: function(responseHandler, projectAreaItemId, stateId, cacheKey, ttl, actualMethod) {
		
		var entry= this._getFromCache(projectAreaItemId, stateId, cacheKey, ttl);
		if (entry != null) {
			responseHandler.doCallback(entry);
			return;
		}

		var self= this;
		var internalHandler= {
			success: function(result) {
				self._updateCache(result, projectAreaItemId, stateId, cacheKey, ttl);
				responseHandler.doCallback(dojo.clone(result));
			},
			failure: function(error) {
				responseHandler.doErrorCallback(error);
			}
		};
		
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		actualMethod(handler);
	},

	_getFromCache: function(projectAreaItemId, stateId, cacheKey, ttl) {
		var sessionKey= cacheKey;
		if (projectAreaItemId != null) {
			sessionKey= cacheKey + "/projectAreaItemId:"  + projectAreaItemId; 
		}

		var sessionEntry= this._sessionCache.get(sessionKey);
		if (sessionEntry != null) {
			return dojo.clone(sessionEntry);
		}

		if (stateId != null) {
			var entry= this._persistentCache.getEntry(projectAreaItemId, stateId);
			if (entry != null && entry[cacheKey] != null) {
				var rawValue = entry[cacheKey];
				var value= this._isMSEdge() ? rawValue : Compression.decompress(rawValue);

				this._sessionCache.put(sessionKey, value, ttl);
				return dojo.clone(value);
			}
		}
		return null;
	},

	// dojo.isIE does not detect MS Edge. This method should be replaced with proper dojo.isIE
	// check if/when we switch to a version of dojo that properly detects MS Edge.
	_isMSEdge: function() {		
		var ua = window.navigator.userAgent;
		var index = ua.indexOf('Edge/');
		return (index >= 0);
	},
	
	_updateCache: function(value, projectAreaItemId, stateId, cacheKey, ttl) {
		if (ttl == SimpleCache.TTL_INFINITE && projectAreaItemId != null && stateId != null) {
			var e= this._persistentCache.getEntry(projectAreaItemId, stateId);
			if (e == null) {
				e= {};
			}
			e[cacheKey]= this._isMSEdge() ? value : Compression.compress(value);
			this._persistentCache.putEntry(projectAreaItemId, stateId, e);
		}
		var sessionKey= cacheKey;
		if (projectAreaItemId != null) {
			sessionKey= cacheKey + "/projectAreaItemId:"  + projectAreaItemId; 
		}
		this._sessionCache.put(sessionKey, value, ttl);
	},

/*
 * Get the project area item id, either from the args or the default
 */
	_setAppArgsProjectAreaItemId: function( appArgs) {
		if (!appArgs) appArgs= {};
		if (appArgs.projectAreaConfigurationStateId === null) {
			delete appArgs.projectAreaConfigurationStateId;
		}
		try {
			if(appArgs.projectAreaItemId) {
				appArgs.projectAreaItemId = appArgs.projectAreaItemId.replace(/\"\"/gi, "\"" ); // TODO: find source of double quotes
				return;
			}
			var context = jazz.app.context.get();
			if(context.itemId){
				// context isn't always a project area
				appArgs.projectAreaItemId = context.itemId;
			}
		} finally {
			if (appArgs.projectAreaConfigurationStateId == null) {
				var data= Initialization.getData(InitializationDataTypes.PROJECT_AREA_INITIALIZATION_DATA);
				if (data != null) {
					var projectArea= data.projectArea;
					if (projectArea.id == appArgs.projectAreaItemId && projectArea.configurationStateId != null) {
						appArgs.projectAreaConfigurationStateId= projectArea.configurationStateId;
					}
				}
			}
		}
	},
	/*
	 * 
	 */
	getWorkItemEditableProperties: function(responseHandler, appArgs) {
		validateWorkItemParams(appArgs, "getWorkItemEditableProperties");
		if (appArgs.newWorkItem) {
			this._setAppArgsProjectAreaItemId(appArgs); // TODO:newonly?
		}
		var dynaCacheTokenArgs = {};
		if (appArgs.teamAreaItemId) {
			dynaCacheTokenArgs["projectAreaItemId"] = appArgs.teamAreaItemId;
			delete appArgs["teamAreaItemId"];
		} else {
			dynaCacheTokenArgs["projectAreaItemId"] = appArgs.projectAreaItemId;
		}
		this._getDynaCacheTokenDeferred(dynaCacheTokenArgs).then(dojo.hitch(this, function(results) {
			// Do not use the server cache when changing type
			var headerArgs = {};
			if (results != null  && !appArgs.typeId) {
				headerArgs[DynaCacheConstants.WORK_ITEM_CACHE_TOKEN_HEADER] = results;
			}
			if (appArgs[DynaCacheConstants.WORK_ITEM_CACHE_TYPE_ID_HEADER]) {
				headerArgs[DynaCacheConstants.WORK_ITEM_CACHE_TYPE_ID_HEADER] = appArgs[DynaCacheConstants.WORK_ITEM_CACHE_TYPE_ID_HEADER];
				delete appArgs[DynaCacheConstants.WORK_ITEM_CACHE_TYPE_ID_HEADER];
			}
			if (appArgs[DynaCacheConstants.WORK_ITEM_CACHE_STATE_ID_HEADER]) {
				headerArgs[DynaCacheConstants.WORK_ITEM_CACHE_STATE_ID_HEADER] = appArgs[DynaCacheConstants.WORK_ITEM_CACHE_STATE_ID_HEADER];
				delete appArgs[DynaCacheConstants.WORK_ITEM_CACHE_STATE_ID_HEADER];
			}
			WorkItemClient._invokeWorkItemService(responseHandler, "getWorkItemEditableProperties", appArgs, headerArgs);
		}));
	},
	
	getConfigurationUriForWorkItemByLinkType: function(responseHandler, appArgs) {
		WorkItemClient._invokeWorkItemService(responseHandler, "getConfigurationUriForWorkItemByLinkType", appArgs);
	},
	
	getHasOSLCLinkTypeMappings: function(responseHandler, appArgs) {
		WorkItemClient._invokeWorkItemService(responseHandler, "getHasOSLCLinkTypeMappings", appArgs);
	},
	
	getAllValues: function(responseHandler, appArgs) {
		validateWorkItemParams(appArgs, "postGetAllValues");
		if (appArgs.newWorkItem) {
			this._setAppArgsProjectAreaItemId(appArgs); // TODO:newonly?
		}
		if (!appArgs.ids) {
			appArgs.ids= [];
		}
		WorkItemClient._invokeWorkItemService(responseHandler, "postGetAllValues", appArgs);
	},
	
	/**
	 * Returns true if the work item's state stored in repository is different from the state in provided in the 
	 * <code>parms</code> object. False is returned otherwise.
	 * 
	 * The work item's state is considered to be different if an attribute value or a link of the work item 
	 * was changed. Even link changes via other work items are tracked by this method.
	 * 
	 * @param responseHandler
	 * 			Handler called when the server request returns.
	 * @param parms
	 * 			Parameter object representing the client's work item, its state and link values
	 */
	getHasWorkItemChanged: function(responseHandler, appArgs) {
		WorkItemClient._invokeWorkItemService(responseHandler, "getHasWorkItemChanged", appArgs);
	},

	getMaxAttachmentSize: function(responseHandler) {
		var appArgs= {};
		this._setAppArgsProjectAreaItemId(appArgs);
		WorkItemClient._invokeCachingWorkItemService(responseHandler, "getMaxAttachmentSize", appArgs, "getMaxAttachmentSize",  SimpleCache.TTL_INFINITE);
	},
	
	guessCategory: function(responseHandler, workingCopy) {
		var messages= dojo.i18n.getLocalization("com.ibm.team.workitem.web.ui.internal", "WidgetMessages");
		var itemSummary= dojo.trim(workingCopy.getValue({ path: "attributes.summary.content" }));
		var itemContent= dojo.trim(workingCopy.getValue({ path: "attributes.description.content" }));
		
		var itemComments= workingCopy.getValue({ path: "attributes.internalComments.items" });
		dojo.forEach(itemComments, function(entry) {
			itemContent += ("\n" + dojo.trim(entry.content));
		});
		
		var currentComment= workingCopy.getValue({ path: ["attributes", workingCopy.SYNTHETIC_COMMENT_ID, "content"] });
		if(currentComment) 
			itemContent += currentComment;
		
		var workItemId= workingCopy.getValue({ path: "itemId" });
		var projectAreaId= workingCopy.getValue({ path: "attributes.projectArea.id" });
		
		var serviceRequest = new ServiceRequest("com.ibm.team.workitem.common.internal.rest.IWorkItemRestService", "postGuessCategory", 
			{ 
				projectAreaItemId : projectAreaId,
				workItemId : workItemId,
				itemSummary : itemSummary,
				itemContent : itemContent
			}
		);
		
		var op = workingCopy.createOperation({
	        message: messages.labelGuessingCategory,
	        id: "guessingCategory"
		});
			
	    var handlerObj = {
		    success: function(result){
					workingCopy.finishOperation(op);
					responseHandler.doCallback(result);
		    },
		    failure: function(errorObj){
		    	var message= "";
		    	responseHandler.doErrorCallback(message);
				workingCopy.finishOperation(op);
		    }
	    }
	    
	    var handler = new ServiceResponseHandler(handlerObj, "success", "failure");
		workingCopy.startOperation(op);
	    TeamServerClient.invokeService(serviceRequest, handler);
	},
	
	getAllQueryableValues: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		WorkItemClient._invokeWorkItemService(responseHandler, "getAllQueryableValues", appArgs);
	},
	
	getValues: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		WorkItemClient._invokeWorkItemService(responseHandler, "getValues", appArgs);
	},
	
	getTime: function(responseHandler, appArgs) {
		WorkItemClient._invokeWorkItemService(responseHandler, "getTime", appArgs);
	},
		
	/**
	 * See IWorkItemRestService#getWorkItemDTO.
	 */
	getWorkItemDTO: function(responseHandler, appArgs) {
		validateWorkItemParams(appArgs, "getWorkItemDTO");
		if (appArgs.newWorkItem) {
			this._setAppArgsProjectAreaItemId(appArgs); // TODO:newonly?
		}
		this._invokeWorkItemService(responseHandler, "getWorkItemDTO", appArgs);
	},
	
	getDynaCacheToken: function(responseHandler, appArgs) {
		var ttl= SimpleCache.TTL_MINUTE * 30;
		this._invokeCachingWorkItemService(responseHandler, "getDynaCacheToken", appArgs, "getDynaCacheToken", ttl);
	},
	
	_getDynaCacheTokenDeferred: function(appArgs) {
		var deferredCacheToken = (new (require("dojo/Deferred"))); //$NON-NLS-1$
		var internalHandler= {
				success: function(token) {	
					deferredCacheToken.resolve(token);
				},
				failure: function(error) {
					var exception= TeamRepositoryExceptionFactory.INSTANCE.createException(error);
					console.log(exception.getMessage());
					deferredCacheToken.resolve(null);
				}
			};
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		this.getDynaCacheToken(handler,appArgs);
		return deferredCacheToken;
	},
	
	getDynaCacheWorkItemStateId: function(appArgs) {
		var deferredDynaCacheWorkItemStateId = (new (require("dojo/Deferred"))); //$NON-NLS-1$
		if (appArgs.newWorkItem) {
			deferredDynaCacheWorkItemStateId.resolve(null);
			return deferredDynaCacheWorkItemStateId;
		}
		
		var internalHandler= {
				success: function(stateId) {	
					deferredDynaCacheWorkItemStateId.resolve(stateId);
				},
				failure: function(error) {
					deferredDynaCacheWorkItemStateId.resolve(null);
					console.debug(error);
				}
			};
		var dynaCacheWorkItemStateIdArgs = {
				id: appArgs.id
		};
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		this._invokeWorkItemService(handler, "getDynaCacheWorkItemStateId", dynaCacheWorkItemStateIdArgs);
		return deferredDynaCacheWorkItemStateId;
	},
	
	/**
	 * See IWorkItemRestService#getWorkItemDTO2.
	 */
	getWorkItem: function(responseHandler, appArgs) {
		if (appArgs.newWorkItem) {
			this._setAppArgsProjectAreaItemId(appArgs); // TODO:newonly?
		}
		var internalHandler= {
				success: function(workItemDTO) {
					if(workItemDTO == null) {
						responseHandler.doCallback(null);
						return;
					}
					var attributeMap= Utils.buildMap(workItemDTO.attributes)
					workItemDTO.attributes= attributeMap;
					responseHandler.doCallback(workItemDTO);
				},
				failure: function(error) {
					responseHandler.doErrorCallback(error);
				}
			};
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		this.getDynaCacheWorkItemStateId(appArgs).then(dojo.hitch(this, function(stateId) {
			if(stateId != null) {
				var headerArgs = {};
				headerArgs[DynaCacheConstants.WORK_ITEM_CACHE_STATE_ID_HEADER] = stateId;
			}
			this._invokeWorkItemService(handler, "getWorkItemDTO2", appArgs, headerArgs);
		}));
	},
	
	/**
	 * See IWorkItemRestService#postWorkItem2.
	 */
	postWorkItem: function(responseHandler, appArgs) {
		if (!appArgs.stateId && !appArgs.projectAreaItemId) {
			this._setAppArgsProjectAreaItemId(appArgs);
		}
		this._internalPostWorkItem("postWorkItem2", responseHandler, appArgs);
	},

	postWorkItemN: function(responseHandler, appArgsList) {
		var serializedAppArgs= dojo.map(appArgsList, function(appArgs) {
			if (!appArgs.stateId && !appArgs.projectAreaItemId) {
				this._setAppArgsProjectAreaItemId(appArgs);
			}
				
			appArgs.attributeValues= dojo.map(appArgs.attributeValues, function(value) {
				if (dojo.isArray(value))
					return dojo.toJson(value);
				return value;
			});

			return dojo.toJson(appArgs);
		}, this);

		var internalHandler= {
			success: function(operationReportDTO) {
				dojo.forEach(operationReportDTO.workItems, function(workItemDTO) {
					workItemDTO.attributes= Utils.buildMap(workItemDTO.attributes)
				});
				
				responseHandler.doCallback(operationReportDTO);
			},
			failure: function(error) {
				console.debug(error);
				responseHandler.doErrorCallback(error);
			}
		};
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		this._invokeWorkItemService(handler, "postWorkItemN", { postWorkItemParams: serializedAppArgs });
	},

	/**
	 * See IWorkItemRestService#postMoveWorkItem.
	 */
	moveWorkItem: function(responseHandler, appArgs) {
		if (appArgs.move == null|| !appArgs.projectAreaItemId) {
			throw new Error("WorkItemClient#moveWorkItem: 'move' (true/false) and 'projectAreaItemId' must be set");
		}
		this._internalPostWorkItem("postMoveWorkItem", responseHandler, appArgs);
	},

	getUsesTimeRemaining: function(responseHandler, appArgs) {
		appArgs = appArgs || {};
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getUsesTimeRemaining", appArgs);
	},
	
	/**
	 * See IWorkItemRestService#postSplitWorkItem.
	 */
	splitWorkItem: function(responseHandler, appArgs) {
    if (!appArgs.stateId && !appArgs.projectAreaItemId) {
      this._setAppArgsProjectAreaItemId(appArgs);
    }
		this._internalPostWorkItem("postSplitWorkItem", responseHandler, appArgs);
	},
	
	_internalPostWorkItem: function(serviceName, responseHandler, appArgs) {
		
		appArgs.attributeValues= dojo.map(appArgs.attributeValues, function(value) {
			if (dojo.isArray(value))
				return dojo.toJson(value);
			return value;
		});
		
		var internalHandler= {
			success: function(operationReportDTO) {
				if (operationReportDTO.severity <= WorkItemClient.INFO) {
					var workItemDTO= operationReportDTO.workItem;
					if(operationReportDTO.infos) {
						workItemDTO.infos = operationReportDTO.infos;
					}
					var attributeMap= Utils.buildMap(workItemDTO.attributes)
					workItemDTO.attributes= attributeMap;
					responseHandler.doCallback(workItemDTO);
				} else {
					responseHandler.doErrorCallback(operationReportDTO);
				}
			},
			failure: function(error) {
				console.debug(error);
				responseHandler.doErrorCallback(error);
			}
		};
		var handler= new ServiceResponseHandler(internalHandler, "success", "failure");
		this._invokeWorkItemService(handler, serviceName, appArgs);
	},
	
	/**
	 * See IWorkItemRestService#postBulkChangeWorkItems.
	 * @param appArgs
	 * 		itemIds: an array of work item itemIds
	 * 		attributeIdentifiers: the identifiers of the attributes to be changed
	 * 		attributeValues: the new values for the attributes. Can be JS arrays of simple values.
	 */
	postBulkChangeWorkItems: function(responseHandler, appArgs) {
		
		appArgs.attributeValues= dojo.map(appArgs.attributeValues, function(value) {
			if (dojo.isArray(value))
				return dojo.toJson(value);
			return value;
		});
		
		this._invokeWorkItemService(responseHandler, "postBulkChangeWorkItems", appArgs);
	},
	
	getResolutions: function(responseHandler, appArgs) {
		if (!(appArgs.status || appArgs.actionId)) {
			throw new Error("WorkItemClient#getResolutions: must set either appArgs.status or appArgs.actionId (none set)");
		}
		if (appArgs.status && appArgs.actionId) {
			throw new Error("WorkItemClient#getResolutions: must set either appArgs.status or appArgs.actionId (both set)");
		}
		
		this._setAppArgsProjectAreaItemId(appArgs);
		appArgs.attributeName= "internalResolution";
		this._invokeWorkItemService(responseHandler, WorkItemClient.GET_VALUES_METHOD, appArgs);
	},
	
	getContributors: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, WorkItemClient.GET_CONTRIBUTORS_METHOD, appArgs);
	},
	
	getTeamMembers: function(responseHandler, appArgs) {
		 appArgs = appArgs || {};
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getTeamMembers", appArgs);
	},
	
	getEditorPresentation: function(responseHandler, appArgs) {
		if (!appArgs.typeId && !appArgs.itemId) {
			throw new Error("WorkItemClient#getEditorPresentation: must set appArgs.typeId (the work item type id) or appArgs.itemId (the work item itemId)");
		}
		this._setAppArgsProjectAreaItemId(appArgs);
		
		var cacheKey = "getEditorPresentation/typeId:" + appArgs.typeId + "/target:" + (appArgs.target ? appArgs.target : "(none)");  			

		this._invokeCachingWorkItemService(responseHandler, "getEditorPresentation", appArgs, cacheKey,  SimpleCache.TTL_INFINITE);

	},
	
	getQuickInformationPresentation: function(responseHandler, appArgs) {
		if (!appArgs.configurationId) {
			throw new Error("WorkItemClient#getQuickInformationPresentation: must set appArgs.configurationId");
		}
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeCachingWorkItemService(responseHandler, "getQuickInformationPresentation", appArgs, ("getQuickInformationPresentation/configurationId:"+ appArgs.configurationId + "/workItemTypeId:"+ appArgs.workItemTypeId),  SimpleCache.TTL_INFINITE);
	},
	
	getDynamicAttributesPropertiesConfiguration: function(responseHandler, appArgs, projectAreaStateId) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeCachingWorkItemService(responseHandler, "getDynamicAttributesPropertiesConfiguration", appArgs, "getDynamicAttributesPropertiesConfiguration:" + projectAreaStateId, SimpleCache.TTL_INFINITE);
	},
		
	getApprovalStates: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeCachingWorkItemService(responseHandler, "getApprovalStates", appArgs, "getApprovalStates", SimpleCache.TTL_INFINITE);
	},
		
	getApprovalTypes: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeCachingWorkItemService(responseHandler, "getApprovalTypes", appArgs, "getApprovalTypes", SimpleCache.TTL_INFINITE);
	},
		
	getLinkTypes: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeCachingWorkItemService(responseHandler, "getLinkTypesForProject", appArgs, "getLinkTypesForProject", SimpleCache.TTL_INFINITE);
	},
	
	getWorkItemTypes: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		appArgs.newWorkItem= true;
		appArgs.ids= ["workItemType"];
		this._invokeCachingWorkItemService(responseHandler, "getAllValues", appArgs, "getWorkItemTypes", SimpleCache.TTL_INFINITE);
	},
	
	getReleases: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getReleases", appArgs);
	},
	
	getWorkItemCategories: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getWorkItemCategories", appArgs);
	},
	
	getWorkItemTimeLines: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getWorkItemTimeLines", appArgs);
	},
	
	getWorkItemTeamAreas: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getWorkItemTeamAreas", appArgs);
	},
	
	postDeleteWorkItems: function(responseHandler, appArgs) {
		this._invokeWorkItemService(responseHandler, "postDeleteWorkItems", appArgs);
	},
	
	postReleases: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "postReleases", appArgs);
	},
	
	postWorkItemCategories: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "postWorkItemCategories", appArgs);
	},

	/*
	 * Post multiple work items
	 * { 
	 *     s: The string form of an {@link IWorkItem} "[itemId];[stateId]".
	 *     a:  list of attributes to return. An empty or null array means all attributes
	 * 	   includeLinks: false: no links will be included (default=false)
	 * 	   includeApprovals: false: no approvals will be included (default=false)
	 * }
	 */
	postGetWorkItems: function(responseHandler, appArgs) {
 		this._setAppArgsProjectAreaItemId(appArgs);
 		this._invokeWorkItemService(responseHandler, "postGetWorkItems", appArgs);
	},
	
	postGetWorkItemBreakdownStructure: function(responseHandler, appArgs) {
 		this._setAppArgsProjectAreaItemId(appArgs);
 		this._invokeWorkItemService(responseHandler, "postGetWorkItemBreakdownStructure", appArgs);
	},
	
	getEnumerationValues: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		this._invokeWorkItemService(responseHandler, "getEnumerationValues", appArgs);
	},
	
	/**
	 * Returns all work item attributes of a project area
	 * @param appArgs.projectAreaItemId the item id of the project area (required)
	 * @param appArgs.filter an optional filter. Currently supported: WorkItemClient.FILTER_HISTOGRAM
	 */
	getAttributes: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		if (appArgs.forceFetch) {
			this._invokeWorkItemService(responseHandler, "getAttributes", appArgs);
		} else {
			var key= "getAttributes/" + dojo.toJson(appArgs);
			this._invokeCachingWorkItemService(responseHandler, "getAttributes", appArgs, key, SimpleCache.TTL_INFINITE);	
		}
	},
	
	getProcessAreaHierarchies: function(responseHandler, appArgs) {
		this._setAppArgsProjectAreaItemId(appArgs);
		var cacheKey= "getProcessAreaHierarchies";
		var ttl= SimpleCache.TTL_MINUTE * 30;
		
		this._invokeCachingWorkItemService(responseHandler, "getProcessAreaHierarchies", appArgs, cacheKey, ttl);
	},
	
	/**
	 * Returns all iterations of the project area. Optionally included archived ones
	 * @param appArgs.projectAreaItemId the item id of the project area (required)
	 * @param appArgs.includeArchived if 'true', also returns the archived iterations
	 */ 
	getIterations: function(responseHandler, appArgs) {
		var ttl= SimpleCache.TTL_MINUTE * 30;
		this._invokeCachingWorkItemService(responseHandler, "getIterations", appArgs, "getIterations", ttl);
	},
	
	/**
	 * For a given project area returns a tree of timelines and child iterations which could be presented to the user. This should include the undefined iteration and backlog iterations as specified for each timeline.
	 * @param appArgs.projectAreaItemId the item id of the project area (required)
	 * @param appArgs.includeArchived if 'true', also returns the archived iterations
	 */ 
	getTimelinesWithIterations: function(responseHandler, appArgs) {
		var ttl= SimpleCache.TTL_MINUTE * 30;
		this._invokeCachingWorkItemService(responseHandler, "getTimelinesWithIterations", appArgs, "getTimelinesWithIterations", ttl);
	},
	
	clearIterations: function(projectAreaItemId) {
		this._sessionCache.clear("getIterations/projectAreaItemId:" + projectAreaItemId);
	},
	
	transformWikiSource: function(responseHandler, appArgs) {
		this._invokeWorkItemService(responseHandler, "postWikiSource", appArgs);
	},
	
	getAllProjectAreas: function(responseHandler, appArgs) {
		this._invokeWorkItemService(responseHandler, "getAllProjectAreas", appArgs);
	},
	
	getAccessContexts: function(responseHandler, appArgs) {
		this._invokeWorkItemService(responseHandler, "getAccessContexts", appArgs);
	},
	
	isAutoLinkArchivingEnabled: function(responseHandler, appArgs) {
        this._invokeWorkItemService(responseHandler, "getIsAutoLinkArchivingEnabled", appArgs);
    },
    
    getIsContributorMemberOf: function(responseHandler, appArgs) {
        this._invokeWorkItemService(responseHandler, "getIsContributorMemberOf", appArgs);
    }
};

// must be defined after com.ibm.team.workitem.web.client.internal.WorkItemClient
var WorkItemClient= com.ibm.team.workitem.web.client.internal.WorkItemClient;

// See IProcessReport
WorkItemClient.INFO= 0;
WorkItemClient.WARNING= 2;
WorkItemClient.ERROR= 4;

})();